(* Copyright (C) 2014, Daniel Wyckoff, except for the portions so labeleled
which I got from Daniel Schepler*)



(*		   GNU LESSER GENERAL PUBLIC LICENSE
                       Version 3, 29 June 2007

 Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.


  This version of the GNU Lesser General Public License incorporates
the terms and conditions of version 3 of the GNU General Public
License, supplemented by the additional permissions listed below.

  0. Additional Definitions. 

  As used herein, "this License" refers to version 3 of the GNU Lesser
General Public License, and the "GNU GPL" refers to version 3 of the GNU
General Public License.

  "The Library" refers to a covered work governed by this License,
other than an Application or a Combined Work as defined below.

  An "Application" is any work that makes use of an interface provided
by the Library, but which is not otherwise based on the Library.
Defining a subclass of a class defined by the Library is deemed a mode
of using an interface provided by the Library.

  A "Combined Work" is a work produced by combining or linking an
Application with the Library.  The particular version of the Library
with which the Combined Work was made is also called the "Linked
Version".

  The "Minimal Corresponding Source" for a Combined Work means the
Corresponding Source for the Combined Work, excluding any source code
for portions of the Combined Work that, considered in isolation, are
based on the Application, and not on the Linked Version.

  The "Corresponding Application Code" for a Combined Work means the
object code and/or source code for the Application, including any data
and utility programs needed for reproducing the Combined Work from the
Application, but excluding the System Libraries of the Combined Work.

  1. Exception to Section 3 of the GNU GPL.

  You may convey a covered work under sections 3 and 4 of this License
without being bound by section 3 of the GNU GPL.

  2. Conveying Modified Versions.

  If you modify a copy of the Library, and, in your modifications, a
facility refers to a function or data to be supplied by an Application
that uses the facility (other than as an argument passed when the
facility is invoked), then you may convey a copy of the modified
version:

   a) under this License, provided that you make a good faith effort to
   ensure that, in the event an Application does not supply the
   function or data, the facility still operates, and performs
   whatever part of its purpose remains meaningful, or

   b) under the GNU GPL, with none of the additional permissions of
   this License applicable to that copy.

  3. Object Code Incorporating Material from Library Header Files.

  The object code form of an Application may incorporate material from
a header file that is part of the Library.  You may convey such object
code under terms of your choice, provided that, if the incorporated
material is not limited to numerical parameters, data structure
layouts and accessors, or small macros, inline functions and templates
(ten or fewer lines in length), you do both of the following:

   a) Give prominent notice with each copy of the object code that the
   Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the object code with a copy of the GNU GPL and this license
   document.

  4. Combined Works.

  You may convey a Combined Work under terms of your choice that,
taken together, effectively do not restrict modification of the
portions of the Library contained in the Combined Work and reverse
engineering for debugging such modifications, if you also do each of
the following:

   a) Give prominent notice with each copy of the Combined Work that
   the Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the Combined Work with a copy of the GNU GPL and this license
   document.

   c) For a Combined Work that displays copyright notices during
   execution, include the copyright notice for the Library among
   these notices, as well as a reference directing the user to the
   copies of the GNU GPL and this license document.

   d) Do one of the following:

       0) Convey the Minimal Corresponding Source under the terms of this
       License, and the Corresponding Application Code in a form
       suitable for, and under terms that permit, the user to
       recombine or relink the Application with a modified version of
       the Linked Version to produce a modified Combined Work, in the
       manner specified by section 6 of the GNU GPL for conveying
       Corresponding Source.

       1) Use a suitable shared library mechanism for linking with the
       Library.  A suitable mechanism is one that (a) uses at run time
       a copy of the Library already present on the user's computer
       system, and (b) will operate properly with a modified version
       of the Library that is interface-compatible with the Linked
       Version. 

   e) Provide Installation Information, but only if you would otherwise
   be required to provide such information under section 6 of the
   GNU GPL, and only to the extent that such information is
   necessary to install and execute a modified version of the
   Combined Work produced by recombining or relinking the
   Application with a modified version of the Linked Version. (If
   you use option 4d0, the Installation Information must accompany
   the Minimal Corresponding Source and Corresponding Application
   Code. If you use option 4d1, you must provide the Installation
   Information in the manner specified by section 6 of the GNU GPL
   for conveying Corresponding Source.)

  5. Combined Libraries.

  You may place library facilities that are a work based on the
Library side by side in a single library together with other library
facilities that are not Applications and are not covered by this
License, and convey such a combined library under terms of your
choice, if you do both of the following:

   a) Accompany the combined library with a copy of the same work based
   on the Library, uncombined with any other library facilities,
   conveyed under the terms of this License.

   b) Give prominent notice with the combined library that part of it
   is a work based on the Library, and explaining where to find the
   accompanying uncombined form of the same work.

  6. Revised Versions of the GNU Lesser General Public License.

  The Free Software Foundation may publish revised and/or new versions
of the GNU Lesser General Public License from time to time. Such new
versions will be similar in spirit to the present version, but may
differ in detail to address new problems or concerns.

  Each version is given a distinguishing version number. If the
Library as you received it specifies that a certain numbered version
of the GNU Lesser General Public License "or any later version"
applies to it, you have the option of following the terms and
conditions either of that published version or of any later version
published by the Free Software Foundation. If the Library as you
received it does not specify a version number of the GNU Lesser
General Public License, you may choose any version of the GNU Lesser
General Public License ever published by the Free Software Foundation.

  If the Library as you received it specifies that a proxy can decide
whether future versions of the GNU Lesser General Public License shall
apply, that proxy's public statement of acceptance of any version is
permanent authorization for you to choose that version for the
Library.*)


(* Todo: Remove unnecessary paramaters, turn le into an operator *)
(* todo: add left and right for zero_prod, zero_sum, etc. *)
(* todo: add to const_eq to allow for adjacent idempotent equalities*)

Require Export Bool.
Require Export Ensembles.
Require Export Constructive_sets.
Require Export Setoid.
Require Import List.


(* I use two records for notational reasons; I haven't figured out if 
   I can combine Bconst and Bool_Alg using notations that don't
   get confused with Arith.*)
Record Bconst : Type :=
  {Btype : Type;
   BS : Ensemble Btype;
   Bplus : Btype -> Btype -> Btype;
   Btimes : Btype -> Btype -> Btype;
   Bone : Btype;
   Bzero : Btype;
   Bcomp : Btype -> Btype}.

Infix "+" := (Bplus _) (at level 50, left associativity).
Infix "*" := (Btimes _) (at level 40, left associativity).

Notation "0" := (Bzero _).
Notation "1" := (Bone _).
Notation "- x" := ((Bcomp _) x).

Record Bool_Alg : Type :=
  {Bc : Bconst;   
   und_set : (BS Bc) = Full_set (Btype Bc);
   assoc_sum : forall n m p:(Btype Bc), n + (m + p) = n + m + p;
   assoc_prod : forall n m p:(Btype Bc), n * (m * p) = n * m * p;
   comm_sum : forall n m:(Btype Bc), n + m = m + n;
   comm_prod : forall n m:(Btype Bc), n * m = m * n;
   abs_sum  : forall n m:(Btype Bc), n + (n * m) = n;
   abs_prod:  forall n m:(Btype Bc), n * (n + m) = n;
   dist_sum : forall n m p:(Btype Bc), p*(n + m) = p * n + p * m;
   dist_prod: forall n m p:(Btype Bc), p+(n * m) = (p+n) * (p+m);
   comp_sum:  forall n:(Btype Bc), n + (- n) = 1;
   comp_prod: forall n:(Btype Bc), n * (- n) = 0}.


(* Transcribed from Coq'Art 16.3.1 *)
(* Binary tree representation of boolean algebra expression *)
Hint Resolve assoc_sum assoc_prod comm_sum comm_prod abs_sum
             abs_prod dist_sum dist_prod comp_sum comp_prod.
             

(*Section Basics.*)
Module Type BAt.
Parameter B: Bool_Alg.
Definition Bt := (Btype (Bc B)).
End BAt.

Module Bscs (BA:BAt).
Import BA.

(*Definition Bp := (Bplus (Bc B)).
Definition Bm := (Btimes (Bc B)).  (* The m stands for meet, since Bt was already taken *)*)

Section assoc.

(*************)
(* All these functions/tactics in "Section assoc" and "Section commut" 
  are adapted from the one-operation versions in Coq'Art Section 16.3.*)

(* Binary tree with two types of nodes corresponding to two binary operations, plus and times,
   and a leaf which contains an integer rank corresponding to its position in a derived
   list.  *)
Inductive bin : Set := pnode : bin -> bin -> bin | 
                        tnode : bin -> bin -> bin |
                        leaf : nat-> bin.

Fixpoint flatten_aux_p (p fin:bin){struct p} : bin :=
  match p with
  | pnode x1 x2 => flatten_aux_p x1 (flatten_aux_p x2 fin)
  | x => pnode x fin
  end.

Fixpoint flatten_aux_t (t fin:bin){struct t} : bin :=
  match t with
  | tnode x1 x2 => flatten_aux_t x1 (flatten_aux_t x2 fin)
  | x => tnode x fin
  end.

Fixpoint flatten (t:bin) : bin :=
  match t with
  | pnode (tnode x1 x2 as y1) (tnode x3 x4 as y2) => pnode (flatten y1) (flatten y2)
  | tnode (pnode x1 x2 as y1) (pnode x3 x4 as y2) => tnode (flatten y1) (flatten y2)
  | tnode x1 x2 => flatten_aux_t (flatten x1) (flatten x2)
  | pnode x1 x2 => flatten_aux_p (flatten x1) (flatten x2)
  | x => x
  end.

Fixpoint bin_Bt (l: list Bt)(def:Bt)(t:bin){struct t}: Bt :=
  match t with
  | pnode t1 t2 => (bin_Bt l def t1) + (bin_Bt l def t2)
  | tnode t1 t2 => (bin_Bt l def t1) * (bin_Bt l def t2)
  | leaf n=> nth n l def
  end.

Theorem flatten_aux_valid_Bt_p :
  forall (l:list Bt)(def:Bt)(t t':bin),
  (bin_Bt l def t) + (bin_Bt l def t') = 
  bin_Bt l def (flatten_aux_p t t').
intros l def t.  elim t; simpl; auto. intros t1 iht1 t2 iht2 t'.
rewrite <- iht1. rewrite <- iht2. rewrite (assoc_sum B).
reflexivity.
Qed.

Theorem flatten_aux_valid_Bt_t :
  forall (l:list Bt)(def:Bt)(t t':bin),
  (bin_Bt l def t) * (bin_Bt l def t') =
  bin_Bt l def (flatten_aux_t t t').
intros l def t.  elim t; simpl; auto.  intros t1 iht1 t2 iht2 t'.
rewrite <- iht1.  rewrite <- iht2. rewrite (assoc_prod B).
reflexivity.
Qed.

Theorem flatten_valid_Bt :
  forall (l:list Bt)(def:Bt)(t:bin),
  bin_Bt l def t = bin_Bt l def (flatten t).
intros l def t. elim t; simpl; auto.
intros t1 iht1 t2 iht2. destruct t1. destruct t2.
rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_p. reflexivity.
rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_p. reflexivity.
rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_p. reflexivity.
destruct t2.  rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_p. reflexivity.
rewrite iht1. rewrite iht2. unfold bin_Bt at 3. fold bin_Bt. reflexivity.
rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_p. reflexivity.
rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_p. reflexivity.


intros t1 iht1 t2 iht2. destruct t1. destruct t2.
unfold bin_Bt at 3.  fold bin_Bt.
rewrite <- iht1; rewrite <- iht2; reflexivity.
rewrite iht1; rewrite iht2; rewrite flatten_aux_valid_Bt_t;  reflexivity.
rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_t.  reflexivity.
rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_t.  reflexivity.
rewrite iht1. rewrite iht2. rewrite flatten_aux_valid_Bt_t.  reflexivity.
Qed.



Theorem flatten_valid_Bt_2 :
  forall (t t':bin)(l:list Bt)(def:Bt),
    bin_Bt l def (flatten t) = bin_Bt l def (flatten t') ->
    bin_Bt l def t = bin_Bt l def t'.
intros t t' l def; rewrite (flatten_valid_Bt l def t); rewrite (flatten_valid_Bt l def t').
trivial.
Qed.
End assoc.

Ltac term_list l v :=
  match v with
  | (?X1 + ?X2) => let l1 := term_list l X2 in term_list l1 X1
  | (?X1 * ?X2) => let l1 := term_list l X2 in term_list l1 X1
  | ?X1 => constr:(cons X1 l)
  end.

Ltac compute_rank l n v :=
  match l with
  | (cons ?X1 ?X2) =>
    let t1 := constr:X2 in
    match constr:(X1 = v) with
    | (?X1 = ?X1) => n
    | _ => compute_rank t1 (S n) v
    end
  end.

Ltac model_aux l v :=
  match v with
  | (?X1 + ?X2) =>
    let r1 := model_aux l X1 with r2 := model_aux l X2 in
      constr:(pnode r1 r2)
  | (?X1 * ?X2) =>
    let r1 := model_aux l X1 with r2 := model_aux l X2 in
      constr:(tnode r1 r2)
  | ?X1 => let n := compute_rank l 0%nat X1 in constr:(leaf n)
  | _ => constr:(leaf 0%nat)
  end.
 
Ltac model_Bt def v :=
  let l := term_list (nil (A:=Bt)) v in
  let t := model_aux l v in
  constr:(bin_Bt l def t).

Ltac assoc_eq:=
  match goal with
  | [|- (?X1 = ?X2) ] =>
  let term1 := model_Bt X1 X1
  with term2 := model_Bt X1 X2 in
  (change (term1 = term2);
  apply flatten_valid_Bt_2; auto)
  end.

(*
Variables a b c d e f g h i j k l: Bt.

Goal ((a+(b+c))*(d+(e+f)))*(g+(h+i)) = ((a+b)+c) * (((d+e)+f)*((g+h)+i)).
assoc_eq.
Qed.
(*Checks*)

Goal ((a*(b*c))+(d*(e*f)))+(g*(h*i)) = ((a*b)*c) + (((d*e)*f)+((g*h)*i)).
assoc_eq.
Qed.
(*Checks*)

Goal (a+(b*(c*d)))+(e*f*(g+h))+(i*j*(k+l)) = (a + ((b*c)*d)) + ((e*(f*(g+h))) + (i*(j*(k+l)))).
assoc_eq.
(*Checks*)

Goal (a+b+c+d)*(e+f+g+h)*(i+j+k+l) = ((a+b)+(c+d))*(((e+f)+(g+h))*((i+j)+(k+l))).
assoc_eq.
(*Checks*)

Goal (a*b*c*d)+(e*f*g*h)+(i*j*k*l) = ((a*b)*(c*d))+(((e*f)*(g*h))+((i*j)*(k*l))).
assoc_eq.
(*Checks*)
*)

Section commut.

Fixpoint nat_eq_bool (m n:nat){struct n} : bool :=
  match m, n with
  | 0, 0 => true
  | S m, 0 => false
  | 0, S n => false
  | S m, S n => nat_eq_bool m n
  end.

Fixpoint nat_lt_bool (m n:nat){struct n} : bool :=
  match m, n with
  | _, 0 => false
  | 0, S n => true
  | S m, S n => nat_lt_bool m n
  end.

Fixpoint bin_eq_bool (t t':bin){struct t'} : bool :=
  match t, t' with
  | (leaf m), (leaf n) => nat_eq_bool m n
  | (pnode p1 p2), (pnode p'1 p'2) => andb (bin_eq_bool p1 p'1) (bin_eq_bool p2 p'2)
  | (tnode t1 t2), (tnode t'1 t'2) => andb (bin_eq_bool t1 t'1) (bin_eq_bool t2 t'2)
  | _, _ => false
  end.

(*The order is arbitrary.  Leaves are less than pnodes are less than tnodes, all other cases
 depend on the constructors. *)

Fixpoint bin_lt_bool (t t':bin){struct t'} : bool :=
  match t, t' with
  | (leaf m), (leaf n) => nat_lt_bool m n
  | (leaf m), _=> true
  | (pnode p1 p2), (pnode p'1 p'2) => if (bin_eq_bool p1 p'1) then (bin_lt_bool p2 p'2)
                                        else (bin_lt_bool p1 p'1)
  | (pnode p1 p2), (leaf m) => false
  | (pnode _ _), (tnode _ _) => true
  | (tnode t1 t2), (tnode t'1 t'2) => if (bin_eq_bool t1 t'1) then (bin_lt_bool t2 t'2)
                                        else (bin_lt_bool t1 t'1)
  | (tnode t1 t2), _ => false
  end.

Fixpoint insert_bin_p (t0:bin)(t:bin){struct t} : bin :=
  match t with
  | leaf m => match (bin_lt_bool t0 t) with
              | true => pnode t0 t
              | false => pnode t t0
              end
  | tnode _ _ => match (bin_lt_bool t0 t) with
                 | true => pnode t0 t
                 | false => pnode t t0
                 end
  | pnode ((leaf m) as l) t' =>
    match (bin_lt_bool t0 l) with
    | true => pnode t0 t
    | false => pnode l (insert_bin_p t0 t')
    end
  | pnode ((tnode _ _) as x) t' =>
    match (bin_lt_bool t0 x) with
    | true => pnode t0 t
    | false => pnode x (insert_bin_p t0 t')
    end
  | t => pnode t0 t
  end.


Fixpoint insert_bin_t (t0:bin)(t:bin){struct t} : bin :=
  match t with
  | leaf m => match (bin_lt_bool t0 t) with
              | true => tnode t0 t
              | false => tnode t t0
              end
  | pnode _ _ => match (bin_lt_bool t0 t) with
                 | true => tnode t0 t
                 | false => tnode t t0
                 end
  | tnode ((leaf m) as l) t' =>
    match (bin_lt_bool t0 l) with
    | true => tnode t0 t
    | false => tnode l (insert_bin_t t0 t')
    end
  | tnode ((pnode _ _) as x) t' =>
    match (bin_lt_bool t0 x) with
    | true => tnode t0 t
    | false => tnode x (insert_bin_t t0 t')
    end
  | t => tnode t0 t
  end.

Fixpoint sort_bin (t:bin) : bin :=
  match t with
  | tnode ((leaf n) as l) t' => insert_bin_t l (sort_bin t')
  | tnode (pnode _ _ as p) t' => insert_bin_t (sort_bin p) (sort_bin t')
  | pnode ((leaf n) as l) p' => insert_bin_p l (sort_bin p')
  | pnode (tnode _ _ as t') p' => insert_bin_p (sort_bin t') (sort_bin p')
  | t => t
end.

Theorem bin_Bt_ref_p : forall (l:list Bt)(def:Bt)(t1 t2:bin), 
  bin_Bt l def (pnode t1 t2) = (bin_Bt l def t1) + (bin_Bt l def t2).
intros l def t1 t2.  unfold bin_Bt. fold bin_Bt. reflexivity.
Qed.

Theorem bin_Bt_ref_t : forall (l:list Bt)(def:Bt)(t1 t2:bin),
  bin_Bt l def (tnode t1 t2) = (bin_Bt l def t1) * (bin_Bt l def t2).
intros l def t1 t2. unfold bin_Bt. fold bin_Bt. reflexivity.
Qed.

Theorem insert_eq_p :
  forall (l:list Bt)(def:Bt)(t0 t:bin),
  bin_Bt l def (insert_bin_p t0 t) = (bin_Bt l def t0) + (bin_Bt l def t).
intros l def t0 t.  induction t; unfold insert_bin_p. case t1.
intros b b0. apply bin_Bt_ref_p.
(*next subgoal*)
fold insert_bin_p. 
intros b b0. case (bin_lt_bool t0 (tnode b b0)).
apply bin_Bt_ref_p.
(*next subgoal*)
rewrite bin_Bt_ref_p. rewrite IHt2.
unfold bin_Bt. fold bin_Bt. 
rewrite (assoc_sum B) with (n:= bin_Bt l def t0).
rewrite (comm_sum B) with (n := bin_Bt l def t0) 
  (m := bin_Bt l def b * bin_Bt l def b0).
apply (assoc_sum B).
(*next subgoal*)
fold insert_bin_p.
intros n.
case (bin_lt_bool t0 (leaf n)).
rewrite bin_Bt_ref_p. reflexivity.
rewrite bin_Bt_ref_p. rewrite IHt2.
rewrite bin_Bt_ref_p.
rewrite (assoc_sum B) with (n := bin_Bt l def t0).
rewrite (comm_sum B) with (n := bin_Bt l def t0)
  (m := bin_Bt l def (leaf n)).
rewrite (assoc_sum B). reflexivity.
case (bin_lt_bool t0 (tnode t1 t2)).
rewrite bin_Bt_ref_p. reflexivity.
rewrite bin_Bt_ref_p. apply (comm_sum B). 
case (bin_lt_bool t0 (leaf n)).
rewrite bin_Bt_ref_p. reflexivity.
rewrite bin_Bt_ref_p. apply (comm_sum B).
Qed.

Theorem insert_eq_t :
  forall (l:list Bt)(def:Bt)(t0 t:bin),
  bin_Bt l def (insert_bin_t t0 t) = (bin_Bt l def t0) * (bin_Bt l def t).
intros l def t0 t.  induction t; unfold insert_bin_t. 
case (bin_lt_bool t0 (pnode t1 t2)).
rewrite bin_Bt_ref_t. reflexivity.
rewrite bin_Bt_ref_t. apply (comm_prod B).
fold insert_bin_t. destruct t1.
case (bin_lt_bool t0 (pnode t1_1 t1_2)).
apply bin_Bt_ref_t.
rewrite bin_Bt_ref_t. 
rewrite IHt2.
unfold bin_Bt. fold bin_Bt.
rewrite (assoc_prod B) with (n := bin_Bt l def t0).
rewrite (comm_prod B) with (n := bin_Bt l def t0)
  (m := (bin_Bt l def t1_1 + bin_Bt l def t1_2)).
apply (assoc_prod B).
unfold bin_Bt.  fold bin_Bt. reflexivity.
case (bin_lt_bool t0 (leaf n)).
apply bin_Bt_ref_t. 
rewrite bin_Bt_ref_t.
rewrite IHt2.
rewrite bin_Bt_ref_t.
rewrite (assoc_prod B) with (n := bin_Bt l def t0).
rewrite (comm_prod B) with (n := bin_Bt l def t0)
  (m := bin_Bt l def (leaf n)).
apply (assoc_prod B).
case (bin_lt_bool t0 (leaf n)).
apply bin_Bt_ref_t.
rewrite bin_Bt_ref_t. apply (comm_prod B).
Qed.

Theorem sort_eq : forall (l: list Bt)(def:Bt)(t:bin),
  bin_Bt l def (sort_bin t) = bin_Bt l def t.
intros l def. induction t.
unfold sort_bin. destruct t1. reflexivity.
fold sort_bin. unfold sort_bin in IHt1. fold sort_bin in IHt1.
rewrite insert_eq_p. rewrite IHt1.  rewrite IHt2. rewrite <- bin_Bt_ref_p.
reflexivity.
fold sort_bin. rewrite insert_eq_p. rewrite IHt2.
rewrite bin_Bt_ref_p. reflexivity.
unfold sort_bin.  destruct t1. fold sort_bin.
unfold sort_bin in IHt1. fold sort_bin in IHt1. rewrite insert_eq_t.
rewrite IHt1. rewrite IHt2. rewrite <- bin_Bt_ref_t.  reflexivity.
reflexivity.
fold sort_bin. rewrite insert_eq_t. rewrite bin_Bt_ref_t. rewrite IHt2.
reflexivity.
unfold sort_bin. reflexivity.
Qed.

Theorem sort_eq_2 :
  forall (l:list Bt)(def:Bt)(t1 t2:bin),
  bin_Bt l def (sort_bin t1) = bin_Bt l def (sort_bin t2) ->
  bin_Bt l def t1 = bin_Bt l def t2.
intros l def t1 t2.
rewrite sort_eq. rewrite sort_eq. trivial.
Qed.

End commut.


Ltac comm_eq :=
  match goal with
  | [ |- (?X1 = ?X2) ] =>
  let l := term_list (nil (A := Bt)) X1 in
  let term1 := model_aux l X1
  with term2 := model_aux l X2 in
  (change (bin_Bt l X1 term1 = bin_Bt l X1 term2);
    apply flatten_valid_Bt_2; 
    apply sort_eq_2;
    auto)
  end.

(*
(*All check!*)
Variables a b c d e f g h i j k l : Bt.
Goal (a*b)+(c*d) = (b*a) +(d*c).
comm_eq.
Goal (a*b)+(d*c)+(e*f*g) = (g*f*e) + (c*d) + (b*a).
comm_eq.
Goal (a*b*c+(d*e*f+(h*i*j)+k) = k + c*b*a + (f*d*e) + (j*i*h)).
comm_eq.
Goal a*(b+c)*d = (c+b)*d*a.
comm_eq.
Goal a+(b*(c+(d*e*f))*g) = ((f*e*d) + c)*b*g + a.
comm_eq.
*)

(*Theorems *)

Lemma dist_sum_r : forall (n m p: Bt), (n + m) * p = n*p + m*p.
intros n m p.  assert (h1: (n + m) * p = p * (n + m)).  apply (comm_prod B).
assert (h2: p * (n + m) = p*n + p*m).  apply (dist_sum B).
assert (h3: p*n + p*m = n*p + m*p). comm_eq. congruence.
Qed.
Hint Resolve dist_sum_r.

Lemma dist_prod_r : forall (n m p: Bt), (n*m) + p = (n+p) * (m + p).
intros n m p.  assert (h1: (n*m) + p = p + (n * m)).  apply (comm_sum B).
assert (h2: p + (n * m) = (p + n) * (p + m)).  apply (dist_prod B). 
assert (h3: (p+n)*(p+m) = (n+p) * (m+p)). comm_eq. congruence.
Qed.
Hint Resolve dist_prod_r.

Lemma idem_sum : forall b : Bt, b + b = b.
intros b.  assert (h1: (b+b = b + b * (b + b))).
assert (h2: b * (b + b) = b). rewrite (abs_prod B). reflexivity. 
rewrite h2. reflexivity. rewrite (abs_sum B) in h1. assumption. 
Qed.
Hint Resolve idem_sum.

Lemma idem_prod: forall (b: Bt), b * b = b.
intros. assert (s: (b*b = b * (b + (b * b)))).
assert (s: b + (b * b) = b). apply (abs_sum B). rewrite s.
reflexivity. rewrite s. rewrite (abs_prod B). reflexivity. 
Qed.
Hint Resolve idem_prod.

Lemma eq_ord: forall (x y: Bt),
  x + y = y <-> x * y = x.
intros. assert (h: x + y = y -> x * y = x).
intros h1. assert (h2: x * y = x * (x + y)). rewrite h1.
reflexivity. assert (h3: x * (x + y) = x).  apply abs_prod.
congruence.
assert (h4: x * y = x -> x + y = y).
intro h5.
assert (h4: x + x * y = x). rewrite h5. apply idem_sum.
assert (h6: y + y * x = y).  apply abs_sum.
assert (h7: y + y * x = x*y + y).  comm_eq.
congruence. tauto.
Qed.
Hint Resolve eq_ord.

Definition le (x y:Bt): Prop :=
x + y = y.
Hint Unfold le.


Lemma refl_le : forall (x: Bt), le x x.
intros.  assert (x + x = x).  rewrite idem_sum. reflexivity.
auto.
Qed.
Hint Resolve refl_le.

Lemma anti_sym_le: forall (x y: Bt), 
  le x y -> le y x -> x = y.
intros x y h1 h2.  assert (h3: x + y = y). intuition.
assert (h4:y + x = x). intuition.  assert (h5: x + y = y + x).
rewrite comm_sum. reflexivity. rewrite <- h3.
rewrite h5. rewrite h4. reflexivity.
Qed.
Hint Resolve anti_sym_le.

Lemma trans_le: forall (x y z : Bt),
  le x y -> le y z -> le x z.
intros x y z h1 h2.  assert (h3: x + z = x + (y + z)). rewrite h2. 
reflexivity.  assert (h4: x + (y + z) = x + y + z). assoc_eq.
assert (h5: x + y + z = y + z). rewrite h1. reflexivity.
assert (h6: y + z = z). rewrite h2. reflexivity. assert (h7: x + z = z).
congruence. auto.
Qed.
Hint Resolve trans_le.

Lemma times_le : forall (x y : Bt), le (x * y) x.
intros x y.
red.
rewrite eq_ord.
rewrite <- assoc_prod.
rewrite (comm_prod _ y x).
rewrite assoc_prod.
rewrite idem_prod.
reflexivity.
Qed.

Lemma le_plus : forall (x y : Bt), le x (x + y).
intros x y.
red.
rewrite assoc_sum.
rewrite idem_sum.
reflexivity.
Qed.

(* b is a lower bound of S in B*)
Definition lb (S: Ensemble Bt) (b: Bt) : Prop :=
  forall (s:Bt), Ensembles.In Bt S s -> le b s.
Hint Unfold lb.

(* b is an upper bound of S in B*)
Definition ub (S: Ensemble Bt) (b: Bt) : Prop :=   
  forall (s:Bt), Ensembles.In Bt S s -> le s b.
Hint Unfold ub.

(* b is the infimum (greatest lower bound) of S in B *)
Definition inf (S: Ensemble Bt) (b: Bt) : Prop :=
  lb S b /\ forall b':Bt, lb S b' -> le b' b.
Hint Unfold inf.

(* b is the supremum (least upper bound) of S in B*)
Definition sup (S: Ensemble Bt) (b: Bt) : Prop :=
  ub S b /\ forall b':Bt, ub S b' -> le b b'.
Hint Unfold sup.

Lemma inf_singleton : forall (x:Bt), inf (Singleton _ x) x.
intro x.
red. split.
(* lower bound *)
red.
intros y h1.
inversion h1.
subst.
auto.
(* greatest lower bound *)
intros b' h1.
red in h1.
pose proof (Singleton_intro _ _ _ (eq_refl x)) as h2.
apply h1. assumption.
Qed.

Lemma sup_singleton : forall (x:Bt), sup (Singleton _ x) x.
intro x.
red. split.
(* lower bound *)
red.
intros y h1.
inversion h1.
subst.
auto.
(* greatest lower bound *)
intros b' h1.
red in h1.
pose proof (Singleton_intro _ _ _ (eq_refl x)) as h2.
apply h1. assumption.
Qed.

(* infimum/supremem unique to do.*)
(* B is a lattice under partial order le (sum portion) *)
Lemma lat_sum : forall (x y: Bt), 
  sup (Couple Bt x y) (x + y).
intros x y.  assert (h1: x + (x + y) = (x + x) + y).
assoc_eq. 
assert (h2: x + x + y = x + y). rewrite idem_sum. reflexivity.
assert (h3: le x (x + y)). unfold le. rewrite h1. rewrite h2. reflexivity.
assert (h4: le y (y + x)). unfold le.  rewrite (assoc_sum B). rewrite idem_sum.
reflexivity. assert (h5: ub (Couple Bt x y) (x+y)). unfold ub. intros s.  
intros h7.  assert (h8: s = x \/ s = y). apply Couple_inv.
apply h7.  destruct h8 as [h8l | h8r]. rewrite h8l. apply h3.  rewrite h8r.
rewrite (comm_sum B). apply h4.  unfold sup.  split.  assumption. intros b' h11.
unfold le.  assert (h15: y + b' = b').  apply h11.  apply (Couple_r Bt).  
assert (h16: x + y + b' = x + (y + b')).  assoc_eq. 
rewrite h16. rewrite h15.  apply h11. apply (Couple_l Bt).
Qed.
Hint Resolve lat_sum.

(* B is a lattice under partial order le (product portion) *)
Lemma lat_prod : forall (x y: Bt),
  inf (Couple Bt x y) (x * y).
intros x y.  assert (h0: lb (Couple Bt x y) (x * y)).  
unfold lb.  intros s h2.  assert (h3: s = x \/ s = y).  apply Couple_inv.
assumption.  destruct h3 as [h3l | h3r]. rewrite h3l.  unfold le.
apply eq_ord.  assert (h4: y * x = x * y). apply (comm_prod B).
assert (h5: x * y * x = x * (y * x)). assoc_eq.   
rewrite h5.  rewrite h4.  assert (h6: x * (x * y) = x * x * y). 
rewrite (assoc_prod B). reflexivity.  rewrite h6.  rewrite idem_prod.
reflexivity.  rewrite h3r.  unfold le.  apply eq_ord.
assert (h7: x * y * y = x * (y * y)).  rewrite (assoc_prod B). reflexivity.
rewrite h7.  rewrite idem_prod. reflexivity. unfold inf.  split.
assumption.  intros z h10. unfold le.  apply eq_ord.  
assert (h11: z * (x * y) = z * x * y).  rewrite <- (assoc_prod B).
reflexivity. rewrite h11.  assert (h12: le z x). apply h10.
apply (Couple_l Bt). assert (h13: z * x = z).
apply eq_ord. apply h12. rewrite h13.  assert (h14: le z y).
apply h10.  apply (Couple_r Bt). apply eq_ord. 
apply h14.
Qed.
Hint Resolve lat_prod.

(* infimum unique *)
Lemma inf_unq : forall (S:Ensemble Bt) (x y: Bt), 
  inf S x -> inf S y -> x = y.
intros S x y h2 h3.  assert (h4: le x y).  
assert (h5: lb S y). apply h3. apply h3. apply h2.
assert (h6: le y x). assert (h7: lb S x).  apply h2. apply h2.
apply h3.  apply anti_sym_le. assumption.
assumption. 
Qed.
Hint Resolve inf_unq.

(* supremum unique *)
Lemma sup_unq : forall (S:Ensemble Bt) (x y: Bt), 
  sup S x -> sup S y -> x = y.
intros S x y h2 h3.  assert (h4: le x y). assert (h5: ub S x).
apply h2. apply h2.  apply h3. assert (h6: le y x).  assert (h7: ub S y).
apply h3. apply h3.  apply h2.  apply anti_sym_le.  assumption.
assumption.
Qed.
Hint Resolve sup_unq.

(*0 <= x*)
Lemma zero_min : forall (x : Bt), le 0 x.
intros x.  assert (h1: inf (Couple Bt x (-x)) 0).
assert (h2: inf (Couple Bt x (-x)) (x * - x)).
apply lat_prod.  assert (h3: x * - x = 0).  apply (comp_prod B).
rewrite <- h3. assumption.  apply h1. apply (Couple_l Bt).
Qed.
Hint Resolve zero_min.

Lemma le_x_0 : forall (x:Bt), le x 0 -> x = 0.
intros x h1.
pose proof (zero_min x) as h2.
apply anti_sym_le; assumption.
Qed.

(* x <= 1 *)
Lemma one_max : forall (x : Bt), le x 1.
intros x.  assert (h1: sup (Couple Bt x (-x)) 1).
assert (h2: sup (Couple Bt x (-x)) (x + (- x))).
apply lat_sum.  assert (h3: x + (-x) = 1).  apply (comp_sum B).
rewrite <- h3. assumption. apply h1.  apply (Couple_l Bt).
Qed.
Hint Resolve one_max.

Lemma le_1_x : forall (x:Bt), le 1 x -> x = 1.
intros x h1.
pose proof (one_max x) as h2.
apply anti_sym_le; assumption.
Qed.

(* x + 0 = x *)
Lemma zero_sum : forall (x : Bt), x + 0 = x.
intros x.  assert (h1: ub (Couple Bt x 0) x).
unfold ub. intros s h2.  assert (h3: s = x \/ s = 0). apply Couple_inv. 
assumption.  destruct h3 as [h3l | h3r].  rewrite h3l. apply idem_sum.
rewrite h3r.  apply zero_min.  assert (h4 : sup (Couple Bt x 0) (x + 0)).
apply lat_sum.  assert (h5: le (x + 0) x). apply h4. assumption. 
assert (h6: le x (x + 0)). apply lat_sum. apply (Couple_l Bt). 
apply anti_sym_le. assumption.  assumption.
Qed.
Hint Resolve zero_sum.

(*  x * 0 = 0 *)
Lemma zero_prod : forall (x : Bt), x * 0 = 0.
intros x.  assert (h1: lb (Couple Bt x 0) 0).
unfold lb.  intros s h2. assert (h3: s = x \/ s = 0). apply Couple_inv.
assumption. destruct h3 as [h3l | h3r].  rewrite h3l. apply zero_min.
rewrite h3r.  unfold le. apply zero_sum.  
assert (h4: inf (Couple Bt x 0) (x * 0)). apply lat_prod.
assert (h5: le 0 (x * 0)).  apply h4. assumption.
assert (h6: le (x * 0) 0). apply lat_prod. apply (Couple_r Bt).
apply anti_sym_le.  assumption. assumption.
Qed.
Hint Resolve zero_prod.

(* x + 1 = 1 *)
Lemma one_sum : forall (x : Bt), x + 1 = 1.
intros x.  assert (h1: ub (Couple Bt x 1) 1).
unfold ub. intros s h2.  assert (h3 : s = x \/ s = 1). apply Couple_inv.
assumption.  destruct h3 as [h3l | h3r].  rewrite h3l.  apply one_max.
rewrite h3r.  unfold le.  apply idem_sum.  
assert (h4: sup (Couple Bt x 1) (x + 1)).  apply lat_sum.
assert (h5: le (x + 1) 1). apply h4. assumption.
assert (h6: le 1 (x+1)).  apply h4. apply (Couple_r Bt).
apply anti_sym_le. assumption. assumption.
Qed.
Hint Resolve one_sum.

(* x * 1 = x *)
Lemma one_prod : forall x : Bt, x * 1 = x.
intros x.  assert (h1: lb (Couple Bt x 1) x).
unfold lb. intros s h2.  assert (h3 : s = x \/ s = 1).  apply Couple_inv.
assumption.  destruct h3 as [h3l | h3r].  rewrite h3l.  apply idem_sum.
rewrite h3r.  apply one_max.  assert (h4: inf (Couple Bt x 1) (x * 1)).
apply lat_prod.  assert (h5: le x (x * 1)). apply h4. assumption.
assert (h6: le (x * 1) x).  apply lat_prod. apply (Couple_l Bt).
apply anti_sym_le.  assumption.  assumption.
Qed.
Hint Resolve one_prod.

Lemma inf_empty : inf (Empty_set Bt) 1.
red. split.
red.
intros.
contradiction.
auto.
Qed.

Lemma sup_empty : sup (Empty_set Bt) 0.
red. split.
red. intros. 
contradiction.
auto.
Qed.

(*The goal of this section is to develop a tactic that reduces equalities that contain
  any 0 or 1 in a disjunction/conjunction or any adjacent combination of x/-x, so that
  0s and 1s and inverses get eliminated on both sides of the equation *)
Section const_eq.
Inductive bin2 : Set := pnode2 : bin2 -> bin2 -> bin2 |
                        tnode2 : bin2 -> bin2 -> bin2 |
                        leaf2 : nat-> bin2 |
                        zero : bin2 |
                        one  : bin2 |
                        popps : nat -> bool -> bin2 |
                        topps : nat -> bool -> bin2.

Fixpoint bin2_Bt (l: list Bt)(def:Bt)(t:bin2){struct t}: Bt :=
  match t with
  | pnode2 t1 t2 => (bin2_Bt l def t1) + (bin2_Bt l def t2)
  | tnode2 t1 t2 => (bin2_Bt l def t1) * (bin2_Bt l def t2)
  | leaf2 n => nth n l def
  | zero => 0
  | one  => 1
  | popps n b => let x := (nth n l def) in 
                    if b then (x + -x) else (-x + x)
  | topps n b => let x := (nth n l def) in 
                    if b then (x * -x) else (-x * x)
  end.

Fixpoint elim_bin2_aux (t:bin2) : bin2 :=
  match t with 
    | pnode2 zero t' => elim_bin2_aux t'
    | pnode2 t' zero => elim_bin2_aux t'
    | pnode2 one t' => one
    | pnode2 t' one => one
    | tnode2 zero t' => zero
    | tnode2 t' zero => zero
    | tnode2 one t' => elim_bin2_aux t'
    | tnode2 t' one => elim_bin2_aux t'
    | pnode2 t1 t2 => pnode2 (elim_bin2_aux t1) (elim_bin2_aux t2)
    | tnode2 t1 t2 => tnode2 (elim_bin2_aux t1) (elim_bin2_aux t2)
    | topps _ _ => zero
    | popps _ _ => one
    | _ => t
  end.

Fixpoint elim_bin2 (t: bin2) : bin2 :=
  match t with
  | pnode2 t1 t2 => elim_bin2_aux (pnode2 (elim_bin2 t1) (elim_bin2 t2))
  | tnode2 t1 t2 => elim_bin2_aux (tnode2 (elim_bin2 t1) (elim_bin2 t2))
  | topps _ _ => zero
  | popps _ _ => one
  | _ => t
  end.

Theorem bin_Bt2_ref_p : forall (l:list Bt)(def:Bt)(t1 t2:bin2),
  bin2_Bt l def (pnode2 t1 t2) = (bin2_Bt l def t1) + (bin2_Bt l def t2).
intros l def t1 t2.  unfold bin2_Bt. fold bin2_Bt. reflexivity.
Qed.

Theorem bin_Bt2_ref_t : forall (l:list Bt)(def:Bt)(t1 t2:bin2),
  bin2_Bt l def (tnode2 t1 t2) = (bin2_Bt l def t1) * (bin2_Bt l def t2).
intros l def t1 t2. unfold bin2_Bt. fold bin2_Bt. reflexivity.
Qed.

Theorem elim_bin2_aux_p : forall (l:list Bt) (def : Bt) (t1 t2:bin2), 
  bin2_Bt l def (elim_bin2_aux (pnode2 t1 t2)) = 
  bin2_Bt l def (pnode2 (elim_bin2_aux t1) (elim_bin2_aux t2)).
intros l def t1 t2.  unfold elim_bin2_aux. fold elim_bin2_aux.
destruct t1. destruct t2. reflexivity. 
(* next subgoal *) reflexivity. 
(* next subgoal *) reflexivity.
(* next subgoal *)
unfold elim_bin2_aux at 3.  rewrite bin_Bt2_ref_p. unfold bin2_Bt at 3.
rewrite zero_sum. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 2.  rewrite bin_Bt2_ref_p. unfold bin2_Bt at 3.
unfold bin2_Bt at 1. rewrite one_sum. reflexivity.
(*next subgoal*)
reflexivity.
(*next subgoal*)
reflexivity.
(*next subgoal*)
destruct t2.  reflexivity.
(*next subgoal*)
reflexivity.
(*next subgoal*)
reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 3.  rewrite bin_Bt2_ref_p. unfold bin2_Bt at 3.
rewrite zero_sum. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 2. unfold bin2_Bt. fold bin2_Bt. 
rewrite one_sum. reflexivity.
(*next subgoal*)
reflexivity.
(*next subgoal*)
reflexivity.
(*next subgoal*)
destruct t2. reflexivity.
(*next subgoal *)
reflexivity.
(*next subgoal *)
reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 3. unfold bin2_Bt. fold bin2_Bt.
rewrite zero_sum. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 2. unfold bin2_Bt. fold bin2_Bt.
rewrite one_sum. reflexivity.
(*next subgoal*)
reflexivity.
(*next subgoal*)
reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 2. unfold bin2_Bt. fold bin2_Bt.
rewrite (comm_sum B). rewrite zero_sum. reflexivity.
(*next subgoal*)
destruct t2;
unfold elim_bin2_aux at 1; unfold bin2_Bt; fold bin2_Bt;
rewrite (comm_sum B); rewrite one_sum; reflexivity.
(*next subgoal*)
destruct t2; try reflexivity.
unfold elim_bin2_aux at 3. unfold bin2_Bt. fold bin2_Bt. 
rewrite zero_sum. reflexivity.
unfold elim_bin2_aux at 2.  unfold bin2_Bt. fold bin2_Bt.
rewrite one_sum. reflexivity.
destruct t2; try reflexivity.
unfold elim_bin2_aux. unfold bin2_Bt. fold bin2_Bt. rewrite zero_sum.
reflexivity.
unfold elim_bin2_aux at 2. unfold bin2_Bt. fold bin2_Bt.
rewrite one_sum. reflexivity.
Qed.

Theorem elim_bin2_aux_t : forall (l:list Bt) (def : Bt) (t1 t2:bin2), 
  bin2_Bt l def (elim_bin2_aux (tnode2 t1 t2)) = 
  bin2_Bt l def (tnode2 (elim_bin2_aux t1) (elim_bin2_aux t2)).
intros l def t1 t2.  unfold elim_bin2_aux. fold elim_bin2_aux.
destruct t1. destruct t2; try reflexivity.
unfold elim_bin2_aux at 2. unfold bin2_Bt. fold bin2_Bt.
rewrite zero_prod. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 3. unfold bin2_Bt. fold bin2_Bt.
rewrite one_prod. reflexivity.
(*next subgoal*)
destruct t2; try reflexivity.
unfold elim_bin2_aux at 2. unfold bin2_Bt. fold bin2_Bt.
rewrite zero_prod. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 3. unfold bin2_Bt. fold bin2_Bt.
rewrite one_prod. reflexivity.
(*next subgoal*)
destruct t2; try reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 2. unfold bin2_Bt. fold bin2_Bt.
rewrite zero_prod. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 3. unfold bin2_Bt. fold bin2_Bt.
rewrite one_prod. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 1. unfold bin2_Bt. fold bin2_Bt.
rewrite (comm_prod B). rewrite zero_prod. reflexivity.
unfold elim_bin2_aux at 5. unfold bin2_Bt. fold bin2_Bt.
rewrite (comm_prod B). rewrite (one_prod). 
destruct t2; try reflexivity.
(*next subgoal*)
destruct t2; try reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 2.  unfold bin2_Bt. fold bin2_Bt.
rewrite zero_prod. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 3.  unfold bin2_Bt. fold bin2_Bt.
rewrite one_prod. reflexivity.
(*next subgoal*)
destruct t2; try reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 2. unfold bin2_Bt. fold bin2_Bt.
rewrite zero_prod. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux at 3. unfold bin2_Bt. fold bin2_Bt.
rewrite one_prod. reflexivity.
Qed.


Theorem elim_bin2_aux_valid : forall (l:list Bt)(def:Bt)(t:bin2),
   bin2_Bt l def t = bin2_Bt l def (elim_bin2_aux t).
intros l def t.  induction t; unfold bin2_Bt; fold bin2_Bt.
rewrite IHt1.  rewrite IHt2. rewrite <- bin_Bt2_ref_p. 
rewrite elim_bin2_aux_p. reflexivity.
(*next subgoal*)
rewrite IHt1.  rewrite IHt2. rewrite <- bin_Bt2_ref_t.
rewrite elim_bin2_aux_t. reflexivity.
unfold elim_bin2_aux. unfold bin2_Bt. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux. unfold bin2_Bt. reflexivity.
(*next subgoal*)
unfold elim_bin2_aux. unfold bin2_Bt. reflexivity.
(*next subgoal*)
case b.  unfold elim_bin2_aux. unfold bin2_Bt. apply (comp_sum B). 
(*next subgoal*)
unfold elim_bin2_aux. unfold bin2_Bt. rewrite (comm_sum B).
apply (comp_sum B).
(*next subgoal*)
case b.  unfold elim_bin2_aux. unfold bin2_Bt. apply (comp_prod B).
(*next subgoal*)
unfold elim_bin2_aux. unfold bin2_Bt. rewrite (comm_prod B). 
apply (comp_prod B). 
Qed.

Theorem  elim_bin2_valid :
  forall (l:list Bt)(def:Bt)(t:bin2),
  bin2_Bt l def t = bin2_Bt l def (elim_bin2 t).
intros l def t.  
induction t; unfold elim_bin2; fold elim_bin2; 
try rewrite <- elim_bin2_aux_valid; unfold bin2_Bt; fold bin2_Bt;
try (rewrite IHt1; rewrite IHt2; reflexivity) || reflexivity.
case b.  apply (comp_sum B).  
(*next subgoal*) 
rewrite (comm_sum B). apply (comp_sum B).
case b. apply (comp_prod B). 
(*next subgoal*)
rewrite (comm_prod B). apply (comp_prod B).
Qed.

Theorem elim_bin2_valid_2 :
  forall (t t':bin2)(l:list Bt)(def:Bt),
    bin2_Bt l def (elim_bin2 t) = bin2_Bt l def (elim_bin2 t') ->
    bin2_Bt l def t = bin2_Bt l def t'.
intros t t' l def.  
rewrite <- elim_bin2_valid.  rewrite <- elim_bin2_valid.
tauto.
Qed.

End const_eq.


(*Tactics to build the const_eq tactic*)
Ltac term_list2 l v :=
  match v with
  | (?X1 + (-?X1)) => constr:(cons X1 l)
  | ((-?X1) + ?X1) => constr:(cons X1 l)
  | (?X1 * (-?X1)) => constr:(cons X1 l)
  | ((-?X1) * ?X1) => constr:(cons X1 l)
  | (?X1 + ?X2) => let l1 := term_list2 l X2 in term_list2 l1 X1
  | (?X1 * ?X2) => let l1 := term_list2 l X2 in term_list2 l1 X1
  | ?X1 => constr:(cons X1 l)
  end.

Ltac model_aux2 l v :=
  match v with
  | (?X1 + (-?X1)) => let  n := compute_rank l 0%nat X1 in 
                    constr:(popps n true)
  | ((-?X1) + ?X1) => let  n := compute_rank l 0%nat X1 in
                    constr:(popps n false)
  | (?X1 * (-?X1)) => let  n := compute_rank l 0%nat X1 in
                    constr:(topps n true)
  | ((-?X1) * ?X1) => let  n := compute_rank l 0%nat X1 in 
                    constr:(topps n false)
  | (?X1 + ?X2) =>
    let r1 := model_aux2 l X1 with r2 := model_aux2 l X2 in
      constr:(pnode2 r1 r2)
  | (?X1 * ?X2) =>
    let r1 := model_aux2 l X1 with r2 := model_aux2 l X2 in
      constr:(tnode2 r1 r2)
  | 0 => constr:(zero)
  | 1 => constr:(one)
  | ?X1 => let n := compute_rank l 0%nat X1 in constr:(leaf2 n)
  | _ => constr:(leaf2 0%nat)
  end.

Ltac const_eq :=
  match goal with
  | [ |- (?X1 = ?X2) ] =>
  
  let l1 := term_list2 (nil (A := Bt)) X1 in
  let l2 := term_list2 l1 X2 in
  
  let term1 := model_aux2 l2 X1
  with term2 := model_aux2 l2 X2 in
  (change (bin2_Bt l2 X1 term1 = bin2_Bt l2 X1 term2);
   apply elim_bin2_valid_2;
   auto)
  end.

(*
(*All tests check *)
Variables a b c d e f g h : Bt.

Goal ((a + 0) = a).
const_eq.
Qed.
Goal ((a + (b*0)) = a).
const_eq.
Qed.
Goal ((a+b+(c*1)*(d+0)) = a + b +(c*d)).
const_eq.
Qed.
Goal (0*(a+b) + c*1) = c.
const_eq.
Qed.
Goal (a+(b*-b)) = a.
const_eq.
Qed.
Goal (a+(c+d)*(-(c+d))) + (b*(c+d*((e+f) + (-(e+f))))) = a + b*(c+d).
const_eq.
Qed.
*)

(* More Theorems *)
(* complement unique *)

Lemma comp_unq : forall (x y : Bt), -x = -y -> x = y.
intros x y h1.  
assert (h2: x = x*y).
  assert (h3: x*y = x*y + x * -x). const_eq.
  assert (h4: x*y + x*-x = x * (y + -x)).  
    rewrite <- (dist_sum B). reflexivity.
  rewrite h1 in h4.
  assert (h5: x * (y + -y) = x).  const_eq.
  congruence.
assert (h6: y = y *x).
  assert (h7 : y*x = y*x + y*-y). const_eq.
  assert (h8: y*x + y*-y = y*(x + -y)).
    rewrite <- (dist_sum B). reflexivity.
  rewrite <- h1 in h8.
  assert (h9: y*(x + -x) = y). const_eq.
  congruence.
rewrite h2. rewrite h6 at 2.  comm_eq.
Qed.
Hint Resolve comp_unq.

(* x = --x *)
Lemma doub_neg : forall (x : Bt), x = --x.
intros x.  assert (h1 : le (---x) (-x)). assert (h2 : -x + --x = 1).
apply (comp_sum B). assert (h3: ---x * 1 = ---x). apply one_prod.
assert (h4: ---x * 1 = ---x * (-x + --x)). rewrite <- h2. reflexivity. 
assert (h5: --x * ---x = 0). apply (comp_prod B). assert (h6: ---x * --x = 0).
assert (h7: ---x * --x = --x * ---x). apply (comm_prod B).  rewrite <- h5. 
assumption. assert (h8: ---x * (-x + --x) = ---x * -x + ---x*--x).
apply (dist_sum B). assert (h9: ---x * --x = 0). rewrite <- h5.
apply (comm_prod B). unfold le. apply eq_ord.  
assert (h10: ---x * (-x + --x) = ---x * -x + 0).  rewrite <- h9.
rewrite h8. reflexivity. assert (h11 : ---x * -x + 0 = ---x * -x).
apply zero_sum. rewrite <- h11. rewrite <- h6. rewrite <- h8.  
rewrite <- h4. assumption.
assert (h12 : le (-x) (---x)).  unfold le.  apply eq_ord.
assert (h13 : -x * 1 = -x).  apply one_prod.
assert (h14: -x * 1 = -x * (--x + ---x)).  assert (h15: --x + ---x = 1).
apply (comp_sum B). rewrite <- h15. reflexivity. 
assert (h16: -x * (--x + ---x) = -x*--x + -x*---x).  apply (dist_sum B).
assert (h17: -x * --x = 0).  apply (comp_prod B).
assert (h18: -x * --x + -x * ---x = 0 + -x * ---x). rewrite <- h17.
reflexivity. assert (h19: 0 + -x*---x = -x*---x).  
assert (h20: -x*---x + 0 = -x*---x). apply zero_sum.
assert (h21: -x * ---x + 0 = 0 + -x * ---x). apply (comm_sum B).
rewrite <- h21. assumption.  rewrite <- h19. rewrite <- h17.
rewrite <- h16. rewrite <- h14. assumption.
assert (h22 : -x = ---x).  apply anti_sym_le. 
assumption.  assumption.  apply comp_unq. assumption.
Qed.
Hint Resolve doub_neg.

(* zero is characterized by the property x + y = x for all y*)
Lemma zero_char : forall (y : Bt), 
  (forall (x : Bt), x + y= x) -> y = 0.
intros y h1.  assert (h2 : 0 + y = 0). apply h1.
assert (h3: y + 0 = y). apply zero_sum.  assert (h4: y + 0 = 0 + y).
apply (comm_sum B).  rewrite <- h3. rewrite h4. assumption.
Qed.
Hint Resolve zero_char.

(* one is characterized by the property x * y = x for all y*)
Lemma one_char : forall (y : Bt), 
  (forall (x : Bt), x * y = x) -> y = 1.
intros y h1.  assert (h2 : 1 * y = 1). apply h1.
assert (h3: y * 1 = y). apply one_prod.  assert (h4: y * 1 = 1 * y).
apply (comm_prod B).  rewrite <- h3.  rewrite h4. assumption.
Qed.
Hint Resolve one_char.

(* complement is characterized by  x * y = 0, x + y =1 *)
Lemma comp_char : forall (x y : Bt), 
  x * y = 0 -> x + y = 1 -> y = -x.
intros x y h1 h2.
assert (h3: y = y * (x + -x)).  const_eq.
assert (h4: y * (x + -x) = (y * x) + (y * -x)). apply (dist_sum B).
assert (h5: (y * x) + (y * -x) = (x * y) + (y * -x)).
comm_eq. rewrite h1 in h5.
assert (h6 : 0 + y * -x = (-x * y) + 0).  comm_eq.
assert (h7: (-x * y) + 0 = -x * y + -x * x).  const_eq.
assert (h8: -x * y + -x * x = -x * (y + x)). 
rewrite (dist_sum B). reflexivity.
assert (h9: -x * (y + x) = -x * (x + y)).  comm_eq.
rewrite h2 in h9.
assert (h10 : -x * 1 = -x).  const_eq.
congruence.
Qed.
Hint Resolve comp_char.

Lemma comp_eq : forall (x y : Bt), -x = -y -> x = y.
intros x y h. assert (h1: x = --x).  apply doub_neg.
assert (h2: --x = --y). rewrite h. reflexivity.  assert (h3 : y = --y).
apply doub_neg. rewrite h3. rewrite <- h2. assumption.
Qed.
Hint Resolve comp_eq.

(* -0 = 1 *)
Lemma zero_comp: -0 = Bone (Bc B).
assert (h1: 0 * 1 = Bzero (Bc B)). apply one_prod.
assert (h2: 0 + 1 = Bone (Bc B)). apply one_sum.
assert (h3: 1 = -(Bzero (Bc B))). apply comp_char.  assumption.  assumption.
congruence.
Qed.
Hint Resolve zero_comp.

(* -1 = 0 *)
Lemma one_comp: -(1) = Bzero (Bc B).
assert (h1: 1 * 0 = Bzero (Bc B)). apply zero_prod. assert (h2: 1 + 0 = Bone (Bc B)).
apply zero_sum. assert (h3: 0 = -(Bone (Bc B))). apply comp_char.
assumption. assumption. rewrite h3. reflexivity.
Qed.
Hint Resolve one_comp.

Lemma de_mor_sum : forall (x y : Bt), -(x + y) = -x * -y.
intros x y.  assert (h0: (x + y ) * (-x * -y) = 0). 
assert (h1: (x + y) * (-x * -y) = x * (-x * -y) + y * (-x * -y)).
apply dist_sum_r.  rewrite h1.  assert (h2: x * (-x * -y) = (x * -x) * -y).
apply (assoc_prod B). rewrite h2.  assert (h3: x * -x = 0). apply (comp_prod B).
rewrite h3.  assert (h4: 0 * -y = 0). assert (h5: -y * 0 = 0).  apply zero_prod.
rewrite <- h5 at 2.  apply (comm_prod B). rewrite h4.
assert (h6: -x * -y = -y * -x).  apply (comm_prod B). rewrite h6.
assert (h7: y * (-y * -x) = (y * -y) * -x). apply (assoc_prod B).
rewrite h7. assert (h8: y * -y = 0).  apply (comp_prod B). rewrite h8.
assert (h9: 0 * -x = 0). assert (h10: (-x * 0) = 0). apply zero_prod.
rewrite <- h10 at 2. apply (comm_prod B). rewrite h9. apply idem_sum.
assert (h11: x + y + -x * -y = 1).  
  assert (h12: x + y + -x * -y = x * (y + -y) + y + -x * -y).
    assert (h13: y + -y = 1).  auto.
    assert (h14: x * 1 = x).  auto.
    congruence.
  assert (h15: x * (y + -y) + y + -x * -y = x * y + x * -y + y + -x * -y).
    assert (h16: x * (y + -y) = x * y + x * -y). auto.
    congruence.
  assert (h16: x * y + x * -y + y + -x * -y = y + x * -y + -x * -y).
    assert (h17: x * y + x * -y + y = x*y + y + x * -y).  comm_eq.
    assert (h18: (x + 1) * y = y).
      assert (h19: x + 1 = 1). auto.
      assert (h20: 1 * y = y * 1). auto.
      assert (h21: y * 1 = y).  auto.
    congruence.
    assert (h22: (x + 1) * y = x * y + y).
      assert (h23: (x + 1) * y = x * y + 1 * y).  auto.
      assert (h24: 1 * y = y * 1).  auto.
      assert (h25: y * 1 = y).  auto.
    congruence.
  congruence.
  assert (h26: y + x * -y + -x * -y = y + -y).
    assert (h27: y + (x * -y + -x * -y) = y + x * -y + -x * -y). auto.
    assert (h28: (x + -x) * -y = x * -y + -x * -y).  auto.
    assert (h29: x + -x = 1).  auto.
    assert (h30: 1 * -y = -y * 1).  auto.
    assert (h31: -y * 1 = -y).  auto.
  congruence. 
  assert (y + -y = 1).  auto.
congruence.
assert (h32: -x * -y = - (x + y)).  auto.
congruence.
Qed.   
Hint Resolve de_mor_sum.

Lemma de_mor_prod : forall (x y : Bt), -(x*y) = -x + -y.
intros x y.
assert (h1 : (x * y) + (-x + -y) = 1).
  assert (h2 : (x * y) + (-x + -y) = (x + (-x + -y))*(y + (-x + -y))).
    apply dist_prod_r.
  assert (h3: (x + (-x + -y)) * (y + (-x + -y)) = 
              ((x + -x) + -y) * ((y + -y) + -x)).
    comm_eq.
  assert (h4 : ((x + -x) + -y) * ((y + -y) + -x) = 1).
    const_eq.
  congruence.
assert (h5 : (x * y) * (-x + -y) = 0).
  assert (h6: (x * y) * (-x + -y) = (x + (y * -y)) * y * (-x + -y)).
    const_eq.
  assert (h7: (x + (y * -y)) * y * (-x + -y) = (x + y) * (x + -y) * y * (-x + -y)).
    rewrite dist_prod with (p := x) (n := y) (m := -y). reflexivity.
  assert (h8: (x + y) * (x+ -y) * y * (-x + -y) = (x + y) * y * (x + -y) * (-x + -y)).
    comm_eq.
  assert (h9: (x + y) * y *(x + -y) *(-x + -y) = y * (x+-y) * (-x + -y)).
    assert (h10 : (x + y) * y*(x + -y) * (-x + -y) = 
                  y * (y + x)  * (x + -y) * (-x + -y)).
      comm_eq.
    assert (h11: y * (y+x) * (x + -y) * (-x + -y) = y*(x + -y) * (-x + -y)).
      rewrite (abs_prod B) with (n := y) (m := x). reflexivity.
    congruence.
  assert (h12: y * (x + -y) * (-x + -y) = y * ((x * -x ) + -y)).
    assert (h13: y * (x + -y) * (-x + -y) = y * ((x + -y) * (-x + -y))).
      assoc_eq.
    rewrite <- dist_prod_r with (n := x) (m := -x) (p := -y) in h13.
    assumption.
  assert (h13 : y * ((x * -x) + -y) = 0).
    assert (h14 : (x * -x) + -y = -y).
      const_eq.
    rewrite h14.
      const_eq.
  congruence.
symmetry.
apply comp_char; auto.
Qed.
Hint Resolve de_mor_prod.    

Lemma times_eq_plus : forall (x y: Bt), x*y = -((-x) + (-y)).
intros x y.
rewrite de_mor_sum with (x := -x) (y := -y).
rewrite <- doub_neg. rewrite <- doub_neg.
reflexivity.
Qed.

Lemma plus_eq_times : forall (x y:Bt), x+y = -((-x)*(-y)).
intros x y.
rewrite de_mor_prod with (x := -x) (y := -y).
rewrite <- doub_neg. rewrite <- doub_neg.
reflexivity.
Qed.

Lemma mono_sum : forall (x x' y y' : Bt), 
  le x x' -> le y y' -> le (x+y) (x' + y').
intros x x' y y' h1 h2.
assert (h3: (x + y) + (x' + y') = x' + y').
assert (h4: (x + y) + (x' + y') = (x+x') + (y + y')).  comm_eq.
assert (h8: x + x' = x').  auto.
assert (h9: y + y' = y').  auto.
congruence. auto.
Qed.

Lemma mono_prod : forall (x x' y y': Bt), 
  le x x' -> le y y' -> le (x * y) (x' * y').
intros x x' y y' h1 h2.
assert (h3: (x * y) * (x' * y') = x * y).
assert (h4: (x * y) * (x' * y') = (x * x') *(y * y')).  comm_eq.
assert (h8: x * x' = x).  
  assert (h8a: x + x' = x'). auto.  apply eq_ord. assumption.
  (* turn this functionality into a separate lemma, to avoid repetition *)
assert (h9: y * y' = y).
  assert (h9a: y + y' = y').  auto.  apply eq_ord. assumption.
congruence.
apply eq_ord.
assumption.
Qed.

Lemma mono_comp : forall (x x' : Btype (Bc B)), le x x' -> le (-x') (-x).
intros x x' h1.
assert (h2: -(x * x') = -x + -x').  apply de_mor_prod.
assert (h3: x * x' = x).
  assert (h4: x + x' = x').  auto.
  apply eq_ord.  assumption.
assert (h5: -x + -x' = -x).
  rewrite <- h2.  congruence.
unfold le.  
assert (h6: -x + -x' = -x' + -x).  auto.
congruence.
Qed.
End Bscs.
(*End Basics.*)

(*Theorems*)
(*
Arguments flatten_aux_valid_Bt_p [B] l def t t'.
Arguments flatten_aux_valid_Bt_t [B] l def t t'.
Arguments flatten_valid_Bt [B] l def t.
Arguments flatten_valid_Bt_2 [B] t t' l def _.
Arguments bin_Bt_ref_p [B] l def t1 t2.
Arguments bin_Bt_ref_t [B] l def t1 t2.
Arguments insert_eq_p [B] l def t0 t.
Arguments insert_eq_t [B] l def t0 t.
Arguments sort_eq [B] l def t.
Arguments sort_eq_2 [B] l def t1 t2 _.
Arguments dist_sum_r [B] n m p.
Arguments dist_prod_r [B] n m p.
Arguments idem_sum [B] b.
Arguments idem_prod [B] b.
Arguments eq_ord [B] x y.
Arguments refl_le [B] x.
Arguments anti_sym_le [B] x y _ _.
Arguments trans_le [B] x y z _ _.
Arguments lat_sum [B] x y.
Arguments lat_prod [B] x y.
Arguments inf_unq [B] S x y _ _.
Arguments sup_unq [B] S x y _ _.
Arguments zero_min [B] x.
Arguments one_max [B] x.
Arguments zero_sum [B] x.
Arguments zero_prod [B] x.
Arguments one_sum [B] x.
Arguments one_prod [B] x.
Arguments comp_unq [B] x y _.
Arguments doub_neg [B] x.
Arguments zero_char [B] y _.
Arguments one_char [B] y _.
Arguments comp_char [B] x y _ _.
Arguments comp_eq [B] x y _.
Arguments zero_comp [B].
Arguments one_comp [B].
Arguments de_mor_sum [B] x y.
Arguments de_mor_prod [B] x y.
Arguments mono_sum [B] x x' y y' _ _.
Arguments mono_prod [B] x x' y y' _ _.
Arguments mono_comp [B] x x' _.
(*Definitions*)
Arguments le [B] x y.
Arguments lb [B] S b.
Arguments ub [B] S b.
Arguments inf [B] S b.
Arguments sup [B] S b.
End bab.*)
