(* Copyright (C) 2014, Daniel Wyckoff, except for the portions so labeleled
which I got from Daniel Schepler*)



(*		   GNU LESSER GENERAL PUBLIC LICENSE
                       Version 3, 29 June 2007

 Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.


  This version of the GNU Lesser General Public License incorporates
the terms and conditions of version 3 of the GNU General Public
License, supplemented by the additional permissions listed below.

  0. Additional Definitions. 

  As used herein, "this License" refers to version 3 of the GNU Lesser
General Public License, and the "GNU GPL" refers to version 3 of the GNU
General Public License.

  "The Library" refers to a covered work governed by this License,
other than an Application or a Combined Work as defined below.

  An "Application" is any work that makes use of an interface provided
by the Library, but which is not otherwise based on the Library.
Defining a subclass of a class defined by the Library is deemed a mode
of using an interface provided by the Library.

  A "Combined Work" is a work produced by combining or linking an
Application with the Library.  The particular version of the Library
with which the Combined Work was made is also called the "Linked
Version".

  The "Minimal Corresponding Source" for a Combined Work means the
Corresponding Source for the Combined Work, excluding any source code
for portions of the Combined Work that, considered in isolation, are
based on the Application, and not on the Linked Version.

  The "Corresponding Application Code" for a Combined Work means the
object code and/or source code for the Application, including any data
and utility programs needed for reproducing the Combined Work from the
Application, but excluding the System Libraries of the Combined Work.

  1. Exception to Section 3 of the GNU GPL.

  You may convey a covered work under sections 3 and 4 of this License
without being bound by section 3 of the GNU GPL.

  2. Conveying Modified Versions.

  If you modify a copy of the Library, and, in your modifications, a
facility refers to a function or data to be supplied by an Application
that uses the facility (other than as an argument passed when the
facility is invoked), then you may convey a copy of the modified
version:

   a) under this License, provided that you make a good faith effort to
   ensure that, in the event an Application does not supply the
   function or data, the facility still operates, and performs
   whatever part of its purpose remains meaningful, or

   b) under the GNU GPL, with none of the additional permissions of
   this License applicable to that copy.

  3. Object Code Incorporating Material from Library Header Files.

  The object code form of an Application may incorporate material from
a header file that is part of the Library.  You may convey such object
code under terms of your choice, provided that, if the incorporated
material is not limited to numerical parameters, data structure
layouts and accessors, or small macros, inline functions and templates
(ten or fewer lines in length), you do both of the following:

   a) Give prominent notice with each copy of the object code that the
   Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the object code with a copy of the GNU GPL and this license
   document.

  4. Combined Works.

  You may convey a Combined Work under terms of your choice that,
taken together, effectively do not restrict modification of the
portions of the Library contained in the Combined Work and reverse
engineering for debugging such modifications, if you also do each of
the following:

   a) Give prominent notice with each copy of the Combined Work that
   the Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the Combined Work with a copy of the GNU GPL and this license
   document.

   c) For a Combined Work that displays copyright notices during
   execution, include the copyright notice for the Library among
   these notices, as well as a reference directing the user to the
   copies of the GNU GPL and this license document.

   d) Do one of the following:

       0) Convey the Minimal Corresponding Source under the terms of this
       License, and the Corresponding Application Code in a form
       suitable for, and under terms that permit, the user to
       recombine or relink the Application with a modified version of
       the Linked Version to produce a modified Combined Work, in the
       manner specified by section 6 of the GNU GPL for conveying
       Corresponding Source.

       1) Use a suitable shared library mechanism for linking with the
       Library.  A suitable mechanism is one that (a) uses at run time
       a copy of the Library already present on the user's computer
       system, and (b) will operate properly with a modified version
       of the Library that is interface-compatible with the Linked
       Version. 

   e) Provide Installation Information, but only if you would otherwise
   be required to provide such information under section 6 of the
   GNU GPL, and only to the extent that such information is
   necessary to install and execute a modified version of the
   Combined Work produced by recombining or relinking the
   Application with a modified version of the Linked Version. (If
   you use option 4d0, the Installation Information must accompany
   the Minimal Corresponding Source and Corresponding Application
   Code. If you use option 4d1, you must provide the Installation
   Information in the manner specified by section 6 of the GNU GPL
   for conveying Corresponding Source.)

  5. Combined Libraries.

  You may place library facilities that are a work based on the
Library side by side in a single library together with other library
facilities that are not Applications and are not covered by this
License, and convey such a combined library under terms of your
choice, if you do both of the following:

   a) Accompany the combined library with a copy of the same work based
   on the Library, uncombined with any other library facilities,
   conveyed under the terms of this License.

   b) Give prominent notice with the combined library that part of it
   is a work based on the Library, and explaining where to find the
   accompanying uncombined form of the same work.

  6. Revised Versions of the GNU Lesser General Public License.

  The Free Software Foundation may publish revised and/or new versions
of the GNU Lesser General Public License from time to time. Such new
versions will be similar in spirit to the present version, but may
differ in detail to address new problems or concerns.

  Each version is given a distinguishing version number. If the
Library as you received it specifies that a certain numbered version
of the GNU Lesser General Public License "or any later version"
applies to it, you have the option of following the terms and
conditions either of that published version or of any later version
published by the Free Software Foundation. If the Library as you
received it does not specify a version number of the GNU Lesser
General Public License, you may choose any version of the GNU Lesser
General Public License ever published by the Free Software Foundation.

  If the Library as you received it specifies that a proxy can decide
whether future versions of the GNU Lesser General Public License shall
apply, that proxy's public statement of acceptance of any version is
permanent authorization for you to choose that version for the
Library.*)




Require Import LogicUtilities.
Require Import TypeUtilities.
Require Import EnsemblesImplicit.
(*I dont' know why I need the following seemingly superfluous Require, 
  which the previous should handle, but I do. *)
Require Import Finite_sets_facts. 
Require Import ClassicalEpsilon.
Require Export List.
Require Import FunctionProperties.
Require Import Constructive_sets.
Require Import DecidableDec.
Require Import FunctionalExtensionality.

(*A list of all sequence lists obtained by choosing one element from each 
  list in [l], in the same order as the lists in [l]*)
Fixpoint list_of_lists_seqs {T:Type} (l:list (list T)) : list (list T) :=
  match l with 
  | nil => cons nil nil
  | al::l' => map (fun x:(T * (list T)) => cons (fst x) (snd x)) 
    (list_prod al (list_of_lists_seqs l'))
  end.

Fixpoint list_to_set {T:Type} (l:list T) : Ensemble T :=
  match l with
  | nil => Empty_set _
  | cons a l' => Add (list_to_set l') a
  end.

Lemma list_to_set_finite : forall {T:Type} (l:list T), 
  Finite _ (list_to_set l).
intros T l.
induction l as [|a l h1].
simpl. auto with sets.
simpl. apply Add_preserves_Finite; assumption.
Qed.

Lemma list_to_set_in_eq : forall {T:Type} (l:list T) (x:T),
  In x l <-> Ensembles.In (list_to_set l) x.
intros T l.
induction l as [|a l h2].
(* nil *)
intros x. split. intro. contradiction. 
intro. contradiction.
(* cons *)
intro x. split.
  (* -> *)
  intro h1. simpl in h1. destruct h1 as [h1l | h1r]. 
    (* h1l *)
    subst. apply Add_intro2.
    (* h1r *)
    rewrite h2 in h1r.
    simpl. apply Add_intro1; assumption.
  (* <- *)
  intro h1. simpl in h1. inversion h1 as [y h3|y h4]. 
    (* h3 *)
    subst. rewrite <- h2 in h3. simpl. right; assumption.
    (* h4 *)
    subst. inversion h4. subst. simpl. left; reflexivity.
Qed.

Lemma list_to_set_bij : forall {T:Type} (l:list T), 
  exists f:{x:T | In x l} -> {x:T | Ensembles.In (list_to_set l) x}, 
  bijective f.
intros T l.
pose proof (list_to_set_in_eq l) as h1.
assert (h2:forall x : T, In x l -> Ensembles.In (list_to_set l) x).
  intros. rewrite <- h1; assumption.
exists (fun t:{x : T | In x l} => exist _ (proj1_sig t) (h2 _ (proj2_sig t))).
red. split.
(* inj *)
red. intros ? ? h3.  apply exist_injective in h3. 
apply proj1_sig_injective. assumption.
(* surj *)
red. intros y.
destruct (h1 (proj1_sig y)) as [? h1r].
specialize (h1r (proj2_sig y)).
exists (exist _ (proj1_sig y) h1r).
simpl. rewrite unfold_sig.
apply existTexist. apply subsetT_eq_compat. reflexivity.
Qed.

Lemma list_to_set_finitet : forall {T:Type} (l:list T), 
  FiniteT {x:T | In x l}.
intros T l.
pose proof (list_to_set_bij l) as h1.
pose proof (list_to_set_finite l) as h2.
pose proof (Finite_ens_type _ h2) as h3.
destruct h1 as [f h1].
pose proof (bijective_impl_invertible _ h1) as h4.
pose proof (invertible_impl_inv_invertible _ h4) as h5.
apply (bij_finite _ _ _ h3 h5).
Qed.

Lemma finite_set_list : forall {T:Type} (E:Ensemble T), Finite _ E ->
  exists (l:list T), E = list_to_set l.
intros T E h1.
induction h1 as [| E h2 h3 t h4].
(* nil *)
exists nil.
unfold list_to_set. reflexivity.
(* Add *)
destruct h3 as [l h5].
exists (t::l).
unfold list_to_set.
unfold list_to_set in h5.
rewrite <- h5.
reflexivity.
Qed.

Definition set_to_list {T:Type} (E:Ensemble T) (pf:Finite _ E) : 
  {l:list T | E = list_to_set l}.
pose proof (finite_set_list E pf) as h1.
refine (constructive_indefinite_description _ h1).
Defined.

Definition map_sig {T:Type} (l:list T) : list {x:T | In x l}.
induction l as [|a l lsig].
refine nil.
pose (map (fun i:{x:T | In x l} => 
  exist (fun x:T => In x (a::l)) 
    (proj1_sig i) (in_cons a _ _ (proj2_sig i))) lsig) as lsig'.
pose (exist (fun x:T => In x (a::l)) a (in_eq a l)) as a'.
refine (a'::lsig').
Defined.

Lemma map_sig_in : forall {T:Type} (l:list T) (i:{x:T | In x l}), 
  In i (map_sig l).
intros T l i.
induction l as [|a l h1]. 
(* nil *)
pose proof (proj2_sig i) as h5. simpl in h5. contradiction.    
(* a :: l *)
simpl. 
pose proof (proj2_sig i) as h5.
simpl in h5.
  destruct h5 as [h5l | h5r].
  (* h5l *)
  left.
  rewrite unfold_sig.
  apply existTexist.
  apply subsetT_eq_compat. assumption.
  (* h5r *)
  right.
  rewrite in_map_iff.
  pose (exist (fun (x:T) => In x l) (proj1_sig i) h5r) as i'.
  exists i'.
  split.
  unfold i'.
  simpl. rewrite unfold_sig. apply existTexist. apply subsetT_eq_compat. 
  reflexivity.
  apply h1.
Qed.

Lemma cons_inj : forall {T:Type} (a1 a2:T) (l1 l2:list T), a1::l1 = a2::l2 ->
  a1 = a2 /\ l1 = l2.
intros ? ? ? ? ? h1.
inversion h1. split; trivial.
Qed.

Lemma cons_eq : forall {T:Type} (a1 a2:T) (l1 l2:list T), 
  a1::l1 = a2::l2 <->  a1 = a2 /\ l1 = l2.
intros T a1 a2 l1 l2.
split.
intro h1. apply cons_inj; assumption.
intro h1. destruct h1; subst. reflexivity.
Qed.

(* Maybe fix, but don't delete just yet *)
(* Commenting out to make make-able *)

(*
Theorem normal_form : forall (l:list Bt) (pf:B' = Gen (list_to_set l)), 
  (forall (p:BA'.Bt), atom p <-> (p <> 0) /\ 
  (exists a:{xl:Bt | In xl l} -> sign, 
  (times_list (map (fun i:{xl:Bt | In xl l} => eps (proj1_sig i)(a i)) (map_sig l))) = 
  (transfer_gen (transfer_eq _ _ pf p)))) /\
  (forall (b:BA'.Bt), exists !(S:Ensemble BA'.Bt),   
   exists (pfs: Finite _ S),  
  (forall x:BA'.Bt, Ensembles.In S x -> atom x) /\    
  plus_list (proj1_sig (set_to_list S pfs)) = b). *)

(*
Lemma seq_to_fun_ex : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (ll:list (T1 * T2)), In ll (list_power l1 l2) -> 
  exists f:{t1:T1 | In t1 l1}->{t2:T2 | In t2 l2}, 
  ll = map (fun t:{t1:T1 | In t1 l1} => ((proj1_sig t), (proj1_sig (f t)))).
*)


(*
Lemma seq_val_lem : forall {T1 T2:Type} {l1:list T1} {l2:list T2} (ll:list (T1*T2)), 
  In ll (list_power l1 l2) -> forall (t1:T1) (pf2:In t1 l1), exists ! t2:T2, 
  In (t1, t2) ll.
intros T1 T2 l1 l2 ll h1 t1 h2.
(*induction l1. contradiction.
unfold list_power in h1. 
rewrite in_flat_map in h1.
destruct h1 as [xl h3].
destruct h3 as [h3l h3r].
rewrite in_map_iff in h3r.
destruct h3r as [t2 h4].
destruct h4 as [h4l h4r].*)


induction ll; induction l1. 
(* nil nil *)
contradiction.
(* nil a::l1 *)
simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [lx h3].
rewrite in_map_iff in h3.
destruct h3 as [h3l h3r].
destruct h3r as [h3rl h3rr].
destruct h3rr.
discriminate.

(* a::ll nil *)
contradiction.
(* a::ll a0::l1 *)
unfold list_power in h1. simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [lx h3].
destruct h3 as [h3l h3r].
rewrite in_map_iff in h3r.
destruct h3r as [t2 h4].
destruct h4 as [h4l h4r].
apply cons_inj in h4l.
destruct h4l; subst.

exists t2.
red.
split.
constructor.
apply in_inv in h2.
destruct h4 as [h4l h4r].
inversion h4l as [h4ll].
subst.
destruct h2 as [h2l | h2r]. subst. reflexivity.
  (* h2l *)
  subst. tauto. 

  (* h2r *)



 T2.
induction l1. contradiction.
apply in_inv in pf2.
destruct pf2.
*)

(* Theorem list_ext *)

(*Definition list_eq {T:Type} (l1 l2:list T) : Prop :=
  match l1 with 
  | nil => l2 = nil
  | a::l1' => *)

(*
Definition fun_seq {T1 T2:Type} {l1:list T1} {l2:list T2} 
  (l:list (T1*T2)) (pf:In l (list_power l1 l2)) (t1:T1) : T2.
induction l.


(* nil *)


destruct (list_power l1 l2). contradiction. simpl in pf.

 simpl in pf.
*)

(*
Lemma list_seq_ext : forall {T1 T2:Type} {l1:list T1} {l2:list T2} (la lb:list (T1*T2)), 
  In la (list_power l1 l2) -> In lb (list_power l1 l2) -> 
  (la = lb <-> forall (t:T1) (pra prb:T1*T2), 

Lemma distinct_list_seqs_differ_at_point : forall {T1 T2:Type} 
  (l1:list T1) (l2:list T2) (la lb:list (T1*T2)), 
  In la (list_power l1 l2) -> In lb (list_power l1 l2) -> la <> lb ->
  exists t1:T1, forall pra prb:T1 * T2, fst pra = t1 -> fst prb = t1 ->
  snd pra <> snd prb.
intros T1 T2 l1 l2 la lb h1 h2 h3.
Check distinct_functions_differ_at_point.
pose (fun t:{t1:T1 | In t la} => 



apply NNPP.
intro h4.
pose proof (not_ex_all_not _ _ h4) as h5.
simpl in h5.
assert (h6:forall n:T1, forall pra prb:T1 * T2, fst pra = n -> fst prb = n ->
  snd pra = snd prb).
firstorder.
tauto.


intros X Y f g h1.
apply NNPP.
intro h2.
pose proof (not_ex_all_not _ _ h2) as h3.
simpl in h3.
assert (h4:forall x:X, f x = g x).
  intro x.
  specialize (h3 x).
  tauto.
pose proof (functional_extensionality _ _ h4).
contradiction.
Qed.
*)

Definition list_fun {T1 T2:Type} (l1:list T1) (l2:list T2) :=
  {x:T1 | In x l1} -> {x:T2 | In x l2}.

Lemma in_list_power1 : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:list (T1 * T2)), In pl (list_power l1 l2) -> forall (pr:T1*T2),
  In pr pl -> In (fst pr) l1.
intros T1 T2 l1 l2. 
induction l1 as [|b l1 h4]. 
intros pl h1. 
simpl in h1.
destruct h1 as [h1l | h1r].
subst. contradiction. contradiction.
intros pl h1.
simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [pl' h5].
destruct h5 as [h5l h5r].
rewrite in_map_iff in h5r.
destruct h5r as [t2 h6].
destruct h6 as [h6l h6r].
specialize (h4 pl' h5l).
intros pr h7.
subst.
destruct h7 as [h7l | h7r].
subst. simpl. left. reflexivity.
specialize (h4 pr h7r).
right.
assumption.
Qed.

Lemma in_list_power2 : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:list (T1 * T2)), In pl (list_power l1 l2) -> forall (pr:T1*T2),
  In pr pl -> In (snd pr) l2.
intros T1 T2 l1 l2.
induction l1 as [|b l1 h4]. 
intros pl h1. 
simpl in h1.
intros pr h2.
destruct h1 as [h1l | h1r].
subst. contradiction. contradiction.
intros pl h2.
simpl in h2.
rewrite in_flat_map in h2.
destruct h2 as [pl' h3].
destruct h3 as [h3l h3r].
rewrite in_map_iff in h3r.
destruct h3r as [t2 h5].
destruct h5 as [h5l h5r].
subst.
intros pr h6.
destruct h6 as [h6l | h6r].
subst. simpl. assumption.
specialize (h4 _ h3l _ h6r).
assumption.
Qed.






(*This just adjusts the sigma condition for a function defined on a cons to the
  popped list *)
Definition f_list_pop {T1 T2:Type} {l:list T1} {b:T1}
  (f:{x:T1 | In x (b::l)}->T2) := 
  fun (i:{x:T1 | In x l}) => 
    (f (exist (fun x : T1 => In x (b::l)) (proj1_sig i)
    (in_cons b (proj1_sig i) l (proj2_sig i)))).

Definition f_list_pop' {T1 T2:Type} {l1:list T1} {l2:list T2} {b:T1}
  (f:list_fun (b::l1) l2) := f_list_pop f.

Definition f_list_push {T1 T2:Type} {l:list T1} (pr:T1*T2) 
  (f:{x:T1 | In x l}->T2) : {x:T1 | In x ((fst pr)::l)} -> T2.
intro i.
destruct (or_to_sumbool _ _ (proj2_sig i)) as [? | h1].
refine (snd pr). 
refine (f (exist _ (proj1_sig i) h1)).
Defined.

Definition f_list_push' {T1 T2:Type} {l1:list T1} {l2:list T2} (pr:T1*T2)
  (f:(list_fun l1 l2)) : (list_fun ((fst pr)::l1) ((snd pr)::l2)).
red.  intro i.
destruct (or_to_sumbool _ _ (proj2_sig i)) as [? | h1].
refine (exist _ (snd pr) (in_eq (snd pr) l2)).
refine (exist (fun x:T2 => In x (snd pr :: l2)) _ 
  (in_cons (snd pr) _ _ (proj2_sig (f (exist _ (proj1_sig i) h1))))).
Defined.

Definition f_list_push'' {T1 T2:Type} {l1:list T1} {l2:list T2} (pr:T1*T2)
  (f:(list_fun l1 ((snd pr)::l2))) : list_fun ((fst pr)::l1) ((snd pr)::l2).
red. intro i.
destruct (or_to_sumbool _ _ (proj2_sig i)) as [? | h1].
refine (exist _ (snd pr) (in_eq (snd pr) l2)).
refine (f (exist _ (proj1_sig i) (h1))).
Defined.

Lemma lfun_to_pair_list_lemma : forall {T1 T2:Type} {l1:list T1} {l2:list T2} 
  (f:(list_fun l1 l2)), exists pl:list (T1*T2), exists h1:(In pl (list_power l1 l2)), 
  (forall (pr:T1*T2) (h2:In pr pl), 
  (snd pr) = (proj1_sig (f (exist _ (fst pr) (in_list_power1 _ h1 _ h2))))).
intros T1 T2 l1 l2 f.
induction l1 as [|a1 l1 h2]; destruct l2 as [| a2 l2].
(* nil nil *)
exists nil.  simpl. exists (or_introl (eq_refl nil)).
intros. contradiction. 
(* nil a2::l2*)
exists nil. simpl. exists (or_introl (eq_refl nil)).
intros. contradiction.  
(* a1::l1 nil *)
pose proof (proj2_sig (f (exist _ a1 (in_eq a1 l1)))) as h1.
simpl in h1. contradiction.
(*a1::l1 a2::l2 *)
pose (f_list_pop f) as f'.
specialize (h2 f').
destruct h2 as [pl h4].
pose (cons (a1, proj1_sig (f (exist _ a1 (in_eq a1 l1)))) pl) as pl'.
exists pl'.
destruct h4 as [h1 h5].
assert (h6:In pl' (list_power (a1 :: l1) (a2 :: l2))).
  simpl. rewrite in_flat_map. exists pl.
  split. assumption. simpl.
  pose proof (proj2_sig 
  (f (exist (fun x : T1 => a1 = x \/ In x l1) a1 (in_eq a1 l1)))) as h6.
  simpl in h6.
  destruct h6 as [h6l | h6r].
  left. rewrite h6l. reflexivity.
  right. rewrite in_map_iff.
  exists (proj1_sig (f (exist (fun x : T1 => In x (a1 :: l1)) a1 (in_eq a1 l1)))).
  simpl. split. unfold pl'. reflexivity. assumption.
exists h6.
intros pr h2.
unfold pl' in h2.
destruct h2 as [h2l | h2r].
(* h2l *)
simpl in h2l.
assert (h7: fst pr = a1). rewrite <- h2l. simpl. reflexivity.
rewrite <- h2l at 1.
assert (h8:(exist (fun x : T1 => a1 = x \/ In x l1) a1 (in_eq a1 l1)) = 
  (exist (fun x : T1 => In x (a1 :: l1)) (fst pr)
        (in_list_power1 pl' h6 pr (or_introl h2l)))).
  apply existTexist.
  apply subsetT_eq_compat. rewrite h7. reflexivity.
rewrite h8. simpl. reflexivity.
(* h2r *)
pose proof (h5 pr h2r) as h7.
rewrite h7.
unfold f'. unfold f_list_pop.
simpl.
unfold pl'.
assert (h8:exist (fun x : T1 => a1 = x \/ In x l1) (fst pr)
        (in_cons a1 (fst pr) l1 (in_list_power1 pl h1 pr h2r)) =   
  exist (fun x : T1 => a1 = x \/ In x l1) (fst pr)
        (in_list_power1
           ((a1,
            proj1_sig
              (f (exist (fun x : T1 => In x (a1 :: l1)) a1 (in_eq a1 l1))))
            :: pl) h6 pr (or_intror h2r))).
  apply existTexist.
  apply subsetT_eq_compat. reflexivity.
rewrite h8.
reflexivity.
Qed.

Definition lfun_to_pair_list {T1 T2:Type} {l1:list T1} {l2:list T2} 
  (f:(list_fun l1 l2)) : 
  {pl:list (T1*T2) | exists h1:(In pl (list_power l1 l2)), 
    forall (pr:T1*T2) (h2:In pr pl), 
    snd pr = proj1_sig (f (exist _ (fst pr) (in_list_power1 _ h1 _ h2)))}.
pose proof (lfun_to_pair_list_lemma f) as h1.
refine (constructive_indefinite_description _ h1).
Defined.

Lemma in_list_power_cons : forall {T1 T2:Type} {b:T1} {l1:list T1} {l2:list T2}
  (pl:(list (T1*T2))), In pl (list_power (b::l1) l2) -> 
  exists (pl':(list (T1*T2))) (c:T2), In c l2 /\ In pl' (list_power l1 l2) /\ 
  pl = (b, c) :: pl'.
intros T1 T2 b l1 l2 pl h1.
simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [pl' h2].
destruct h2 as [h2 h3].
rewrite in_map_iff in h3.
destruct h3 as [c h4].
destruct h4 as [h4l h4r].
symmetry in h4l.
exists pl'.
exists c.
repeat split; assumption.
Qed.

Lemma in_list_power_cons' : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:list (T1*T2)) (pr:T1 * T2) (c:T1), In (pr::pl) (list_power (c::l1) l2) ->
  In pl (list_power l1 l2).
intros T1 T2 l1 l2 pl pr c h1.
simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [pl' h1].
destruct h1 as [h1l h1r].
rewrite in_map_iff in h1r.
destruct h1r as [x h2].
destruct h2 as [h2l h2r].
rewrite cons_eq in h2l.
destruct h2l. subst. assumption.
Qed.

(*Try to do unique existence, but not right away!*)

Lemma in_list_power_fun : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl : list (T1*T2)), In pl (list_power l1 l2) -> 
  forall a:T1, In a l1 -> exists b:T2, In b l2 /\ In (a, b) pl.
intros T1 T2.
induction l1 as [|c l1 h1]; induction pl as [|ap pl h2].
(* nil nil *)
intros; contradiction.
(* nil cons *)
intro h1. simpl in h1. destruct h1; [discriminate | contradiction].
(* cons nil *)
intro h2. simpl in h2. rewrite in_flat_map in h2.
destruct h2 as [x [? h2]].
rewrite in_map_iff in h2.
destruct h2 as [? []]; discriminate.
(* cons cons *)
intros h3 a h4.
pose proof (in_eq ap pl) as h5.
pose proof (in_list_power1 _ h3 _ h5) as h6.
destruct h4 as [h4l | h4r]; destruct h6 as [h6l | h6r].
  (*h4l h6l*)
  subst. exists (snd ap). split. 
    apply (in_list_power2 _ h3 _ h5).
    subst. rewrite <- surjective_pairing. auto with datatypes.
  (*h4l h6r*)
  subst.
  pose proof (in_list_power_cons' _ _ _ h3) as h7.
  pose proof (h1 _ _ h7 _ h6r) as h8.
  destruct h8 as [b [h9l h9r]].
  pose proof (in_list_power1 _ h3 _ h5) as h10.
    destruct h10 as [h10l | h10r].
    (* h10l *)
    subst. exists b. split; [assumption | apply in_cons; assumption].
    (* h10r *)
    pose proof (h1 _ _ h7 _ h10r) as h11.
    simpl in h3. rewrite in_flat_map in h3. destruct h3 as [pl' [h12 h13]].
    rewrite in_map_iff in h13. destruct h13 as [b' [h13l h13r]].
    rewrite cons_eq in h13l. destruct h13l; subst.
    exists b'. split; assumption.
  (*h4r h6l*)
  subst.
  pose proof (in_list_power_cons' _ _ _ h3) as h6.
  pose proof (h1 _ _ h6 _ h4r) as h7.
  destruct h7 as [b [h7l h7r]].
  exists b. split; [assumption | apply in_cons; assumption].
  (*h4r h6r*)
  pose proof (in_list_power_cons' _ _ _ h3) as h6.
  pose proof (h1 _ _ h6 _ h4r) as h7.
  destruct h7 as [b [h7l h7r]].
  exists b. split; [assumption | apply in_cons; assumption].
Qed.


Definition pair_list_to_lfun {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl : list (T1*T2)) (pf:In pl (list_power l1 l2)) : (list_fun l1 l2).
red. intro x1. 
pose proof (in_list_power_fun _ pf _ (proj2_sig x1)) as h1.
pose proof (constructive_indefinite_description _ h1) as x2'.
destruct x2' as [x2' [h2l h2r]].
refine (exist _ x2' h2l).
Defined.

Lemma pair_list_to_lfun_in : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl : list (T1*T2)) (pf:In pl (list_power l1 l2)) (x1:{t1:T1 | In t1 l1}),
  In (proj1_sig x1, proj1_sig ((pair_list_to_lfun pl pf) x1)) pl.
intros T1 T2 l1 l2 pl h1 x1.
unfold pair_list_to_lfun.
destruct (constructive_indefinite_description
           (fun b : T2 => In b l2 /\ In (proj1_sig x1, b) pl)
           (in_list_power_fun pl h1 (proj1_sig x1) (proj2_sig x1))) as [b h3].
destruct h3 as [h3l h3r].
simpl.
assumption.
Qed.

(*
Lemma pair_list_to_lfun_cons : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:list (T1*T2)) (pr:T1*T2) (pf:In (pr::pl) (list_power ((fst pr)::l1) ((snd pr)::l2))), 
  pair_list_to_lfun (pr::pl) pf = 
  f_list_push'' pr (pair_list_to_lfun pl (in_list_power_cons' _ _ _ pf)).
intros T1 T2 l1 l2 pl pr h1.
apply functional_extensionality.
intro x.
pose proof (pair_list_to_lfun_in _ h1 x) as h2.
unfold f_list_push''.
destruct (or_to_sumbool (fst pr = proj1_sig x)  
    ((fix In (a : T1) (l : list T1) {struct l} : Prop :=
        match l with
        | nil => False
        | b :: m => b = a \/ In a m
        end) (proj1_sig x) l1) (proj2_sig x)) as [h3 | h4].
(* h3 *)
  destruct h2 as [h2l | h2r].
  (* h2l *)
  rewrite surjective_pairing in h2l at 1.
  inversion h2l. (*A seeming bug prevents me from naming these hypotheses*)
  apply proj1_sig_injective. simpl. 
  rewrite H1. reflexivity.
  (* h2r *)
  apply proj1_sig_injective. simpl. subst.

(* h3 *)
destruct h3 as [h3l | h3r].
  (* h3l *)
  destruct h2 as [h2l | h2r].
    (*h2l *)
    pose proof (f_equal (@snd T1 T2) h2l) as h4.
    simpl in h4. rewrite h4. reflexivity.
    (* h2r *)
    apply proj1_sig_injective. simpl. 

rewrite h2l at 2.

rewrite h2l.




destruct (constructive_indefinite_description
       (fun b : T2 =>
        (snd pr = b \/ In b l2) /\
        (pr = (proj1_sig x, b) \/ In (proj1_sig x, b) pl))
       (in_list_power_fun (pr :: pl) h1 (proj1_sig x) (proj2_sig x))) as [b h2].
destruct (constructive_indefinite_description
       (fun b : T2 =>
        (snd pr = b \/ In b l2) /\
        (pr = (proj1_sig x, b) \/ In (proj1_sig x, b) pl))
       (in_list_power_fun (pr :: pl) h1 (proj1_sig x) (proj2_sig x))) as [c h3].
destruct 
     (constructive_indefinite_description
       (fun b0 : T2 =>
        (snd pr = b0 \/ In b0 l2) /\
        (pr = (proj1_sig x, b0) \/ In (proj1_sig x, b0) pl))
       (in_list_power_fun (pr :: pl) h1 (proj1_sig x) (proj2_sig x))).




Lemma pair_list_to_lfun_compat : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl : list (T1*T2)) (pf:In pl (list_power l1 l2)), 
  proj1_sig (lfun_to_pair_list (pair_list_to_lfun pl pf)) = pl.
intros T1 T2.
induction l1 as [|a l1 h1]; induction pl as [|ap pl h2]. 
(*nil nil*)
intro h2.
assert (h3:pair_list_to_lfun nil h2 = 
  fun x1:{t1:T1 | In t1 nil} => (False_rect {t2:T2 | In t2 l2} (proj2_sig x1))).
  apply functional_extensionality. intro x. pose proof (proj2_sig x) as h3.
  simpl in h3. contradiction.
rewrite h3.
pose proof (proj2_sig (lfun_to_pair_list 
  (fun x1 : {t1 : T1 | In t1 nil} =>
      False_rect {t2 : T2 | In t2 l2} (proj2_sig x1)))) as h4.
simpl in h4.
destruct h4 as [h4l h4r]. simpl.
destruct h4l as [h4la | ?]; try contradiction.
rewrite  h4la. reflexivity.
(*nil cons*)
intro h1. simpl in h1. destruct h1; [discriminate | contradiction].
(*cons nil*)
intro h2. simpl in h2. 
pose proof h2 as h3.
rewrite in_flat_map in h3.
destruct h3 as [pl [h3l h3r]].
rewrite in_map_iff in h3r. destruct h3r as [? []]; discriminate.
(* cons cons *)
intro h3.
destruct l2 as [|c l2].
  (* nil *)
  pose proof h3 as h4. simpl in h4. rewrite in_flat_map in h4.
  destruct h4 as [? []]; contradiction.  
  (* cons *)
  pose proof (in_list_power_cons' _ _ _ h3) as h4.
  specialize (h1 _ _ h4).
  rewrite <- h1 at 2.



(fun x1 : {_ : T1 | False} =>
      False_rect {t2 : T2 | In t2 l2} (proj2_sig x1)))).

 ) as h4.

pose proof (pair_list_to_lfun_in _ h2).


simpl in h2.
pose proof (proj2_sig (lfun_to_pair_list (pair_list_to_lfun nil h2))) as h3.
simpl in h1.
destruct h1 as [h1 h2].
simpl in h2.

*)

Lemma list_power_hd : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:list (T1*T2)) (a a':T1) (b:T2), 
  In ((a, b)::pl) (list_power (a'::l1) l2) -> a = a'.
intros T1 T2 l1 l2 pl a a' b h1.
simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [pl' [h2l h2r]].
rewrite in_map_iff in h2r.
destruct h2r as [b' [h3l h3r]].
rewrite cons_eq in h3l.
destruct h3l as [h3la h3lb].
inversion h3la; subst.
reflexivity.
Qed.


(*
Lemma list_power_no_dups : forall {T1 T2 : Type} {l1:list T1} {l2:list T2}
  (pl:list (T1*T2)) (a:T1) (b1 b2:T2), In ((a, b1)::pl) (list_power l1 l2) ->
  In (a, b2) pl -> False.
intros T1 T2. 
induction l1 as [|c l1 h1]; induction pl as [|ap pl h2].
(* nil nil *)
intros; contradiction.
(* nil cons *)
intros a b1 b2 h1 h3.
simpl in h1.  destruct h1; [discriminate | contradiction].
(* cons nil *)
intros; contradiction.
(* cons cons *)
intros a b1 b2 h3 h4.
pose proof (in_list_power_cons' _ _ _ h3) as h5.
destruct h4 as [h4l | h4r].
  (* h4l *)
  subst.


simpl in h3.
rewrite in_flat_map in h3.
destruct h3 as [pl' h5].
induction ((map (fun y : T2 => (c, y) :: pl') l2)) as [|pl'' lp h6]. 
destruct h5; contradiction.
destruct h5 as [h5l h5r].
destruct h5r as [h5ra | h5rb].


simpl in h3.


induction l1 as [|c l1 h1]; induction pl as [|ap pl h2].
(* nil nil*)
intros a b1 b2 h1. simpl in h1. destruct h1; [discriminate | contradiction].
(* nil cons *)
intros a b1 b2 h1. simpl in h1. destruct h1; [discriminate | contradiction].
(* cons nil *)
intros; contradiction.






(* nil *)
intros a b1 b2 h1.
simpl in h1.
destruct h1; [discriminate | contradiction].
(* cons *)
(*intros l2 pl a b1 b2 h2 h3.*)
induction pl as  [|ap pl h2].
  (* nil *)
  intros. contradiction.
  (* cons *)
  intros a b1 b2 h3 h4.
  pose proof (in_list_power_cons' _ _ _ h3) as h5.
  destruct h4 as [h4l | h4r]. 
    (* h4l *)
    rewrite surjective_pairing in h4l at 1.
    inversion h4l.
    subst.
    Check surjective_pairing.
    pose proof (surjective_pairing ap) as h6.
    rewrite h6 in h5.
    pose proof (h1 l2 pl (fst ap) (snd ap) b1

    simpl in h3. rewrite in_flat_map in h3.
    destruct h3 as [pl' [h3l h3r]].
    rewrite in_map_iff in h3r. destruct h3r as [x [h3ra h3rb]].
    rewrite cons_eq in h3ra.
    destruct 

 
    

  destruct h4 as [h4l | h4r].
    subst.
 pl a b1 b2 h2 h3.

*)


(*
Lemma list_power_fst_eq_eq : forall {T1 T2 : Type} {l1: list T1} {l2:list T2}
  (pl:list (T1*T2)), In pl (list_power l1 l2) -> forall (a:T1) (b:T2), 
  In (a, b) pl -> forall (b':T2), In (a, b') pl -> b = b'.
intros T1 T2.
(*
Check pair_list_to_lfun.
pose proof (proj2_sig (lfun_to_pair_list ((pair_list_to_lfun pl h1)))) as h2.
simpl in h2.



induction pl as [|ap pl h1]. intros; contradiction.
induction l1 as [|c l1 h2].
  intro h2. simpl in h2. destruct h2; [discriminate | contradiction].
intros h3 a b h4 b' h5.
*)

induction l1 as [|a l1 h1].
(* nil *)
  intros l2 pl h2. simpl in h2. 
  induction pl as [|ap pl h3].
  (* nil *) intros; contradiction.
  (* cons *)
  destruct h2; [discriminate | contradiction].
(* cons *)
  revert l1 h1.
  induction pl as [|ap pl h6].
intros h2 a' b h4 b' h5.
  (* nil *) contradiction.
  (* cons *)
  intros h2 a' b h4 b' h5.
  pose proof (in_list_power_cons' _ _ _ h2) as h7.
  destruct h4 as [h4l | h4r]; destruct h5 as [h5l | h5r].
  (* h4l h5l *)
  subst. inversion h5l; subst. reflexivity. 
  (* h4l h5r *)
  subst. pose proof (list_power_hd _ _ _ _ h2). subst.



  simpl in h2. rewrite in_flat_map in h2.
  destruct h2 as [pl' [h2l h2r]].
  simpl in h2r. rewrite in_map_iff in h2r.
  destruct h2r as [b'' [h2r1 h2r2]].
  rewrite cons_eq in h2r1. destruct h2r1 as [h3 ?]; subst.
  inversion h3; subst.


  pose proof (in_cons
  pose proof (h1 l2

  


rewrite surjective_pairing in h5l.
  simpl in h5l.


  intros l1 h7 h8.
  pose proof (in_list_power_cons' _ _ _ h8) as h9.
    destruct h3 as [h3l | h3r]; destruct h4 as [h4l | h4r].
    (* h3l h4l *) subst; trivial.
    (* h3l h4r *) 
    subst.
    pose proof (h1 (pr1 :: pl))


(* partially completed *)



revert a b1 b2 l1 h2 h3.
induction pl as [|ap pl h4].
(* nil *) contradiction.
(* cons *)
intros a b1 b2 l1 h5 h6.
revert pl h4 h5 h6.
induction l1 as [|c l1 h7].
  (* nil *)
  intros pl h1 h2. simpl in h2. destruct h2; [discriminate | contradiction].
  (* cons *)
  intros pl h8 h9 h10.
  destruct h10 as [h10l | h10r].
    (* h10l *)
    subst.
    
    simpl in h9.
    rewrite in_flat_map in h9.
    destruct h9 as [pl' h10].
    destruct h10 as [h10l h10r].
    


  pose proof (in_list_power_cons' _ _ _ h9) as h11.  
  pose proof (h8 (fst ap) (snd ap) b2 l1


  pose proof (h7 pl h8) as h11.

  pose proof (h7 (ap::pl)) as h11.


  apply h7 with pl. assumption.



destruct h6 as [h6l | h6r].
  (* h3l *)
  destruct l1 as [|c l1 h7].
    (* nil *)
    simpl in h5. destruct h5; [discriminate | contradiction].
    (* cons *)
    pose proof (in_list_power_cons' _ _ _ h5) as h7.   
    pose proof (h4 (f
    pose proof (h4 a b1 b2)



  pose proof (in_eq (a, b2) pl) as h8.
  rewrite <- h3l in h6 at 2.
  

destruct l1 as [|c l1 h6]; destruct h3 as [h3l | h3r].
  (* nil steps *)
  simpl in h5. destruct h5; [discriminate | contradiction].
  simpl in h5. destruct h5; [discriminate | contradiction].
  (* cons steps *)
   h3l.

induction l1 as [|c l1 h1].
(* nil *)
intros pl a b1 b2 h1 h2.
simpl in h1. destruct h1; [discriminate | contradiction].
(* cons *)

pose proof (in_list_power_cons' _ _ _  h2) as h4.
  induction pl as [|ap pl h5].
  (* nil *) contradiction.
  (* cons *)
  destruct h3 as [h3l | h3r].
   (* h3l *)
   pose proof (surjective_pairing ap) as h6.
   rewrite h3l in h6 at 1.
   inversion h6 as [h7].
   

simpl in h2.
rewrite in_flat_map in h2.
destruct h2 as [pl' h2].
destruct h2 as [h2l h2r].
rewrite in_map_iff in h2r.
destruct h2r as [x h2r].

*)

(*
Lemma in_list_power_fun_unique : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl : list (T1*T2)), In pl (list_power l1 l2) -> 
  forall a:T1, In a l1 -> exists ! b:T2, In b l2 /\ In (a, b) pl.
intros T1 T2 l1 l2 pl h1 a h2.
pose proof (in_list_power_fun pl h1 a h2) as h3.
destruct h3 as [b h3].
exists b. red. split; try assumption.
intros b' h4.

*)


(*
Lemma not_nil_in_list_power_cons2 : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (b:T2), In nil (list_power l1 (b::l2)) -> False.
intros T1 T2. 
induction l1 as [|a l1 h1]; induction l2 as [| b l2 h2].
intros b h2. simpl in h2.
*)

(*
Lemma in_list_power_cons'' : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:list (T1*T2)) (pr:T1 * T2) (c:T1) (d:T2), In (pr::pl) (list_power (c::l1) (d::l2)) ->
  In pl (list_power l1 l2).
intros T1 T2 l1 l2. 
induction pl as [|ap pl h1].
(* nil *)
intros pr c d h1.
simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [pl' h1].
destruct h1 as [h1l h1r].
destruct h1r as [h1ra | h1rb].
rewrite cons_eq in h1ra. destruct h1ra; subst.
pose proof (@in_list_power2).
rewrite in_map_iff in h1r.
destruct h1r as [x h2].
destruct h2 as [h2l h2r].
rewrite cons_eq in h2l.
destruct h2l. subst. assumption.
*)







(*
Lemma in_list_power_cons_hd_eq : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:(list (T1*T2)))  (pr:(T1 * T2)) (a:T1) (b:T2), 
  In (pr::pl) (list_power (a::l1) (b::l2)) -> pr = (a, b).
intros T1 T2 l1 l2 pl pr a b h1.
pose proof (in_eq pr pl) as h2.
pose proof (in_list_power1 _ h1 _ h2) as h3.
pose proof (in_list_power2 _ h1 _ h2) as h4.


simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [x h2].
*)

(*
Lemma pair_list_to_lfun_lemma : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl : list (T1*T2)), In pl (list_power l1 l2) -> exists f:(list_fun l1 l2), 
  pl = proj1_sig (lfun_to_pair_list f).
intros T1 T2.  
  induction l1 as [|a l1 h1]; induction l2 as [|b l2 ih2].
  (* nil nil *)
    induction pl as [|ap pl h2].
    (* nil *)
    intros h3. (*simpl in h3. *)
    pose (fun x:{t1 : T1 | In t1 nil} => 
      (False_rect {t2:T2 | In t2 nil} (proj2_sig x))) as f.
    exists f.
    pose proof (proj2_sig (lfun_to_pair_list f)) as h4.
    destruct h4 as [h5 h6].
    simpl in h5. destruct h5; [assumption | contradiction].
    (* cons *)
    intro h3.
    simpl in h3. destruct h3; [discriminate | contradiction].
  (* nil cons *)
    induction pl as [|ap pl h2].
    (* nil *)
    intro h2. 
    pose (fun x:{t1 : T1 | In t1 nil} => 
      (False_rect {t2:T2 | In t2 (b::l2)} (proj2_sig x))) as f.
    exists f.
    pose proof (proj2_sig (lfun_to_pair_list f)) as h4.
    destruct h4 as [h5 h6].
    simpl in h5. destruct h5; [assumption | contradiction].
    (* cons *)
    intro h3. simpl in h3. destruct h3; [discriminate | contradiction].
  (* cons nil *)
    intros pl h2.
    simpl in h2. rewrite in_flat_map in h2. destruct h2 as [? h3].
    destruct h3; contradiction.
  (* cons cons *)
    induction pl as [|ap pl h3].
    (* nil *)
    intro h2. simpl in h2. rewrite in_flat_map in h2.
    destruct h2 as [pl h3]. destruct h3 as [? h4].
    simpl in h4. rewrite in_map_iff in h4.
    destruct h4 as [|h4]; try discriminate.
    destruct h4 as [? []]; try discriminate.
    (* cons *)
    intro h2. 
    pose proof (in_list_power_cons' _ _ _ h2) as h4.
    specialize (h1 _ _ h4).
    destruct h1 as [f h5].
    red in f.
    pose 
      (@f_list_push T1 {x:T2 | In x (b::l2)} l1 (a, (exist _ b (in_eq b l2))) f)
      as f'.
(*    simpl in f'.*)
(*    pose (fun t:{x:T1 | In x (a::l1)} => exist _ (f' *)
    exists f'.
    rewrite h5.
    pose proof (proj2_sig (lfun_to_pair_list f')) as h6.
    simpl in h6.
    destruct h6 as [h6 h7].

    unfold f'.
    pose proof (proj2_sig 
(lfun_to_pair_list
     (f_list_push (a, exist (fun x : T2 => b = x \/ In x l2) b (in_eq b l2))
        f))).

    unfold lfun_to_pair_list.


    fold list_fun in f'.
    pose proof (proj2_sig (lfun_to_pair_list f')).

    Check lfun_to_pair_list.





    Print 
(*simpl in h2. rewrite in_flat_map in h2.
    destruct h2 as [pl h3].
    destruct h3 as [h3l h3r].
    rewrite in_map_iff in h3r. destruct h3r as [x h5].
    destruct h5; discriminate.
    intro  h3.*)
    pose proof (in_list_power_cons' _ _ _ h2) as h4.
    specialize (h1 _ _ h4).
    destruct h1 as [f h5].
    Print In.
    pose proof (in_eq ap pl) as h6.
    Check in_list_power2.
    pose (fun x:{t1:In t1 (a::l1)} => 
      match (proj2_sig x) with 
      | (a = t1) => exist _ (snd ap) (in_list_power2 _  h3 _ h6)
      | _ => exist _ (f 


    destruct h5 as [h6 h7].
    simpl in h6. destruct h6 as [h6l | h6r].
      (* h6l *)
      subst.
    apply proj2_sig

l2 pl h1.






(*partially completed *)

Lemma list_power_functional_cons : forall {T1 T2:Type} {l1:list T1} {l2:list T2}
  (pl:(list (T1*T2))) (a:T1) (b1 b2:T2), In ((a, b1)::pl) (list_power l1 l2) -> 
  In (a, b2) ((a, b1)::pl) -> b1 = b2.
intros T1 T2 l1 l2 pl a b1 b2 h1 h2.
induction l1 as [|c l1 h3].
(* nil *)
(*intros pl a b1 b2 h1 h2.*)
simpl in h1. destruct h1; [discriminate | contradiction].
(* cons *)
(*intros pl a b1 b2 h1 h2.*)
pose proof (in_list_power_cons' _ _ _ h1) as h4.
induction pl as [|ap pl h5].
  (* nil *)
  simpl in h2. destruct h2 as [h2 |]; try contradiction.
  inversion h2. subst. reflexivity.
  destruct h2 as [h2l | h2r].
    (* h2l *)
    inversion h2l. subst. reflexivity.
    (* h2r *)
    pose proof (in_cons (a, b1) _ _ h2r) as h6.





    pose proof (surjective_pairing ap) as h6.
    rewrite h6 in h4.
    pose proof (h3 pl (fst ap) (snd ap) b2 h4 h2r).
  pose proof (h3 (ap :: pl)
  pose proof (in_list_power_cons' _ _ _ h1) as h6.




simpl in h1.
rewrite in_flat_map in h1.
destruct h1 as [pl'' h1].
destruct h1 as [h1l h1r].
rewrite in_map_iff in h1r.
destruct h1r as [x h5].
destruct h5 as [h5l h5r].
rewrite cons_eq in h5l.
destruct h5l as [h5l1 h5l2].
subst.
inversion h5l1. subst.
destruct h2 as [h2l | h2r].  




induction pl as [|ap pl' h4].
  (* nil *)
  simpl in h2. destruct h2 as [h4 | ?]; try contradiction.
  inversion h4; subst. reflexivity.
  (* cons *)
  simpl in h1. rewrite in_flat_map in h1.
  destruct h1 as [pl'' h1].
  destruct h1 as [h1l h1r].
  rewrite in_map_iff in h1r.
  destruct h1r as [x h5].
  destruct h5 as [h5l h5r].
  rewrite cons_eq in h5l.
  destruct h5l as [h5l1 h5l2].
  subst.
  inversion h5l1. subst.
  destruct h2 as [h2l | h2r].  
    (* h2l *)
    inversion h2l. subst. reflexivity.
    (* h2r *)
    destruct h2r as [h2r1 | h2r2].
      (* h2r1 *)
      pose proof (surjective_pairing ap) as h6.
      rewrite h2r1 in h6 at 1. inversion h6.
      subst. inversion h6.
      pose proof (h3 pl' (fst ap) (snd ap) b2 h1l ).



    rewrite h6 in h1l.
(*    pose proof (h3 pl' (fst ap) (snd ap) b2 h1l h6).*)
    destruct h2r as [h2r1 | h2r2].
      (* h2r1 *)
      subst.  simpl in h1l. 
      pose proof (in_eq (a, b2) pl') as h7.
      pose proof (h3 pl'

inversion h6. subst.
      pose proof (in_eq ap pl') as h7.
      subst.
      (*rewrite h2r1 in h7. *)
      pose proof (h3 pl' (fst ap) (snd ap) b2 h1l h7).


(* partially completed *)


Lemma forall_in_list_power : forall {T1 T2:Type} {l1:list T1} {l2: list T2} 
  (pl1 pl2 : list (T1 * T2)), In pl1 (list_power l1 l2) -> In pl2 (list_power l1 l2)
  -> ((forall pr:T1 * T2, In pr pl1 -> In pr pl2) <-> (pl1 = pl2)).
intros T1 T2 l1 l2 pl1 pl2 h1 h2.
split.
(* -> *)
revert pl1 pl2 h1 h2.
induction l1 as [| a l1 ihl1].
(* nil l1 *)
intros pl1 pl2 h3 h4.
destruct pl1; destruct pl2.
  (* nil steps *)
  auto.
  simpl in h4. destruct h4; [discriminate | contradiction].
  simpl in h3. destruct h3; [discriminate | contradiction].
  simpl in h3. destruct h3; [discriminate | contradiction].
(* cons l1 *)
intros pl1 pl2 h1' h2' h3.
pose proof h1' as h1. pose proof h2' as h2.
simpl in h1. simpl in h2.
rewrite in_flat_map in h1. rewrite in_flat_map in h2.
destruct h1 as [pl1' h1]. destruct h2 as [pl2' h2].
destruct h1 as [h1l h1r]. destruct h2 as [h2l h2r].
rewrite in_map_iff in h1r. rewrite in_map_iff in h2r. 
destruct h1r as [x1 h1r]. destruct h2r as [x2 h2r].
destruct h1r as [h1ra h1rb]. destruct h2r as [h2ra h2rb]. 
subst.
rewrite cons_eq.
split.
specialize (ihl1 _ _ h1l h2l).
assert (h4: forall pr : T1 * T2, In pr pl1' -> In pr pl2').
  intros pr h5.
  Check in_cons.
  pose proof (in_cons (a, x1) _ _ h5) as h6.
  pose proof (h3 _ h6) as h7.
  simpl in h7.
  destruct h7 as [h7l | h7r].
    (* h3l *)
    subst.


  

rewrite cons_eq in h1ra. rewrite cons_eq in h2ra.
destruct h1ra as [h1ra1 h1ra2]. destruct h2ra as [h2ra1 h2ra2].
subst.




 
  induction pl1 as [| ap pl1 h3]; destruct pl2 as [| bp pl2 h4]; auto.
(* nil steps *)
simpl in h2. destruct h2; [discriminate | contradiction].
simpl in h1. destruct h1; [discriminate | contradiction].
simpl in h1. destruct h1; [discriminate | contradiction].
simpl in h1. rewrite in_flat_map in h1. destruct h1 as [pr h3].
destruct h3 as [h3l h3r]. rewrite in_map_iff in h3r.
destruct h3r as [x h6]. destruct h6; discriminate.
simpl in h2. rewrite in_flat_map in h2. destruct h2 as [pr h6].
destruct h6 as [h6l h6r]. rewrite in_map_iff in h6r.
destruct h6r as [x h7]. destruct h7; discriminate.


(*simpl in h1. destruct h1; [discriminate | contradiction].
simpl in h2. rewrite in_flat_map in h2. destruct h2 as [pr h6].
destruct h6 as [h6l h6r]. rewrite in_map_iff in h6r.
destruct h6r as [x h7]. destruct h7; discriminate.
simpl in h1. destruct h1; [discriminate | contradiction].*)
(* cons steps *)
intro h6.
simpl in h1. simpl in h2.
rewrite in_flat_map in h1. rewrite in_flat_map in h2.
destruct h1 as [pl1' h1]. destruct h2 as [pl2' h2].
destruct h1 as [h1l h1r]. destruct h2 as [h2l h2r].
rewrite in_map_iff in h1r. rewrite in_map_iff in h2r. 
destruct h1r as [x1 h1r]. destruct h2r as [x2 h2r].
destruct h1r as [h1ra h1rb]. destruct h2r as [h2ra h2rb]. 
rewrite cons_eq in h1ra. rewrite cons_eq in h2ra.
destruct h1ra as [h1ra1 h1ra2]. destruct h2ra as [h2ra1 h2ra2].
subst.
specialize

auto.
intro h3. contradiction.

intros pr h3 h4.




Lemma pair_list_cons : forall {T1 T2:Type} {b:T1} {l1 : list T1} {l2 : list T2} 
  (f:(list_fun (b::l1) l2)), (proj1_sig (lfun_to_pair_list f)) = 
  (b, (proj1_sig (f (exist (fun x:T1 => In x (b::l1)) b (in_eq b l1))))) ::
  (proj1_sig (lfun_to_pair_list (f_list_pop f))).
intros T1 T2 b l1 l2 f.

(*
induction l1 as [|a l1 ih].
(* nil *)
pose proof (proj2_sig (lfun_to_pair_list (f_list_pop f))) as h1.
simpl in h1.
destruct h1 as [h1 h2].
destruct h1 as [h1 | ?]; try contradiction.
simpl.
rewrite <- h1.
pose proof (proj2_sig (lfun_to_pair_list f)) as h3.
(*pose proof h3 as h3'.*)
(*simpl in h3.*)
destruct h3 as [h3 h4].
Check in_map_iff.
pose proof h3 as h5.
simpl in h5.
rewrite app_nil_r in h5.
rewrite in_map_iff in h5.
destruct h5 as [x h6].
destruct h6 as [h6l h6r].
pose proof (in_eq (b, x) nil) as h7.
rewrite h6l in h7.
specialize (h4 (b,x) h7).
simpl in h4.
rewrite <- h6l.
rewrite h4.
rewrite cons_eq.
split.
assert (h5: exist (fun x0 : T1 => b = x0 \/ False) b
        (in_list_power1 (proj1_sig (lfun_to_pair_list f)) h3 (b, x) h7) = 
  exist (fun x0 : T1 => b = x0 \/ False) b (in_eq b nil)).
  apply existTexist.
  apply subsetT_eq_compat. reflexivity.
simpl in h5.
rewrite h5.
reflexivity. reflexivity.
(* cons *)
*)
pose proof (proj2_sig (lfun_to_pair_list f)) as h1.
(*simpl in h1.*)
destruct h1 as [h1 h2].
pose proof (in_power_list_cons _ h1) as h3.
destruct h3 as [pl' h4].
destruct h4 as [c h5].
destruct h5 as [h5a [h5b h5c]].
pose proof (in_eq (b, c) pl') as h6.
rewrite <- h5c in h6.
pose proof (h2 (b, c) h6) as h2'.
simpl in h2'.
rewrite h2' in h5c.
rewrite h5c.
rewrite cons_eq.
split.
assert (h7:exist (fun x : T1 => b = x \/ In x l1) b
        (in_list_power1 (proj1_sig (lfun_to_pair_list f)) h1 (b, c) h6) = 
  exist (fun x : T1 => In x (b :: l1)) b (in_eq b l1)).
  apply existTexist.
  apply subsetT_eq_compat; reflexivity.
simpl in h7.
rewrite h7.
reflexivity.
pose proof (proj2_sig (lfun_to_pair_list (f_list_pop f))) as h7.
simpl in h7.
destruct h7 as [h7 h8].


assert (h5:(exists t2:T2, 
  ((fun y:T2 => (b, y) :: nil) t2) = (proj1_sig (lfun_to_pair_list f))
  /\ In t2 (l2 ++ nil))).
  rewrite <- in_map_iff.
  assumption.
simpl in h3.

(*assert (h5:l2 ++ nil = l2). auto with datatypes.*)
rewrite h5 in h3.

assert
pose proof in_map_iff.
*)(*
pose proof (proj2_sig (lfun_to_pair_list f)) as h1.
simpl in h1.
destruct h1 as [h2 h3].
pose proof (in_power_list_cons _ h2) as h4.
destruct h4 as [pl' h5].
destruct h5 as [c h6].
destruct h6 as [h6a [h6b h6c]].



Check in_flat_map.
pose proof h2 as h4.
rewrite in_flat_map in h4.
destruct h4 as [pl h4].
destruct h4 as [h4l h4r].
rewrite in_map_iff in h4r.
destruct h4r as [c h5].
destruct h5 as [h5l h5r].
pose proof (in_eq (b, c) pl) as h6.
rewrite h5l in h6.
pose proof (h3 (b, c) h6) as h7.
simpl in h7.
rewrite h7 in h5l.
Check (proj1_sig (lfun_to_pair_list f)).
rewrite <- h5l.
assert (h8:pl = proj1_sig (lfun_to_pair_list (f_list_pop f))).
  induction pl.
  pose proof (proj2_sig (lfun_to_pair_list (f_list_pop f))) as h8.
  simpl in h8.
(*  admit.*)

  pose proof (@in_list_power1).


unfold f_list_pop. simpl.


simpl.
simpl in h5l.



rewrite <- h5l.
simpl.

Check cons_inj.

pose proof (proj2_sig (lfun_to_pair_list f)) as h5.
destruct h5 as [h5l h5r].
clear h5l.

unfold lfun_to_pair_list in h3l.

Print eq.




apply eq_refl.


simpl.



(pl : list (T1 * T2))*)


