(* Copyright (C) 2014, Daniel Wyckoff, except for the portions so labeleled
which I got from Daniel Schepler*)



(*		   GNU LESSER GENERAL PUBLIC LICENSE
                       Version 3, 29 June 2007

 Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.


  This version of the GNU Lesser General Public License incorporates
the terms and conditions of version 3 of the GNU General Public
License, supplemented by the additional permissions listed below.

  0. Additional Definitions. 

  As used herein, "this License" refers to version 3 of the GNU Lesser
General Public License, and the "GNU GPL" refers to version 3 of the GNU
General Public License.

  "The Library" refers to a covered work governed by this License,
other than an Application or a Combined Work as defined below.

  An "Application" is any work that makes use of an interface provided
by the Library, but which is not otherwise based on the Library.
Defining a subclass of a class defined by the Library is deemed a mode
of using an interface provided by the Library.

  A "Combined Work" is a work produced by combining or linking an
Application with the Library.  The particular version of the Library
with which the Combined Work was made is also called the "Linked
Version".

  The "Minimal Corresponding Source" for a Combined Work means the
Corresponding Source for the Combined Work, excluding any source code
for portions of the Combined Work that, considered in isolation, are
based on the Application, and not on the Linked Version.

  The "Corresponding Application Code" for a Combined Work means the
object code and/or source code for the Application, including any data
and utility programs needed for reproducing the Combined Work from the
Application, but excluding the System Libraries of the Combined Work.

  1. Exception to Section 3 of the GNU GPL.

  You may convey a covered work under sections 3 and 4 of this License
without being bound by section 3 of the GNU GPL.

  2. Conveying Modified Versions.

  If you modify a copy of the Library, and, in your modifications, a
facility refers to a function or data to be supplied by an Application
that uses the facility (other than as an argument passed when the
facility is invoked), then you may convey a copy of the modified
version:

   a) under this License, provided that you make a good faith effort to
   ensure that, in the event an Application does not supply the
   function or data, the facility still operates, and performs
   whatever part of its purpose remains meaningful, or

   b) under the GNU GPL, with none of the additional permissions of
   this License applicable to that copy.

  3. Object Code Incorporating Material from Library Header Files.

  The object code form of an Application may incorporate material from
a header file that is part of the Library.  You may convey such object
code under terms of your choice, provided that, if the incorporated
material is not limited to numerical parameters, data structure
layouts and accessors, or small macros, inline functions and templates
(ten or fewer lines in length), you do both of the following:

   a) Give prominent notice with each copy of the object code that the
   Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the object code with a copy of the GNU GPL and this license
   document.

  4. Combined Works.

  You may convey a Combined Work under terms of your choice that,
taken together, effectively do not restrict modification of the
portions of the Library contained in the Combined Work and reverse
engineering for debugging such modifications, if you also do each of
the following:

   a) Give prominent notice with each copy of the Combined Work that
   the Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the Combined Work with a copy of the GNU GPL and this license
   document.

   c) For a Combined Work that displays copyright notices during
   execution, include the copyright notice for the Library among
   these notices, as well as a reference directing the user to the
   copies of the GNU GPL and this license document.

   d) Do one of the following:

       0) Convey the Minimal Corresponding Source under the terms of this
       License, and the Corresponding Application Code in a form
       suitable for, and under terms that permit, the user to
       recombine or relink the Application with a modified version of
       the Linked Version to produce a modified Combined Work, in the
       manner specified by section 6 of the GNU GPL for conveying
       Corresponding Source.

       1) Use a suitable shared library mechanism for linking with the
       Library.  A suitable mechanism is one that (a) uses at run time
       a copy of the Library already present on the user's computer
       system, and (b) will operate properly with a modified version
       of the Library that is interface-compatible with the Linked
       Version. 

   e) Provide Installation Information, but only if you would otherwise
   be required to provide such information under section 6 of the
   GNU GPL, and only to the extent that such information is
   necessary to install and execute a modified version of the
   Combined Work produced by recombining or relinking the
   Application with a modified version of the Linked Version. (If
   you use option 4d0, the Installation Information must accompany
   the Minimal Corresponding Source and Corresponding Application
   Code. If you use option 4d1, you must provide the Installation
   Information in the manner specified by section 6 of the GNU GPL
   for conveying Corresponding Source.)

  5. Combined Libraries.

  You may place library facilities that are a work based on the
Library side by side in a single library together with other library
facilities that are not Applications and are not covered by this
License, and convey such a combined library under terms of your
choice, if you do both of the following:

   a) Accompany the combined library with a copy of the same work based
   on the Library, uncombined with any other library facilities,
   conveyed under the terms of this License.

   b) Give prominent notice with the combined library that part of it
   is a work based on the Library, and explaining where to find the
   accompanying uncombined form of the same work.

  6. Revised Versions of the GNU Lesser General Public License.

  The Free Software Foundation may publish revised and/or new versions
of the GNU Lesser General Public License from time to time. Such new
versions will be similar in spirit to the present version, but may
differ in detail to address new problems or concerns.

  Each version is given a distinguishing version number. If the
Library as you received it specifies that a certain numbered version
of the GNU Lesser General Public License "or any later version"
applies to it, you have the option of following the terms and
conditions either of that published version or of any later version
published by the Free Software Foundation. If the Library as you
received it does not specify a version number of the GNU Lesser
General Public License, you may choose any version of the GNU Lesser
General Public License ever published by the Free Software Foundation.

  If the Library as you received it specifies that a proxy can decide
whether future versions of the GNU Lesser General Public License shall
apply, that proxy's public statement of acceptance of any version is
permanent authorization for you to choose that version for the
Library.*)



Require Export EnsemblesImplicit.
Require Export ImageImplicit.
Require Import Classical.
Require Import Image.
Require Import LogicUtilities.
Require Import Setoid.
Require Import FunctionProperties.

Lemma inclusion_intersection : forall {T:Type} (P S R : Ensemble T), 
  (Included P S /\ Included P R) <-> Included P (Intersection S R).

intros T P S R.
split.
(*left*)
intro h1.
destruct h1 as [h2 h3].
red.
intros x h4.
constructor; auto with sets.
(*right*)
intro h1.
split; red.
  (*left*)
  intros x h5.
  red in h1.
  pose proof (h1 x h5) as h6.
  inversion h6; trivial.
  (*right*)
  intros x h7.
  pose proof (h1 x h7) as h6. 
  inversion h6; trivial.
Qed.

Lemma intersection_preserves_inclusion : forall {T:Type} (P Q R:Ensemble T), 
  Included P Q -> Included (Intersection R P) (Intersection R Q).
intros T P Q R h1.
red.
intros x h2.
destruct h2 as [x h2l h2r].
auto with sets.
Qed.

Lemma inclusion_id_intersection : forall {T:Type} (P Q:Ensemble T), 
  Included P Q <-> Intersection P Q = P.
intros T P Q.
split.
(* -> *)
intro.
apply Extensionality_Ensembles.
red. split; auto with sets.
(* <= *)
intro h1.
red.
intros x h2.
rewrite <- h1 in h2.
destruct h2; assumption.
Qed.

Lemma empty_inclusion : forall {T:Type} (S:Ensemble T), Included (Empty_set _) S.
intros T S.
red. intros; contradiction.
Qed.

Lemma intersection_full_set : forall {T:Type} (S:Ensemble T), 
  Intersection S (Full_set T) = S.
intros T S.
apply Extensionality_Ensembles.
red. split.
  (* <= *) red. intros ? h. destruct h; assumption.
  (* >= *) red. intros x ?. constructor. assumption. constructor.
Qed.

Lemma included2_impl_union_included : forall {T:Type} (A B C:Ensemble T), 
  ((Included A C) /\ (Included B C)) <-> Included (Union A B) C.
intros T A B C.
split.
(* -> *)
unfold Included.
intro h1.
destruct h1 as [h1 h2].
intros x h3.
apply Union_inv in h3.
destruct h3; auto with sets.
(* <- *)
intro h1.
split.
assert (Included A (Union A B)). auto with sets.
auto with sets.
assert (Included B (Union A B)). auto with sets.
auto with sets.
Qed.



Section complement.
Lemma complement_inclusion: forall {Y:Type} (S T:Ensemble Y),
  Included S T -> Included (Ensembles.Complement T) (Ensembles.Complement S).
Proof.
intros.
red; intros.
red; intro.
contradiction H0.
auto with sets.
Qed.

Lemma excl_middle_empty : (forall {T:Type} (S:(Ensemble T)), Intersection 
  S (Ensembles.Complement S) = Empty_set _).
intros T S.
apply Extensionality_Ensembles. red. split.
(*left*)
red.
intros x h1.
inversion h1; contradiction.
(*right*)
red.
intros; contradiction.
Qed.

Lemma excl_middle_full : forall {T:Type} (S:Ensemble T), Union
  S (Ensembles.Complement S) = Full_set _.
intros T S.
apply Extensionality_Ensembles. red. split.
(*left*)
red. intros. constructor. auto with sets.
(*right*)
red. intros x h1.
case (classic (In S x)) as [h2 | h3].
  apply Union_introl; trivial.
  apply Union_intror.
  auto with sets.
Qed.

Lemma complement_full : forall T:Type, Ensembles.Complement (Full_set T) = Empty_set T.
intro T.
apply Extensionality_Ensembles. red. unfold Included. split.
(* left *)
intros x h2.
unfold Ensembles.Complement in h2. red in h2.
elim h2. apply Full_intro.
(* right *)
intros x h3. contradiction.
Qed.

Lemma complement_empty : forall T:Type, Ensembles.Complement (Empty_set T) = Full_set T.
intro T.
pose proof (Complement_Complement _ (Full_set T)) as h1.
rewrite complement_full in h1.
assumption.
Qed.

Lemma complement_meets_non_subset : forall {T:Type} (P Q : Ensemble T), 
  (Inhabited P) ->
  (~Included P Q <-> Inhabited (Intersection (Ensembles.Complement Q) P)).
intros T P Q h1.
split.
(*left*)
intro h2. 
unfold Included in h2.
pose proof (not_all_ex_not _ _ h2) as h4.
elim h4. intros x h5.
assert (h6: (In P x) /\ (~In Q x)).
  tauto.
destruct h6 as [h7 h8].
assert (h9: In (Intersection (Ensembles.Complement Q) P) x).
  constructor; auto with sets. 
apply Inhabited_intro with x; trivial.
(*right*)
intro h2.
unfold Included.
inversion h2 as [x h3].
assert (h4: exists x:T, ~(In P x -> In Q x)).
  exists x.
  assert (h5: (In P x) /\ (~ In Q x)).
    split.
    inversion h3 as [h6 h7]. assumption.
    inversion h3 as [y h6 h7].
    compute in h7. auto.
  tauto.
apply ex_not_not_all; trivial.
Qed.

Lemma included_empty_complement_int : forall {T:Type} (A B:Ensemble T), 
  Included A B <-> Intersection A (Ensembles.Complement B) = Empty_set _.
intros T A B.
split.
(* -> *)
intros h1.
apply Extensionality_Ensembles. red.
  split.
  (* <= *)
  red in h1.
  red. intros x h2.
  destruct h2 as [x h2l h2r].
  specialize (h1 _ h2l).
  assert (h3:In (Intersection B (Ensembles.Complement B)) x).
    auto with sets.
  rewrite excl_middle_empty in h3. assumption.
  (* >= *) auto with sets.
(* <- *)
intro h1.
red. 
intros x h2.
apply NNPP.
intro h3.
assert (h4:In (Intersection A (Ensembles.Complement B)) x).
  constructor; assumption.
rewrite h1 in h4.
contradiction.
Qed.

(*see if I can automate this *)
Lemma setminus_int_complement : forall {T:Type} (A B:Ensemble T), Setminus A B = 
  Intersection A (Ensembles.Complement B).
intros T A B.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros ? h1.
destruct h1 as [h2 h3].
constructor; auto with sets. 
(* >= *)
red. intros x h1.
destruct h1 as [h2 h3].
constructor; auto with sets.
Qed.

Lemma setminus_sub_sup : forall {T:Type} (A B:Ensemble T), Included A B -> 
  Setminus A B = Empty_set _.
intros T A B h1.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h2.
destruct h2 as [h3 h4].
red in h1.
apply h1 in h3.
contradiction.
(* >= *) auto with sets.
Qed.

End complement.



Section family1.


(*Todo: Maybe only use indexed families and not plain families, still not sure
  if/when I should make it more consistent *)
(*Variables Itype T: Type.*)
(*replace calls to Union/Intersection_intro with 
auto with sets in Power_set_Algebra_axioms *)

(*Include infinite versions of distributive and demorgan laws*)

Variable It T:Type.
Definition Family := Ensemble (Ensemble T).
(*Variable F:Family.*)

(*Copied from Zorn's Lemma Coq user contribution *)
Inductive FamilyUnion (F:Family): Ensemble T :=
  | family_union_intro: forall (S:Ensemble T) (x:T),
    In F S -> In S x -> In (FamilyUnion F) x.

Inductive FamilyIntersection (F:Family) : Ensemble T :=
  | family_intersection_intro: forall x:T,
    (forall S:Ensemble T, In F S -> In S x) ->
    In (FamilyIntersection F) x.

Definition IndexedFamily := It -> Ensemble T.
Variable FI:IndexedFamily.

Inductive IndexedUnion : Ensemble T :=
  | indexed_union_intro: forall (i:It) (x:T),
    In (FI i) x -> In IndexedUnion x.

Inductive IndexedIntersection : Ensemble T :=
  | indexed_intersection_intro: forall (x:T),
    (forall i:It, In (FI i) x) -> In IndexedIntersection x.
Definition ImageFamily : Family:=
  Im (Full_set _) FI.

Lemma indexed_to_family_union: IndexedUnion = FamilyUnion (ImageFamily).
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
destruct H.
apply family_union_intro with (FI i).
apply Im_intro with i.
constructor.
reflexivity.
assumption.

destruct H.
destruct H.
apply indexed_union_intro with x0.
rewrite <- H1.
assumption.
Qed.                                                        

Lemma indexed_to_family_intersection:
  IndexedIntersection = FamilyIntersection ImageFamily.
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
constructor.
intros.
destruct H.
destruct H0.
rewrite H1.
apply H.

constructor.
intro.
destruct H.
apply H.
apply Im_intro with i.
constructor.
reflexivity.
Qed.
End family1.

Lemma family_union_add : forall {T:Type} (F:Family T) (S:Ensemble T),
  FamilyUnion _ (Add F S) = Union S (FamilyUnion _ F).
intros T F S.
apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h4.
  unfold Add in h4.
  inversion h4 as [A y h5 h6].
  pose proof (Union_inv _ _ _ A h5) as h7.
  case h7 as [h8 | h9].
  (*h8*)
  apply Union_intror. 
  apply family_union_intro with A; assumption.
  (*h9*)
  left. 
  pose proof (Singleton_inv _ _ _ h9) as h12.
  rewrite <- h12 in h6; assumption.
  (* >= *)
  red. intros x h2.
  pose proof (Union_inv _ _ _ x h2) as h3.
  case h3 as [h4 | h5].
  (*h4*)
  pose proof (Add_intro2 _ F S) as h5.
  apply family_union_intro with S; assumption.
  (*h5*)
  inversion h5 as [A y h6 h7]. 
  pose proof (Add_intro1 _ F S) as h8.
  pose proof (h8 A h6) as h9.
  apply family_union_intro with A; assumption.
Qed.

Section family2.
Lemma empty_indexed_intersection: forall {T:Type}
  (F:IndexedFamily False T),
  IndexedIntersection _ _ F = (Full_set _).
Proof.
intros.
apply Extensionality_Ensembles; red; split; red; intros;
  auto with sets.
constructor.
constructor.
destruct i.
Qed.

Variable T:Type.
Lemma empty_family_union: FamilyUnion T (Empty_set _) =
                          (Empty_set _).
Proof.
apply Extensionality_Ensembles.
unfold Same_set.
unfold Included.
intuition.
destruct H.
contradiction H.
contradiction H.
Qed.

Definition chain (F:Family T) : Prop := 
  forall (S1 S2:Ensemble T), In F S1 -> 
    In F S2 -> Included S1 S2 \/ Included S2 S1.

End family2.

Arguments FamilyUnion [T] _ _.
Arguments FamilyIntersection [T] _ _.
Arguments IndexedUnion _ [T] _ _.
Arguments IndexedIntersection _ [T] _ _.

Section Power_set_Algebra_axioms.
Variable Xt:Type.
Lemma assoc_sum_psa : forall N M P:(Ensemble Xt), (Union N (Union M P)) = 
  (Union (Union N M) P).
intros N M P.
apply Extensionality_Ensembles.  unfold Same_set.
split; unfold Included.
(*left*)
intros x H.
inversion H.
assert (h1 : In (Union N M) x). 
apply Union_introl. assumption.
apply Union_introl. assumption.
inversion H0.
assert (h1 : In (Union N M) x). 
apply Union_intror. assumption.
apply Union_introl. assumption.
apply Union_intror. assumption.
(*right*)
intros x H.
inversion H. inversion H0.
apply (Union_introl _ _ _ _ H2).
assert (h2: In (Union M P) x).
apply (Union_introl _ _ _ _ H2).
apply (Union_intror _ _ _ _ h2).
assert (h2: In (Union M P) x).
apply (Union_intror _ _ _ _ H0).
apply (Union_intror _ _ _ _ h2).
Qed.

Lemma assoc_prod_psa : forall N M P:(Ensemble Xt), 
  (Intersection N (Intersection M P)) = 
  (Intersection (Intersection N M) P).
intros N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
inversion h3 as [x'' h5 h6 h7].
assert (h8: In (Intersection N M) x).
apply Intersection_intro; assumption.
apply Intersection_intro; assumption.
(*Learn how to automate the above duplication*)
(*right*)
intros x h9.
inversion h9 as [x' h10 h11 h12].
inversion h10 as [x'' h13 h14 h15].
assert (h16: In (Intersection M P) x).
apply Intersection_intro; assumption.
apply Intersection_intro; assumption.
Qed.

Lemma comm_sum_psa : forall N M:(Ensemble Xt), Union N M = Union M N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set. 
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
apply Union_intror; assumption.
apply Union_introl; assumption.
(*right*)
intros x h6.
inversion h6 as [x' h7 h8 | x'' h9 h10].
apply Union_intror; assumption.
apply Union_introl; assumption.
Qed.

Lemma comm_prod_psa : 
  forall N M:(Ensemble Xt), Intersection N M = Intersection M N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
apply Intersection_intro; assumption.
(*right*)
intros x h5.
inversion h5 as [x' h6 h7 h8].
apply Intersection_intro; assumption.
Qed.

Lemma abs_sum_psa  : 
  forall N M:(Ensemble Xt), Union N (Intersection N M) = N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
assumption.
inversion h4 as [x''' h6 h7].
assumption.
(*right*)
intros x h8.
apply Union_introl. assumption.
Qed.

Lemma abs_prod_psa  : 
  forall N M:(Ensemble Xt), Intersection N (Union N M) = N.
intros N M.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
assumption.
(*right*)
intros x h5.
assert (h6:In (Union N M) x).
apply Union_introl; assumption.
apply Intersection_intro; assumption.
Qed.

Lemma dist_sum_psa :
  forall N M P:(Ensemble Xt), 
    Intersection P (Union N M) = 
    Union (Intersection P N) (Intersection P M).
intros N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 h4].
inversion h3 as [x'' h5 h6 | x'' h7 h8].
assert (h9: In (Intersection P N) x).
apply Intersection_intro; assumption.
apply Union_introl; assumption.
assert (h10: In (Intersection P M) x).
apply Intersection_intro; assumption.
apply Union_intror; assumption.
(*right*)
intros x h11.
inversion h11 as [x' h12 h13 | x'' h14 h15].
inversion h12 as [x''' h16 h17 h18].
assert (h19 : In (Union N M) x).
apply Union_introl; assumption.
apply Intersection_intro; assumption.
inversion h14 as [x''' h20 h21 h22].
assert (h23: In (Union N M) x).
apply Union_intror; assumption.
apply Intersection_intro; assumption.
Qed.


Lemma dist_prod_psa : 
  forall N M P:(Ensemble Xt), 
    Union P (Intersection N M) = 
    Intersection (Union P N) (Union P M).
intros N M P.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
inversion h1 as [x' h2 h3 | x'' h4 h5].
assert (h6: In (Union P N) x). auto with sets.
assert (h7: In (Union P M) x). auto with sets.
auto with sets.
inversion h4.
assert (h8: In (Union P N) x). auto with sets.
assert (h9: In (Union P N) x). auto with sets.
auto with sets.
(*right*)
intros x h10.
inversion h10 as [x' h11 h12 h13].
inversion h11. 
auto with sets.
inversion h12.
auto with sets.
assert (h14: In (Intersection N M) x). 
auto with sets. auto with sets.
Qed.

(*Uses Law of Excluded Middle*)
Lemma comp_sum_psa: 
  forall N:(Ensemble Xt), Union N (Ensembles.Complement N) = Full_set Xt.
intros N.
apply Extensionality_Ensembles.
unfold Same_set. 
split; unfold Included.
(*left*)
intros x h1.
apply Full_intro.
(*right*)
intros x h2.
assert (h3: ~(In N x) \/ (In N x)). 
    tauto.
elim h3; auto with sets.
Qed.

Lemma comp_prod_psa:
  forall N:(Ensemble Xt), Intersection N (Ensembles.Complement N) = Empty_set Xt.
intros N.
apply Extensionality_Ensembles.
unfold Same_set.
split; unfold Included.
(*left*)
intros x h1.
contradict h1.
intro h2.
inversion h2 as [x' h3 h4].
unfold In in h3.  unfold In in h4. unfold Ensembles.Complement in h4.
tauto.
(*right*)
intros x h5.
inversion h5.
Qed.
End Power_set_Algebra_axioms.

Section DeMorgan.
Variable T:Type.

Lemma comp_union : forall (M N : Ensemble T), Ensembles.Complement (Union M N) =
  Intersection (Ensembles.Complement M) (Ensembles.Complement N).
intros M N.
apply Extensionality_Ensembles. red. split.
(*left*)
red. intros x h1.
assert (h2: Included M (Union M N)).
  auto with sets.
assert (h3: Included N (Union M N)).
  auto with sets.
split.
pose proof (complement_inclusion _ _ h2) as h4.
auto with sets.
pose proof (complement_inclusion _ _ h3) as h5.
auto with sets.
(*right*)
red. intros x h6.
inversion h6 as [y h7 h8].
unfold Ensembles.Complement in h7. red in h7.
unfold Ensembles.Complement in h8. red in h8.
unfold Ensembles.Complement. red.
intro h9.
inversion h9.
contradiction. contradiction.
Qed.

Lemma comp_int : forall (M N : Ensemble T), Ensembles.Complement (Intersection M N) =
  Union (Ensembles.Complement M) (Ensembles.Complement N).
intros M N.
pose proof (Complement_Complement _ M) as h1.
pose proof (Complement_Complement _ N) as h2.
rewrite <- h1 at 1.
rewrite <- h2 at 1.
pose proof (comp_union (Ensembles.Complement M) (Ensembles.Complement N)) as h3.
rewrite <- h3.
apply Complement_Complement.
Qed.
End DeMorgan.

Section Infinite_analogues.
Variable It T:Type.
Variable FI : (IndexedFamily It T).

Definition Int_Fam_Ens (P:Ensemble T) (Q:Family T) : Family T := 
  (Im Q (fun (S:Ensemble T) => (Intersection P S))).
Definition Union_Fam_Ens (P:Ensemble T) (Q:Family T) : Family T :=
  (Im Q (fun (S:Ensemble T) => (Union P S))).

Definition Int_Ind_Ens (P:Ensemble T) (i:It) : (Ensemble T) :=
  Intersection P (FI i).

Definition Union_Ind_Ens (P:Ensemble T) (i:It) : (Ensemble T) :=
  Intersection P (FI i).
  
 
Lemma dist_union_infnt_1 : 
  forall (P:Ensemble T) (Q:Family T) , 
    Intersection P (FamilyUnion Q) = FamilyUnion (Int_Fam_Ens P Q).
intros P Q.
apply Extensionality_Ensembles.
unfold Same_set.
split.
(*left*)
unfold Included.
intros x h1.
inversion h1 as [y h2 h3 h4].
inversion h3 as [S z h5 h6 h7].
assert (h8: In (Intersection P S) x).
  auto with sets.
apply family_union_intro with (S := (Intersection P S)).
unfold Int_Fam_Ens.
apply Im_intro with (x := S).
assumption. reflexivity. assumption.
(*right*)
unfold Included.
intros x h8.
inversion h8 as [S y h9 h10 h11].
unfold Int_Fam_Ens in h9.
apply Intersection_intro.
(*inversion h9.*)
inversion h9 as [A h12 B h13 h14].
rewrite h13 in h10.
inversion h10.
assumption.
inversion h9 as [A h12 B h13 h14].
rewrite h13 in h10.
assert (h15: In A x).
  inversion h10. assumption.
apply family_union_intro with (S := A).
assumption. assumption.
Qed.

Lemma dist_int_infnt_1 : 
  forall (P:Ensemble T) (Q:Family T), 
    Union P (FamilyIntersection Q) = FamilyIntersection (Union_Fam_Ens P Q).
intros P Q.
apply Extensionality_Ensembles.
unfold Same_set.
split.
(*left*)
unfold Included.
intros x h1.
inversion h1 as [y h2 | y h3].
apply family_intersection_intro.
intros S h4.
unfold Union_Fam_Ens in h4.
inversion h4 as [S' z h5 h6 h7].
assert (h8: In (Union P S') x). auto with sets.
rewrite h6. assumption.
inversion h3 as [z h9 h10].
apply family_intersection_intro.
intros S h11.
unfold Union_Fam_Ens in h11.
inversion h11 as [A h12 B h13 h14].
assert (h15: In A x).
  apply h9.  assumption.
assert (h16: In (Union P A) x).  auto with sets.
rewrite h13. assumption.
(*right*)
unfold Included.
intros x h17.
inversion h17 as [y h18 h19].
unfold Union_Fam_Ens in h18.
assert (h20: forall (S':Ensemble T), (In Q S') -> In (Union P S') x).
  intros S' h50.
  apply h18.
  apply Im_intro with (x := S'). 
  assumption. reflexivity.
assert (h21 : 
    forall S : Ensemble T, In Q S -> In P x \/ In S x).
  intros S h22.
  apply Union_inv.
  apply h20. assumption.
assert (h22: In P x \/ (forall S : Ensemble T, (In Q S -> In S x))).
  apply prop_dis_forall.
  intro S.
  assert (h23: (In Q S -> In P x \/ In S x)).
    apply h21.
  tauto.
case h22.
  auto with sets.
intro h24.
apply Union_intror.
apply family_intersection_intro.
assumption.
Qed.

Lemma dist_union_infnt_1_ind : 
  forall (P:Ensemble T), 
    Intersection P (IndexedUnion _ FI) = IndexedUnion _ (Int_Ind_Ens P).
intro P.
rewrite (indexed_to_family_union _ _ FI).
rewrite indexed_to_family_union.
rewrite (dist_union_infnt_1 P (ImageFamily It T FI)).
apply Extensionality_Ensembles. unfold Same_set.
unfold Int_Fam_Ens.
unfold Int_Ind_Ens.
unfold Included.
split.
(*left*)
intros x h1.
inversion h1.
inversion H.
unfold ImageFamily.
apply family_union_intro with (S := S).
unfold ImageFamily in H2.
rewrite H3.
inversion H2.
apply Im_intro with (X := Full_set It) (x := x2).
assumption.
rewrite H6. reflexivity.
assumption.
(*right*)
intros x h20.
unfold ImageFamily in h20.
inversion h20.
apply family_union_intro with (S := S).
inversion H.
rewrite H3.
apply Im_intro with (x := (FI x1)).
unfold ImageFamily.
apply Im_intro with (x := x1).
assumption.
reflexivity.
reflexivity.
assumption.
Qed.
End Infinite_analogues.

(*Section Characteristic. Commented to keep the notation top-level*)

Inductive characteristic_function_abstraction {X:Type} (P:X->Prop) (x:X) : Prop :=
  | intro_characteristic_sat: P x ->
    In (characteristic_function_abstraction P) x.

Definition characteristic_function_to_ensemble {X:Type} (P:X->Prop) : Ensemble X :=
  characteristic_function_abstraction P.

Notation "[ x : X | P ]" :=
  (characteristic_function_to_ensemble (fun x:X => P))
  (x ident).

Lemma characteristic_function_to_ensemble_is_identity:
  forall {X:Type} (P:X->Prop),
    [ x:X | P x ] = P.
Proof.
intros.
apply Extensionality_Ensembles; split; red; intros.
destruct H.
exact H.
constructor.
exact H.
Qed.

(*End Characteristic.*)


Section Finite_Families.
Lemma Finite_Finite_Union : forall {T:Type} (F:Family T), 
  (forall (S:(Ensemble T)), In F S -> Finite _ S) ->
  Finite _ F -> Finite _ (FamilyUnion F).
intros T F h1 h2.
induction h2.
(*Empty*)
pose proof (empty_family_union T) as h3.
rewrite h3.
constructor.
(*Add*)
rename x into S.
rename A into F.
rename H into h3.
rename IHh2 into h4.
pose proof (family_union_add F S) as h5.
rewrite h5.
apply Union_preserves_Finite.
apply h1.
auto with sets.
assert (h6:forall S:Ensemble T, In F S -> Finite _ S).
  intros S0 h7.
  apply h1.
  auto with sets.
apply (h4 h6).
Qed.

(* Singleton Family decomposition of a set *)
Definition SingF {T:Type} (S:Ensemble T) := Im S (@Singleton T).  

Lemma union_singF_eq : forall {T:Type} (S:Ensemble T), 
  S = FamilyUnion (SingF S).
intros T S.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h1.
assert (h2:In (SingF S) (Singleton x)).
  unfold SingF.
  apply Im_intro with x; trivial.
pose proof (In_singleton _ x) as h3.
apply family_union_intro with (Singleton x); assumption.
(* >= *)
red. intros x h1.

inversion h1 as [S0 ? h2 h3].
unfold SingF in h2.
inversion h2 as [? h4 ? h5].
rewrite h5 in h3.
inversion h3 as [h6].
rewrite <- h6; trivial.
Qed.

Lemma fin_fin_singf : forall {T:Type} (S:Ensemble T), Finite _ S 
  -> Finite _ (SingF S).
intros.
apply finite_image; assumption.
Qed.

End Finite_Families.

Section SurjectiveEq.

Lemma surj_eq : forall {T U:Type} (f:T->U), surjective f <-> Im (Full_set T) f = 
  Full_set U.
intros T U f. 
split.
(* -> *)
intro h1.
apply Extensionality_Ensembles.
red. split.
  (* <= *)
  red.
  intros x ?. constructor.
  (* >= *)
  red.
  intros u ?.
  red in h1.
  pose proof (h1 u) as h2.
  destruct h2 as [t].
  apply Im_intro with t.
  constructor.
  symmetry; assumption.
(* <- *)
intro h1.
red.
intro u.
pose proof (Full_intro _ u) as h2.
rewrite <- h1 in h2.
inversion h2 as [t].
exists t. symmetry. assumption.
Qed.

End SurjectiveEq.

Section RandomFacts.
Lemma full_sig : forall {T:Type} (S:Ensemble T), 
  S = Im (Full_set {x:T | In S x}) (@proj1_sig _ _).
intros T S.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x h1.
apply Im_intro with (exist _ x h1).
constructor. simpl. reflexivity.
(* >= *)
red. intros x h1.
inversion h1 as [x' h2 ? h3 h4].
rewrite h3.
apply proj2_sig.
Qed.


(* In the middle for now :
Lemma card_full_sig : forall {T:Type} (A:Ensemble T) (n:nat), 
  cardinal _ A n <-> cardinal _ (Full_set {t:T | In A t}) n.
intros T A n.
split.
intro h1.
Check injective_preserves_cardinal.
induction h1 as [|A n h2 h3 t h4 h5].
(* 0 *)
assert (h2:Full_set {t:T | In (Empty_set T) t} = Empty_set _).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *) red. intros x. inversion x. contradiction.
  (* >= *) red. intros x h1. inversion h1.
rewrite h2. constructor.
(* S n *)
Print cardinal.
pose proof (card_add _ _ _ h2 _ h4) as h5.


Print cardinal.

pose proof (injective_preserves_cardinal _ 

apply injective_preserves_cardinal.

rewrite (full_sig S) in h1.
*)

End RandomFacts.

Arguments FamilyUnion [T] F _.
Arguments FamilyIntersection [T] F _.
Arguments IndexedUnion [It] [T] FI _.
Arguments IndexedIntersection [It] [T] FI _.
Arguments assoc_sum_psa [Xt] N M P.
Arguments assoc_prod_psa [Xt] N M P.
Arguments comm_sum_psa [Xt] N M. 
Arguments comm_prod_psa [Xt] N M.
Arguments abs_sum_psa [Xt] N M. 
Arguments abs_prod_psa [Xt] N M.
Arguments dist_sum_psa [Xt] N M P.
Arguments dist_prod_psa [Xt] N M P.
Arguments comp_sum_psa [Xt] N. 
Arguments comp_prod_psa [Xt] N.
(*Arguments Ex_Ensemble [U] P.*)

