(* Copyright (C) 2014, Daniel Wyckoff, except for the portions so labeleled
which I got from Daniel Schepler*)


(*		   GNU LESSER GENERAL PUBLIC LICENSE
                       Version 3, 29 June 2007

 Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.


  This version of the GNU Lesser General Public License incorporates
the terms and conditions of version 3 of the GNU General Public
License, supplemented by the additional permissions listed below.

  0. Additional Definitions. 

  As used herein, "this License" refers to version 3 of the GNU Lesser
General Public License, and the "GNU GPL" refers to version 3 of the GNU
General Public License.

  "The Library" refers to a covered work governed by this License,
other than an Application or a Combined Work as defined below.

  An "Application" is any work that makes use of an interface provided
by the Library, but which is not otherwise based on the Library.
Defining a subclass of a class defined by the Library is deemed a mode
of using an interface provided by the Library.

  A "Combined Work" is a work produced by combining or linking an
Application with the Library.  The particular version of the Library
with which the Combined Work was made is also called the "Linked
Version".

  The "Minimal Corresponding Source" for a Combined Work means the
Corresponding Source for the Combined Work, excluding any source code
for portions of the Combined Work that, considered in isolation, are
based on the Application, and not on the Linked Version.

  The "Corresponding Application Code" for a Combined Work means the
object code and/or source code for the Application, including any data
and utility programs needed for reproducing the Combined Work from the
Application, but excluding the System Libraries of the Combined Work.

  1. Exception to Section 3 of the GNU GPL.

  You may convey a covered work under sections 3 and 4 of this License
without being bound by section 3 of the GNU GPL.

  2. Conveying Modified Versions.

  If you modify a copy of the Library, and, in your modifications, a
facility refers to a function or data to be supplied by an Application
that uses the facility (other than as an argument passed when the
facility is invoked), then you may convey a copy of the modified
version:

   a) under this License, provided that you make a good faith effort to
   ensure that, in the event an Application does not supply the
   function or data, the facility still operates, and performs
   whatever part of its purpose remains meaningful, or

   b) under the GNU GPL, with none of the additional permissions of
   this License applicable to that copy.

  3. Object Code Incorporating Material from Library Header Files.

  The object code form of an Application may incorporate material from
a header file that is part of the Library.  You may convey such object
code under terms of your choice, provided that, if the incorporated
material is not limited to numerical parameters, data structure
layouts and accessors, or small macros, inline functions and templates
(ten or fewer lines in length), you do both of the following:

   a) Give prominent notice with each copy of the object code that the
   Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the object code with a copy of the GNU GPL and this license
   document.

  4. Combined Works.

  You may convey a Combined Work under terms of your choice that,
taken together, effectively do not restrict modification of the
portions of the Library contained in the Combined Work and reverse
engineering for debugging such modifications, if you also do each of
the following:

   a) Give prominent notice with each copy of the Combined Work that
   the Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the Combined Work with a copy of the GNU GPL and this license
   document.

   c) For a Combined Work that displays copyright notices during
   execution, include the copyright notice for the Library among
   these notices, as well as a reference directing the user to the
   copies of the GNU GPL and this license document.

   d) Do one of the following:

       0) Convey the Minimal Corresponding Source under the terms of this
       License, and the Corresponding Application Code in a form
       suitable for, and under terms that permit, the user to
       recombine or relink the Application with a modified version of
       the Linked Version to produce a modified Combined Work, in the
       manner specified by section 6 of the GNU GPL for conveying
       Corresponding Source.

       1) Use a suitable shared library mechanism for linking with the
       Library.  A suitable mechanism is one that (a) uses at run time
       a copy of the Library already present on the user's computer
       system, and (b) will operate properly with a modified version
       of the Library that is interface-compatible with the Linked
       Version. 

   e) Provide Installation Information, but only if you would otherwise
   be required to provide such information under section 6 of the
   GNU GPL, and only to the extent that such information is
   necessary to install and execute a modified version of the
   Combined Work produced by recombining or relinking the
   Application with a modified version of the Linked Version. (If
   you use option 4d0, the Installation Information must accompany
   the Minimal Corresponding Source and Corresponding Application
   Code. If you use option 4d1, you must provide the Installation
   Information in the manner specified by section 6 of the GNU GPL
   for conveying Corresponding Source.)

  5. Combined Libraries.

  You may place library facilities that are a work based on the
Library side by side in a single library together with other library
facilities that are not Applications and are not covered by this
License, and convey such a combined library under terms of your
choice, if you do both of the following:

   a) Accompany the combined library with a copy of the same work based
   on the Library, uncombined with any other library facilities,
   conveyed under the terms of this License.

   b) Give prominent notice with the combined library that part of it
   is a work based on the Library, and explaining where to find the
   accompanying uncombined form of the same work.

  6. Revised Versions of the GNU Lesser General Public License.

  The Free Software Foundation may publish revised and/or new versions
of the GNU Lesser General Public License from time to time. Such new
versions will be similar in spirit to the present version, but may
differ in detail to address new problems or concerns.

  Each version is given a distinguishing version number. If the
Library as you received it specifies that a certain numbered version
of the GNU Lesser General Public License "or any later version"
applies to it, you have the option of following the terms and
conditions either of that published version or of any later version
published by the Free Software Foundation. If the Library as you
received it does not specify a version number of the GNU Lesser
General Public License, you may choose any version of the GNU Lesser
General Public License ever published by the Free Software Foundation.

  If the Library as you received it specifies that a proxy can decide
whether future versions of the GNU Lesser General Public License shall
apply, that proxy's public statement of acceptance of any version is
permanent authorization for you to choose that version for the
Library.*)




(*Todo: provide names for automatic variables in inversion/destruct/etc.*)
Require Import SetUtilities.
Require Import LogicUtilities.
Require Export BoolAlgBasics.
Require Import FieldsOfSets.
Require Import InfiniteOperations.
Require Import TypeUtilities.
Require Import Arith.
Require Import FunctionProperties.
Require Import Equality.
Require Import FunctionalExtensionality.
Require Import IndefiniteDescription.

Module SA (Import BA:BAt).
Include Bscs(BA).

Section A_sec.
Variable A:(Ensemble Bt). 

Definition SubBtype := {a:Bt | In A a}.

Definition Bplus_sub (x y:SubBtype) := (proj1_sig x) + (proj1_sig y).
Definition Btimes_sub (x y:SubBtype) := (proj1_sig x) * (proj1_sig y).
Definition Bcomp_sub (x:SubBtype) := -(proj1_sig x).

Notation  "x +|' y" := (Bplus_sub x y) (at level 50, left associativity).
Notation  "x *|' y" := (Btimes_sub x y) (at level 40, left associativity).
Notation "-|' x" := (Bcomp_sub x) (at level 30).

Definition Plus_closed_sub : Prop := (forall (x y:SubBtype), 
               In A (x +|' y)).
Definition Times_closed_sub : Prop := (forall (x y:SubBtype), 
               In A (x *|' y)).
Definition One_closed_sub  : Prop := In A 1.
Definition Zero_closed_sub : Prop := In A 0.
Definition Comp_closed_sub : Prop := (forall x:SubBtype, 
               In A (-|' x)).

(*Redo this part, probably eliminate some of these trivial lemmas *)

Definition Plus_closed'' (x y:Bt) :
  Plus_closed_sub -> In A x -> In A y -> In A (x+y).
intros h1 h2 h3.
unfold Plus_closed_sub in h1.
apply h1 with (x := exist (In A) x h2) (y := exist (In A) y h3).
Qed.

Definition Times_closed'' (x y:Bt) :
  Times_closed_sub -> In A x -> In A y -> In A (x*y).
intros h1 h2 h3.
unfold Times_closed_sub in h1.
apply h1 with (x := exist (In A) x h2) (y := exist (In A) y h3).
Qed.

Definition Comp_closed'' (b:Bt) : Comp_closed_sub -> (In A b) ->
  In A (-b).
intros h1 h2.
unfold Comp_closed_sub in h1.
apply h1 with (x := exist (In A) b h2).
Qed.

Record A_closed := {
  P_c : Plus_closed_sub;
  T_c : Times_closed_sub;
  O_c : One_closed_sub;
  Z_c : Zero_closed_sub;
  C_c : Comp_closed_sub}.

Variable Ac : A_closed.

Definition Bc' := 
  Build_Bconst SubBtype
    (Full_set SubBtype)
    (fun (x y:SubBtype) => exist (In A) (x +|' y)
      (P_c Ac x y))
    (fun (x y:SubBtype) => exist (In A) (x *|' y)
      (T_c Ac x y))
    (exist (In A ) 1 (O_c Ac))
    (exist (In A ) 0 (Z_c Ac))
    (fun (x:SubBtype) => (exist (In A) (-|' x)
      (C_c Ac x))).

Section Bc''.

(*Maybe transfer the closure requirements to a single object in the definition of Bc'*)

Infix "+|" := (Bplus Bc')  (at level 50, left associativity). 
Infix "*|" := (Btimes Bc') (at level 40, left associativity).
Notation "0|" := (Bzero Bc').
Notation "1|" := (Bone Bc').
Notation "-| x" := ((Bcomp Bc') x) (at level 30).


(*These are just all the basic axioms applied to the subset type SubBtype, derived from
  the main boolean algebra B *)
Lemma assoc_sum' : forall n m p:SubBtype, n +| (m +| p) = n +| m +| p.
intros n m p.  
red in n.
red in m.
red in p.
inversion n as [x]. 
inversion m as [y]. 
inversion p as [z].
simpl.
unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply assoc_sum.
Qed.

Lemma assoc_prod' : forall n m p:SubBtype, n *| (m *| p) = n *| m *| p.
intros n m p.
red in n.
red in m.
red in p.
inversion n as [x]. 
inversion m as [y]. 
inversion p as [z].
simpl.
unfold Btimes_sub.
simpl.
apply proj1_sig_injective.
simpl. apply assoc_prod.
Qed.

Lemma comm_sum' : forall n m:SubBtype, n +| m = m +| n.
intros n m.
red in n. red in m.
inversion n. inversion m.
simpl.
unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply comm_sum.
Qed.

Lemma comm_prod' : forall n m:SubBtype, n *| m = m *| n.
intros n m.
red in n. red in m.
inversion n. inversion m.
simpl.
unfold Btimes_sub.
simpl.
apply proj1_sig_injective.
simpl. apply comm_prod.
Qed.

Lemma abs_sum'  : forall n m:SubBtype, n +| (n *| m) = n.
intros n m.
red in n. red in m.
inversion n. inversion m.
simpl.
unfold Btimes_sub. unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply abs_sum.
Qed.

Lemma abs_prod':  forall n m:SubBtype, n *| (n +| m) = n.
intros n m.
red in n. red in m.
inversion n. inversion m.
simpl.
unfold Btimes_sub. unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply abs_prod.
Qed.

Lemma dist_sum' : forall n m p:SubBtype, p *| (n +| m) = p *| n + p *| m.
intros n m p.
red in n. red in m. red in p.
inversion n. inversion m. inversion p.
simpl.
unfold Btimes_sub. unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply dist_sum.
Qed.

Lemma dist_prod': forall n m p:SubBtype, p +| (n *| m) = (p +| n) *| (p +| m).
intros n m p.
red in n. red in m. red in p.
inversion n. inversion m. inversion p.
simpl.
unfold Btimes_sub. unfold Bplus_sub.
simpl.
apply proj1_sig_injective.
simpl. apply dist_prod.
Qed.

Lemma comp_sum':  forall n:SubBtype, n +| (-| n) = 1.
intros n.
red in n. inversion n.
simpl.
unfold Bplus_sub. unfold Bcomp_sub.
simpl.
apply proj1_sig_injective.
simpl. apply comp_sum.
Qed.

Lemma comp_prod': forall n:SubBtype, n *| (-| n) = 0.
intros n.
red in n. inversion n.
simpl.
unfold Bplus_sub. unfold Bcomp_sub.
simpl.
apply proj1_sig_injective.
simpl. apply comp_prod.
Qed.

Lemma und_set' : (BS Bc') = Full_set (Btype Bc').
unfold Bc'.
simpl. reflexivity.
Qed.

Definition SubAlg := Build_Bool_Alg Bc' und_set' assoc_sum' 
  assoc_prod' comm_sum' comm_prod' abs_sum' abs_prod' dist_sum'
  dist_prod' comp_sum' comp_prod'.


(*Being closed under 2 operations is sufficient to ensure that a subset is a subalgebra *)
Lemma two_ops_imp_times : Comp_closed_sub -> Plus_closed_sub -> Times_closed_sub.
intros h1 h2. unfold Times_closed_sub.
intros x y.
unfold SubBtype in x.   unfold SubBtype in y.
assert (h3: proj1_sig x * proj1_sig y = -(-(proj1_sig x) + -(proj1_sig y))).
apply times_eq_plus. unfold Btimes_sub. rewrite h3.
apply Comp_closed'' with (b:= (-proj1_sig x + -proj1_sig y)).
apply h1.
apply Plus_closed'' with (x := -proj1_sig x) (y := -proj1_sig y).
apply h2.
apply Comp_closed''.  apply h1. 
apply (proj2_sig x).
apply Comp_closed''.  apply h1.
apply (proj2_sig y).
Qed.

Lemma two_ops_imp_plus : Comp_closed_sub -> Times_closed_sub -> Plus_closed_sub.
intros h1 h2. unfold Plus_closed_sub.
intros x y.
unfold SubBtype in x.   unfold SubBtype in y.
assert (h3: proj1_sig x + proj1_sig y = -(-(proj1_sig x) * -(proj1_sig y))).
apply plus_eq_times. unfold Bplus_sub. rewrite h3.
apply Comp_closed'' with (b:= (-proj1_sig x * -proj1_sig y)).
apply h1.
apply Times_closed'' with (x := -proj1_sig x) (y := -proj1_sig y).
apply h2.
apply Comp_closed''.  apply h1. 
apply (proj2_sig x).
apply Comp_closed''.  apply h1.
apply (proj2_sig y).
Qed.

End Bc''.
End A_sec.

(*Document this case where even though two subsequent sections use the same section variables
, it's good to separate the sections if you need to use generalized versions of the first
section's variables in the second section.*)

Section Families_Ind_Families_Subalgebras.
Variable It : Type.
Variable A:Ensemble Bt.

(*Instead of duplicating the closure requirements, maybe assume  a priori the existence 
  of a function from the indexing type to the type of SubAlgebras of B*)
Record SubAlgI := {  
  sai_A:It->(Ensemble Bt);
  sai_closed:forall (i:It), A_closed (sai_A i);
  sai_B := fun (i:It) =>  SubAlg 
    (sai_A i) (sai_closed i)
  }.

(*Subalgebra Family*)
Record SubAlgF := {
  saf_A:Family Bt;
  saf_closed : forall (S:Ensemble Bt), In saf_A S -> A_closed S;
  saf_B := fun (S:{S':Ensemble Bt | In saf_A S'}) =>
    SubAlg (proj1_sig S) (saf_closed (proj1_sig S) (proj2_sig S))}.

Variable SI:SubAlgI.
Variable SF:SubAlgF.

Definition int_sai_A := IndexedIntersection (sai_A SI).
Definition int_saf_A := FamilyIntersection (saf_A SF).

Lemma Plus_closed_i : Plus_closed_sub int_sai_A.
red.
intros x y.
assert (h1: In int_sai_A (proj1_sig x)).
apply proj2_sig.
assert (h2: In int_sai_A (proj1_sig y)).
apply proj2_sig.
inversion h1. inversion h2.
assert (h3: forall (i:It), Plus_closed_sub (sai_A SI i)).
  intro i.
  apply (P_c _ (sai_closed SI i)).
apply indexed_intersection_intro.
intro i.
apply Plus_closed''. 
apply (h3 i). 
apply (H i). 
apply (H1 i).
Qed.

Lemma Plus_closed_F : Plus_closed_sub int_saf_A.
red.
intros x y.
constructor.
intros S h1.
apply Plus_closed''.
apply (P_c S (saf_closed SF S h1)).
destruct x as [? h2].
simpl.
inversion h2 as [? h3].
apply (h3 S h1).
destruct y as [? h4].
simpl.
inversion h4 as [? h5].
apply (h5 S h1).
Qed.


(*Dual*)
Lemma Times_closed_i : Times_closed_sub int_sai_A.
red.
intros x y.
assert (h1: In int_sai_A (proj1_sig x)).
apply proj2_sig.
assert (h2: In int_sai_A (proj1_sig y)).
apply proj2_sig.
inversion h1. inversion h2.
assert (h3: forall (i:It), Times_closed_sub (sai_A SI i)).
  intro i.
  apply (T_c _ (sai_closed SI i)).
apply indexed_intersection_intro.
intro i.
apply Times_closed''. 
apply (h3 i). 
apply (H i). 
apply (H1 i).
Qed.

Lemma Times_closed_F : Times_closed_sub int_saf_A.
red.
intros x y.
constructor.
intros S h1.
apply Times_closed''.
apply (T_c S (saf_closed SF S h1)).
destruct x as [? h2].
simpl.
inversion h2 as [? h3].
apply (h3 S h1).
destruct y as [? h4].
simpl.
inversion h4 as [? h5].
apply (h5 S h1).
Qed.

Lemma Comp_closed_i : Comp_closed_sub int_sai_A.
red.  intro x.
assert (h1: In int_sai_A (proj1_sig x)).
apply proj2_sig.
assert (h2: forall (i:It), Comp_closed_sub (sai_A SI i)).
  intro i.
  apply (C_c _ (sai_closed SI i)).
apply indexed_intersection_intro.
intro i.
apply Comp_closed''.
apply (h2 i).
inversion h1.
apply (H i).
Qed.

Lemma Comp_closed_F : Comp_closed_sub int_saf_A.
red.
intros x.
constructor.
intros S h1.
apply Comp_closed''.
apply (C_c S (saf_closed SF S h1)).
destruct x as [? h2]. simpl.
inversion h2 as [? h3]. 
apply (h3 S h1).
Qed.

Lemma One_closed_i : One_closed_sub int_sai_A.
red.
apply indexed_intersection_intro.
intro i.
pose proof (O_c _ (sai_closed SI i)) as h1.
unfold One_closed_sub in h1.
assumption.
Qed.

Lemma One_closed_F : One_closed_sub int_saf_A.
red.
constructor.
intros S h1.
apply (O_c S (saf_closed SF S h1)).
Qed.

Lemma Zero_closed_i : Zero_closed_sub int_sai_A.
red.
apply indexed_intersection_intro.
intro i.
pose proof (Z_c _ (sai_closed SI i)) as h1.
unfold Zero_closed_sub in h1.
assumption.
Qed.

Lemma Zero_closed_F : Zero_closed_sub int_saf_A.
red.
constructor.
intros S h1.
apply (Z_c S (saf_closed SF S h1)).
Qed.

(*Maybe turn SI into a section variable, and int into a definition*)

Lemma int_sai_A_closed : A_closed int_sai_A.
constructor.
apply Plus_closed_i.
apply Times_closed_i.
apply One_closed_i.
apply Zero_closed_i.
apply Comp_closed_i.
Qed.

Lemma int_saf_A_closed : A_closed int_saf_A.
constructor.
apply Plus_closed_F.
apply Times_closed_F.
apply One_closed_F.
apply Zero_closed_F.
apply Comp_closed_F.
Qed.


Definition int_sub_algi := SubAlg _ int_sai_A_closed.
Definition int_sub_algF := SubAlg _ int_saf_A_closed.


(*This definition is for the indexed family of all subalgebras that contain A*)
(*Maybe create new type for indexed families in set-theoretic notation [|] *) 

(*This definition defines a condition to ensure that SI exhausts all subalgebras
  containing A, and it's bijectively defined of course.*)

Definition S_clo_cont_A := [S:Ensemble Bt | Included A S /\ A_closed S].
Definition SA_cont_A : SubAlgF. 
pose S_clo_cont_A as F.
red in F.
assert (h1: forall (S:Ensemble Bt), In F S -> A_closed S).
  intros S h2.
  inversion h2 as [h3].  destruct h3 as [h3l h3r]. assumption.
refine (Build_SubAlgF F h1).
Defined.

Lemma saf_A_eq : saf_A SA_cont_A = S_clo_cont_A.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros S h1.
inversion h1.
constructor; assumption.
(* >= *)
red.
intros S h1.
unfold S_clo_cont_A in h1.
inversion h1.
simpl. constructor; assumption.
Qed.

(*
Definition A_Incl_SI : Prop := 
  (forall (i:It), Included A (sai_A SI i)) /\
  (forall (S:Ensemble Bt) (pf:A_closed S)
   (SB:(SubAlg S pf)), (exists i:It, SB = (sai_B SI i) /\
      Included A (BS (Bc SB)))).

Definition Cont_SI (A:(Ensemble Bt)) (SI:SubAlgI) : Prop := 
  (forall (i:It), Included (sai_A SI i) A).
*)
End Families_Ind_Families_Subalgebras.

Definition Gen (A:Ensemble Bt) : Bool_Alg.
pose (SA_cont_A A) as saf.
pose (int_saf_A saf) as FI.
refine (SubAlg FI (int_saf_A_closed saf)).
Defined.

(*Figure out why Coq doesn't recognize this notation*)
(*Notation "< A >" := (Gen A).*)

Lemma gen_minimal : forall (A S:Ensemble Bt), let F := (S_clo_cont_A A) in 
  In F S ->  Included (FamilyIntersection F) S.
intros A S F. 
(* -> *)
intro h1.
red.
intros x h2.
inversion h2 as [? h3].
apply (h3 S h1).
Qed.  

(*adjust this, seems redundant *)

Definition Gen_Ens (A:Ensemble Bt) : Ensemble Bt := 
  FamilyIntersection (S_clo_cont_A A).

(*
(*This is an alternative step-by-step definition, just in case it's useful *) 
Definition und_set_gen' (A:Ensemble Bt) : Ensemble Bt.
pose (Gen A) as G.
unfold Gen in G.
unfold SubAlg in G.
pose (Bc G) as Bc0.
simpl in Bc0.
unfold Bc' in Bc0.
pose (BS Bc0) as BS.
simpl in BS.
unfold SubBtype in BS.
refine (Im BS (@proj1_sig Bt _)).
Defined.
*)

(*adjust this seems redundant*)
(*Maybe include a definition on one subalgebra being included in another*)
Lemma gen_ens_minimal : forall (A S:Ensemble Bt), 
  let F := (S_clo_cont_A A) in In F S -> Included (Gen_Ens A) S.
intros. 
unfold Gen_Ens.
apply gen_minimal. assumption.
Qed.

Lemma gen_ens_includes : forall (S:Ensemble Bt), Included S (Gen_Ens S).
intro S.
red.
intros x h1.
constructor.
intros S0 h2.
inversion h2 as [h3].
destruct h3. auto with sets.
Qed.

Lemma gen_minimal'': forall (A S:Ensemble Bt), A_closed S ->
  Included A S -> 
  Included (int_saf_A (SA_cont_A A)) S.
intros A S h1 h2.
unfold int_saf_A.
pose proof (saf_A_eq A) as h3.
rewrite h3.
red.
intros x h4.
inversion h4 as [? h5].
apply h5. constructor. split; assumption.
Qed.

Section Gen_examples.

Example A_closed_two : A_closed (Couple 0 1).
  constructor.
  (* + *)
  red. intros x y.
  destruct x as [x h2].  destruct y as [y h3].
  unfold Bplus_sub. simpl.
  pose proof (Couple_inv _ _ _ _ h2) as h4.
  pose proof (Couple_inv _ _ _ _ h3) as h5.
  destruct h4 as [h4l | h4r].
    (*h4l*)
    destruct h5 as [h5l | h5r].
      (*h5l*)
      rewrite h4l. rewrite h5l.
      pose proof (zero_sum 0) as h6.
      rewrite h6.
      apply Couple_l.
      (*h5r*)
      rewrite h4l. rewrite h5r.
      pose proof (one_sum 0) as h6.
      rewrite h6.
      apply Couple_r.
    (*h4r*)
    destruct h5 as [h5l | h5r].
      (*h5l*)
      rewrite h4r. rewrite h5l.
      pose proof (zero_sum 1) as h6.
      rewrite h6.
      apply Couple_r.
      (*h5r*)
      rewrite h4r. rewrite h5r.
      pose proof (one_sum 1) as h6.
      rewrite h6.
      apply Couple_r.
  (* * *)
  red.  
  intros x y.
  destruct x as [x h1]. destruct y as [y h2].
  unfold Btimes_sub. simpl.
  pose proof (Couple_inv _ _ _ _ h1) as h3.
  pose proof (Couple_inv _ _ _ _ h2) as h4.
  destruct h3 as [h3l | h3r].
  (*h3l*)
    destruct h4 as [h4l | h4r].
    (*h4l*)
    rewrite h3l. rewrite h4l.
    pose proof (zero_prod 0) as h5.
    rewrite h5.
    apply Couple_l.
    (*h4r*)
    rewrite h3l. rewrite h4r.
    pose proof (one_prod 0) as h5.
    rewrite h5.
    apply Couple_l.
  (*h3r*)
    destruct h4 as [h4l | h4r].
    (*h4l*)
    rewrite h3r. rewrite h4l.
    pose proof (zero_prod 1) as h5.
    rewrite h5.
    apply Couple_l.
    (*h4r*)
    rewrite h3r. rewrite h4r.
    pose proof (one_prod 1) as h5.
    rewrite h5.
    apply Couple_r.
  (* 1 *)
  red. apply Couple_r.
  (* 0 *)
  red. apply Couple_l.
  (* - *)
  red. intro x.
  destruct x as [x h1].
  unfold Bcomp_sub. simpl.
  pose proof (Couple_inv _ _ _ _ h1) as h2.
  destruct h2 as [h2l | h2r].
  (*h2l*)
  rewrite h2l.
  pose proof (zero_comp) as h3.
  rewrite h3.
  apply Couple_r.
  (*h2r*)
  rewrite h2r.
  pose proof (one_comp) as h3.
  rewrite h3.
  apply Couple_l.
Qed.


Definition two : Bool_Alg.
pose (@Couple Bt 0 1) as A.
pose proof (A_closed_two) as h1.
refine (SubAlg A h1).
Defined.

Example two_smallest_subalgebra : forall (A:Ensemble Bt), A_closed A -> 
  let C := (Couple 0 1) in Included C A.
intros A h1 C.
red.
intros x h2.
pose proof (Couple_inv _ _ _ _ h2) as h3.
destruct h3 as [h3l | h3r].
rewrite h3l.
apply (Z_c _ h1).
rewrite h3r.
apply (O_c _ h1).
Qed.

Lemma gen_ens_closed_eq : forall (S:Ensemble Bt), A_closed S -> Gen_Ens S = S.
intros S h1.
(*unfold Gen_Ens.*)
apply Extensionality_Ensembles.
red. split.
(* <= *)
apply gen_ens_minimal. 
constructor. split. auto with sets. assumption.
(* >= *)
apply gen_ens_includes.
Qed.

(*"if E is empty and A is not degenerate, then the
subalgebra generated by E is the smallest possible subalgebra of A, namely 2."*)

Example gen_ens_empty : Gen_Ens (Empty_set Bt) = (Couple 0 1).
unfold Gen_Ens.
rewrite <- saf_A_eq.
apply Extensionality_Ensembles.
red. split.
(* <= *)
apply gen_minimal.
unfold S_clo_cont_A.
constructor.
split.
  auto with sets.  apply A_closed_two.
(* >= *)
red.
intros x h1.
constructor.
intros S h2.
inversion h2 as [h3].
destruct h3 as [? h4].
pose proof (Couple_inv _ _ _ _ h1) as [h5 | h6].
rewrite h5.
apply (Z_c _ h4).
rewrite h6.
apply (O_c _ h4).
Qed.

Example gen_empty : Gen (Empty_set Bt) = two.
unfold Gen. unfold two.
pose proof gen_ens_empty as h1.
unfold Gen_Ens in h1.
rewrite <- saf_A_eq in h1.
unfold int_saf_A.
pose proof (subsetT_eq_compat _ _ _ _ 
  (int_saf_A_closed (SA_cont_A (Empty_set Bt))) 
  A_closed_two h1) as h2.
dependent rewrite -> h2.
reflexivity.
Qed.

End Gen_examples.

(*"A simple but useful remark for subsets E and F of a Boolean algebra A is
that if F is included in E, then the subalgebra generated by F is included (as
a subalgebra) in the subalgebra generated by E."*)

Lemma gen_ens_preserves_inclusion : forall (S U: Ensemble Bt), Included S U ->
  Included (Gen_Ens S) (Gen_Ens U).
intros S U h1.
unfold Gen_Ens.
red.
intros x h2.
inversion h2 as [? h3].
assert (forall S0 : Ensemble Bt, In (S_clo_cont_A U) S0 -> In S0 x).
  intros S0 h5.
  assert (h6:In (S_clo_cont_A S) S0).
  inversion h5 as [h6].
  destruct h6.
  assert (h7:Included S S0).
    auto with sets.
  constructor. split; assumption.
  apply h3; assumption.
constructor; assumption.
Qed.

(*Try to adjust so that the section variable B or module variable is abstracted.*)

(*"The relation of one Boolean algebra being a subalgebra of another is a
partial order on the class of all Boolean algebras."  Technically I don't define
[subalg_of] exactly like that, since I presume an underlying algebra B, but 
my definition is equivalent to the informal definition given in the quotes, 
so I know this definition and lemma are pretty vacuous.*)

Definition subalg_of (A1 A2:{A:Ensemble Bt | A_closed A}) : Prop := 
  (Included (proj1_sig A1) (proj1_sig A2)).

Lemma order_subalg_of : Order _ subalg_of.
constructor.
(*refl*)
red.
intro.
red.
auto with sets.
(*trans*)
red. unfold subalg_of. intros.
auto with sets.
(*antisymm*)
red. unfold subalg_of. intros A1 A2 h1 h2.
cut ((proj1_sig A1) = (proj1_sig A2)).
apply proj1_sig_injective.
auto with sets.
Qed.

(* Skip for now "It is an interesting and occasionally useful observation that the class of all
subalgebras of a Boolean algebra A is a complete lattice under the relation
of being a subalgebra."*)

(*"The family is said to be directed if any two members B_i
and B_j of the family are always subalgebras of some third member Bk."*)

Definition directed (F:SubAlgF) : Prop := forall (S1 S2:Ensemble Bt), 
  In (saf_A F) S1 -> In (saf_A F) S2 -> (exists (S3:Ensemble Bt),
    In (saf_A F) S3 /\
    Included S1 S3 /\ Included S2 S3).

(*"Lemma 1. The union of a non-empty, directed family of subalgebras is again
a subalgebra."*)

Lemma union_directed_SubAlgf_closed : forall (F:SubAlgF), directed F -> 
  Inhabited (saf_A F) -> A_closed (FamilyUnion (saf_A F)).
intros F h1 h2.
constructor.
(* + *)
red.
intros p q.
destruct p as [p h3]. destruct q as [q h4].
unfold Bplus_sub. simpl.
inversion h3 as [B_i ? h5 h6].
inversion h4 as [B_j ? h7 h8].
unfold directed in h1.
pose proof (h1 _ _ h5 h7) as h9.
elim h9. intros B_k h10.
destruct h10 as [h10 [h11l h11r]].
assert (h12:In B_k p).  auto with sets.
assert (h13:In B_k q).  auto with sets.
pose proof (saf_closed F B_k h10) as h14.
pose proof (P_c _ h14) as h15.
red in h15.
pose proof (h15 (exist _ p h12) (exist _ q h13)) as h16.
unfold Bplus_sub in h16.
simpl in h16.
apply family_union_intro with B_k; assumption.
(* * *)
red.
intros p q.
destruct p as [p h3]. destruct q as [q h4].
unfold Btimes_sub. simpl.
inversion h3 as [B_i ? h5 h6].
inversion h4 as [B_j ? h7 h8].
unfold directed in h1.
pose proof (h1 _ _ h5 h7) as h9.
elim h9. intros B_k h10.
destruct h10 as [h10 [h11l h11r]].
assert (h12:In B_k p).  auto with sets.
assert (h13:In B_k q).  auto with sets.
pose proof (saf_closed F B_k h10) as h14.
pose proof (T_c _ h14) as h15.
red in h15.
pose proof (h15 (exist _ p h12) (exist _ q h13)) as h16.
unfold Btimes_sub in h16.
simpl in h16.
apply family_union_intro with B_k; assumption.
(* 1 *)
red.
inversion h2 as [S h3].
pose proof (saf_closed F S h3) as h4.
pose proof (O_c _ h4) as h5.
red in h5.
apply family_union_intro with S; assumption.
(* 0 *)
red.
inversion h2 as [S h3].
pose proof (saf_closed F S h3) as h4.
pose proof (Z_c _ h4) as h5.
red in h5.
apply family_union_intro with S; assumption.
(* - *)
red. intro p.
destruct p as [p h3].
unfold Bcomp_sub. simpl.
inversion h3 as [B_i ? h4 h5].
pose proof (saf_closed F B_i h4) as h6.
pose proof (C_c _ h6) as h7.
red in h7.
pose proof (h7 (exist _ p h5)) as h8.
red in h8. unfold Bcomp_sub in h8.
simpl in h8.
apply family_union_intro with B_i; assumption.
Qed.

(* "More precisely, a family {Bi } of subalgebras is a chain if for any two members Bi
and Bj of the family, either Bi is a subalgebra of Bj , or vice versa." *)

Definition chain_subalg (F:SubAlgF) : Prop := chain _ (saf_A F).

Lemma chain_impl_directed : forall (F:SubAlgF), chain_subalg F -> directed F.
intros F h1.
red in h1.
red.
intros S1 S2 h2 h3.
red in h1.
pose proof (h1 _ _ h2 h3) as h4.
case h4 as [h4l | h4r].
  (* left *)
  exists S2. split. assumption.
  split. assumption. auto with sets.
  (* right *)
  exists S1. split. assumption.
  split. auto with sets. assumption. 
Qed. 

  
Lemma union_chain_SubAlgf_closed : forall (F:SubAlgF), chain_subalg F -> 
  Inhabited (saf_A F) -> A_closed (FamilyUnion (saf_A F)).
intros F h1 h2.
apply union_directed_SubAlgf_closed.
apply chain_impl_directed; assumption.
assumption.
Qed.

(** "Corollary 1. Let $A$ be a Boolean algebra generated by a set $E$, and for each finite subset $F$ of $E$, let $B_F$ be the subalgebra of $A$ generated by $F$. The family $\{B_F:F \subseteq E \text{ and }  F \text{ is finite}\}$ is directed, and its union is $A$."
[or in my terminology]: 
Assume B is generated by a set E, and for each
finite subset S of E, let A be the subalgebra of B generated by S. The family
{A : A is generated by S and S <= E and S is finite} is directed, and its union is B. *)

Lemma family_included_subs_directed_eq_B : forall (E:Ensemble Bt) (F:SubAlgF), 
  (Full_set Bt) = Gen_Ens E -> (saf_A F) = 
  [A:Ensemble Bt | exists (S:Ensemble Bt), 
    A = Gen_Ens S /\ Included S E /\ Finite _ S] -> 
  directed F /\ FamilyUnion (saf_A F) = (Full_set Bt).
intros E F h1 h2.
assert (h3:directed F).
  red.
  intros S1' S2' h3 h4.
  rewrite h2.
  rewrite h2 in h3. rewrite h2 in h4.
  inversion h3 as [h5].
  inversion h4 as [h6].
  elim h5. intros S1 h7. destruct h7 as [h7a [h7b h7c]].
  elim h6. intros S2 h8. destruct h8 as [h8a [h8b h8c]].
  assert (h9 :Included S1 (Union S1 S2)). auto with sets.
  assert (h10:Included S2 (Union S1 S2)). auto with sets.
  pose proof (gen_ens_preserves_inclusion S1 (Union S1 S2) h9)  as h11.
  pose proof (gen_ens_preserves_inclusion S2 (Union S1 S2) h10) as h12.
  exists (Gen_Ens (Union S1 S2)).
  constructor. constructor.
  exists (Union S1 S2). split. reflexivity.
  split. auto with sets. apply Union_preserves_Finite;assumption.
  rewrite h7a. rewrite h8a.
  split; assumption.

split. assumption.
(* F's union is Full_set *)
pose proof (union_directed_SubAlgf_closed) as h4.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. intros x ?. constructor.
(* >= *)
assert (h5:Included E (FamilyUnion (saf_A F))).
  red. intros x h6.
  pose proof (In_singleton _ x) as h7.
  apply family_union_intro with (Gen_Ens (Singleton x)).
  rewrite h2.
  constructor. exists (Singleton x). split. reflexivity.
  split. red. intros x' h8.
  pose proof (Singleton_inv _ _ _ h8) as h9.
  rewrite <- h9. assumption.
  apply Singleton_is_finite.
  constructor. intros S h8.
  inversion h8 as [h9]. destruct h9 as [h9l h9r].
  auto with sets. 
rewrite h1.
apply gen_ens_minimal.
constructor.
split. assumption.
pose proof (Inhabited_intro _ (saf_A F) (Couple 0 1)) as h6.
assert (h7:In (saf_A F) (Couple 0 1)).
  rewrite h2.
  constructor. exists (Empty_set Bt).
  split.
  (*left*) 
  rewrite <- gen_ens_empty. reflexivity.
  (*right*)
  split. auto with sets. constructor.
pose proof (h6 h7) as h8.
pose proof (h4 _ h3 h8). assumption. 
Qed.


(*"For any Boolean algebra A [B in my terminology], 
apply the preceding corollary to the set E = A [B]
to conclude that every Boolean algebra is the (directed) union of its finitely
generated subalgebras."*)

Corollary family_included_subs_directed_eq_B_full_set : 
  forall (F:SubAlgF),
  (saf_A F) = [A:Ensemble Bt | exists (S:Ensemble Bt), 
    A = Gen_Ens S /\ Finite _ S] -> 
  directed F /\ FamilyUnion (saf_A F) = Full_set Bt.
Proof.
intros F h1.
pose proof (family_included_subs_directed_eq_B (Full_set Bt) F) as h2.
assert (h3:Full_set Bt = Gen_Ens (Full_set Bt)).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x ?.
  unfold Gen_Ens.
  constructor.
  intros S h4.
  inversion h4 as [h5].
  destruct h5; auto with sets.
  (* >= *)
  red. intros.  constructor. 
assert (h4:saf_A F = [A : Ensemble Bt
     | exists S : Ensemble Bt,
         A = Gen_Ens S /\ Included S (Full_set Bt) /\ Finite Bt S]).
  rewrite h1.
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros S h4.
  inversion h4 as [h5].
  destruct h5 as [E h6].
  destruct h6.
  constructor.
  exists E.
  repeat split; auto with sets.
  (* >= *)
  red.  intros S h4.
  inversion h4 as [h5].
  destruct h5 as [E h6].
  destruct h6 as [? []].
  constructor.
  exists E.
  split; assumption.
apply h2; assumption.
Qed.

Parameter B':Bool_Alg.
Declare Module BA':BAt with Definition B:=B'.
Module Bscs' := Bscs BA'.
Declare Module I:INFNT with Definition B:=B with Module BA:=BA.

Require Import ListUtilities. 
Definition atom (b:BA'.Bt) : 
  Prop := b <> 0 /\ forall a:BA'.Bt, Bscs'.le a b -> a = b.

Inductive sign:Set := pls|mns.
Definition signl:list sign := pls::mns::nil.

(*Whereas Givant/Halmos uses "p", I use the Handbook's notation epsilon, but with
  arguments in the same order as G/H's "p" *)
Definition eps(b:Bt)(sgn:sign) := if sgn then b else -b.
Definition eps'(pr:Bt*sign) := eps (fst pr) (snd pr).

Fixpoint times_list {B0:Bool_Alg} (l:list (Btype (Bc B0))) : (Btype (Bc B0)) := 
  match l with 
  | nil => 1
  | cons a l' => a * (times_list l')
  end.

Fixpoint plus_list {B0:Bool_Alg} (l:list (Btype (Bc B0))) : (Btype (Bc B0)) := 
  match l with 
  | nil => 0
  | cons a l' => a + (plus_list l')
  end.

Lemma plus_list_app : forall (l1 l2:list Bt), 
  plus_list (l1 ++ l2) = plus_list l1 + plus_list l2.
intros l1 l2.
induction l1 as [|a l1' h1].
simpl. rewrite comm_sum. rewrite zero_sum. reflexivity.
(* a::l1'*)
simpl.
rewrite h1.
apply assoc_sum.
Qed.

Lemma times_list_app : forall (l1 l2:list Bt), 
  times_list (l1 ++ l2) = times_list l1 * times_list l2.
intros l1 l2.
induction l1 as [|a l1' h1].
simpl. rewrite comm_prod. rewrite one_prod. reflexivity.
(* a::l1'*)
simpl.
rewrite h1.
apply assoc_prod.
Qed.


(* maybe turn Bt into a general Boolean algebraic type *)
Lemma dist_list_sing_plus : forall (l:list Bt) (x:Bt), 
  x * (plus_list l) = plus_list (map (fun y:Bt => (x*y)) l).
intros l x.
induction l as [|a l h1].
(* nil *)
simpl.  apply zero_prod.
simpl.
rewrite dist_sum.
rewrite h1.
reflexivity.
Qed.

Lemma dist_list_sing_times : forall (l:list Bt) (x:Bt), 
  x + (times_list l) = times_list (map (fun y:Bt => (x+y)) l).
intros l x.
induction l as [|a l h1].
(* nil *)
simpl.  apply one_sum.
simpl.
rewrite dist_prod.
rewrite h1.
reflexivity.
Qed.

Lemma dist_list_2_plus : forall (l1 l2:list Bt), 
  (plus_list l1) * (plus_list l2) = 
  plus_list (map (fun p:Bt*Bt => (fst p * snd p)) (list_prod l1 l2)).
intros l1 l2.
induction l1 as [|a l1' h1]; destruct l2 as [|b l2']; simpl.
(* nil nil *)
auto. 
(* nil b::l2' *)
rewrite comm_prod.
apply zero_prod.
(* a::l1' nil *)
simpl in h1.
rewrite <- h1.
do 2 rewrite zero_prod.
reflexivity.
(* a::l1' b::l2'*)
simpl in h1.
rewrite comm_prod at 1.
rewrite dist_sum.
rewrite comm_prod in h1.
rewrite h1.
rewrite comm_prod.
rewrite dist_sum.
rewrite map_app.

assert (h3:a*plus_list l2' = 
  plus_list (map (fun p:Bt * Bt => fst p * snd p) (map (fun y:Bt => (a, y)) l2'))).
  rewrite dist_list_sing_plus.
  assert (h4:map (fun y:Bt => a * y) l2' = 
    map (fun p : Bt * Bt => fst p * snd p) (map (fun y:Bt => (a, y)) l2')).
    rewrite map_map. simpl. reflexivity.  
  rewrite h4.
  reflexivity.
rewrite h3.
rewrite plus_list_app.
rewrite <- assoc_sum.
reflexivity.
Qed.


Lemma dist_list_2_times : forall (l1 l2:list Bt), 
  (times_list l1) + (times_list l2) = 
  times_list (map (fun p:Bt*Bt => (fst p + snd p)) (list_prod l1 l2)).
intros l1 l2.
induction l1 as [|a l1' h1]; destruct l2 as [|b l2']; simpl.
(* nil nil *)
auto. 
(* nil b::l2' *)
rewrite comm_sum.
apply one_sum.
(* a::l1' nil *)
simpl in h1.
rewrite <- h1.
do 2 rewrite one_sum.
reflexivity.
(* a::l1' b::l2'*)
simpl in h1.
rewrite comm_sum at 1.
rewrite dist_prod.
rewrite comm_sum in h1.
rewrite h1.
rewrite comm_sum.
rewrite dist_prod.
rewrite map_app.

assert (h3:a+times_list l2' = 
  times_list (map (fun p:Bt * Bt => fst p + snd p) (map (fun y:Bt => (a, y)) l2'))).
  rewrite dist_list_sing_times.
  assert (h4:map (fun y:Bt => a + y) l2' = 
    map (fun p : Bt * Bt => fst p + snd p) (map (fun y:Bt => (a, y)) l2')).
    rewrite map_map. simpl. reflexivity.  
  rewrite h4.
  reflexivity.
rewrite h3.
rewrite times_list_app.
rewrite <- assoc_prod.
reflexivity.
Qed.

Definition plus_times_list_of_lists (l:list (list Bt)) : Bt :=
  plus_list (map (fun l':(list Bt) => times_list l') l).

Definition times_plus_list_of_lists (l:list (list Bt)) : Bt :=
  times_list (map (fun l':(list Bt) => plus_list l') l).

(* This is the complete distributive rule for finite sets, as 
 quoted in (8.3), but mine is more general since it allows you to 
 take the meet of the joins of lists of arbitrary length
 (not fixed as in the book) *)

Lemma complete_dist_list_times : forall (l:list (list Bt)), 
  times_plus_list_of_lists l = plus_times_list_of_lists (list_of_lists_seqs l).
intro l.
induction l as [|al l' h1].
(* nil *)
unfold plus_times_list_of_lists. simpl. 
rewrite zero_sum. reflexivity.
(* al::l'*)
simpl.
unfold times_plus_list_of_lists.
simpl.
unfold times_plus_list_of_lists in h1.
rewrite h1.
unfold plus_times_list_of_lists.
rewrite dist_list_2_plus.
rewrite map_map. simpl.
induction al as [|a t h2].
  (*nil*)
  simpl. reflexivity.
  (*a::t*)
  simpl.
  rewrite map_map.
  rewrite map_app.
  rewrite plus_list_app.
  rewrite h2.
  rewrite map_app.
  rewrite plus_list_app.
  do 2 rewrite map_map.
  simpl.
  reflexivity.
Qed.

Lemma complete_dist_list_plus : forall (l:list (list Bt)), 
  plus_times_list_of_lists l = times_plus_list_of_lists (list_of_lists_seqs l).
intro l.
induction l as [|al l' h1].
(* nil *)
unfold times_plus_list_of_lists. simpl. 
unfold plus_times_list_of_lists. simpl.
rewrite comm_prod.
rewrite zero_prod. reflexivity.
(* al::l'*)
simpl.
unfold plus_times_list_of_lists.
simpl.
unfold plus_times_list_of_lists in h1.
rewrite h1.
unfold times_plus_list_of_lists.
rewrite dist_list_2_times.
rewrite map_map. simpl.
induction al as [|a t h2].
  (*nil*)
  simpl. reflexivity.
  (*a::t*)
  simpl.
  rewrite map_map.
  rewrite map_app.
  rewrite times_list_app.
  rewrite h2.
  rewrite map_app.
  rewrite times_list_app.
  do 2 rewrite map_map.
  simpl.
  reflexivity.
Qed.


Lemma inf_union : forall (U V:Ensemble Bt) (u v:Bt), inf U u -> inf V v ->
  (inf (Union U V) (u * v)).
intros U V u v h1 h2.
red in h1. destruct h1 as [h1l h1r].
red in h2. destruct h2 as [h2l h2r].
red.
split.
(* lower bound *)
red. red in h1l. red in h2l.
intros s h3.
pose proof (Union_inv _ _ _ _ h3) as h4.
  destruct h4 as [h4l | h4r].
  (* h4l *)
  (* maybe turn into a separate lemma *)
  assert (h5:le (u * v) u).
    red. rewrite eq_ord. rewrite comm_prod. rewrite assoc_prod.
    rewrite idem_prod. reflexivity.
  specialize (h1l _ h4l). apply (trans_le _ _ _ h5 h1l).
  assert (h5:le (u * v) v).
    red. rewrite eq_ord. rewrite <- assoc_prod. rewrite idem_prod. 
    reflexivity. 
  specialize (h2l _ h4r). apply (trans_le _ _ _ h5 h2l).
(* greatest lower bound *)
intros b h3.
red in h3.
assert (h4:lb U b).
  red.
  intros s h5.
  assert (Ensembles.In (Union U V) s). auto with sets.
  apply h3; assumption.
assert (h5:lb V b).
  red.
  intros s h6.
  assert (Ensembles.In (Union U V) s). auto with sets.
  apply h3; assumption.
specialize (h1r b h4).
specialize (h2r b h5).
pose proof (mono_prod _ _ _ _ h1r h2r) as h6.
rewrite idem_prod in h6.
assumption.
Qed.

Lemma sup_union : forall (U V:Ensemble Bt) (u v:Bt), sup U u -> sup V v ->
  (sup (Union U V) (u + v)).
intros U V u v h1 h2.
red in h1. destruct h1 as [h1l h1r].
red in h2. destruct h2 as [h2l h2r].
red.
split.
(* upper bound *)
red. red in h1l. red in h2l.
intros s h3.
pose proof (Union_inv _ _ _ _ h3) as h4.
  destruct h4 as [h4l | h4r].
  (* h4l *)
  (* maybe turn into a separate lemma *)
  assert (h5:le u (u + v)).
    red. rewrite assoc_sum. rewrite idem_sum. reflexivity. 
  specialize (h1l _ h4l). apply (trans_le _ _ _ h1l h5).
  assert (h5:le v (u + v)).
    red. rewrite comm_sum.  rewrite <- assoc_sum. rewrite idem_sum. reflexivity. 
  specialize (h2l _ h4r). apply (trans_le _ _ _ h2l h5).
(* least upper bound *)
intros b h3.
red in h3.
assert (h4:ub U b).
  red.
  intros s h5.
  assert (Ensembles.In (Union U V) s). auto with sets.
  apply h3; assumption.
assert (h5:ub V b).
  red.
  intros s h6.
  assert (Ensembles.In (Union U V) s). auto with sets.
  apply h3; assumption.
specialize (h1r b h4).
specialize (h2r b h5).
pose proof (mono_sum _ _ _ _ h1r h2r) as h6.
rewrite idem_sum in h6.
assumption.
Qed.


Lemma decompose_inf : forall (S X:Ensemble Bt) (s x x':Bt), Included X S -> inf S s ->
  inf (Setminus S X) x' -> inf X x -> inf (Couple x x') s.
intros S X s x x' h1 h2 h3 h4.
red.
red in h2. destruct h2 as [h2l h2r].
red in h3. destruct h3 as [h3l h3r].
red in h4. destruct h4 as [h4l h4r].
split.
(* lower bound *)
red.
intros b h5.
pose proof (Couple_inv _ _ _ _ h5) as h6.
  destruct h6 as [h6l | h6r].
  (* b = x *)
  subst.
  red in h2l.
  assert (h6: forall y:Bt, Ensembles.In X y -> Ensembles.In S y). auto with sets.
  assert (h7: forall y:Bt, Ensembles.In X y -> le s y).  
    intros y h8. specialize (h6 _ h8).
    apply h2l; assumption.
  assert (h8: lb X s). red. assumption.
  apply h4r; assumption.
  (* b = x'*)
  subst.
  assert (h6:Included (Setminus S X) S). 
    red. intros b h7. unfold Setminus in h7. red in h7. 
    tauto.
  red in h3l.
  assert (h7: forall y:Bt, Ensembles.In (Setminus S X) y -> Ensembles.In S y). auto with sets.
  assert (h8: forall y:Bt, Ensembles.In (Setminus S X) y -> le s y).
    intros y h9. specialize (h7 _ h9).
    apply h2l; assumption.
  assert (h9: lb (Setminus S X) s). red. assumption.
  apply h3r; assumption.
(* greatest lower bound *)
intros a h5.
assert (h6: lb S a).
  red in h5.
  pose proof (Couple_l _ x x') as h6.
  pose proof (Couple_r _ x x') as h7.
  pose proof (h5 _ h6) as h8.
  pose proof (h5 _ h7) as h9.
  red in h3l. red in h4l.
  red.
  intros c h10.
  case (classic (Ensembles.In X c)) as [h11 | h12].
  (* In X c *)
  pose proof (h4l _ h11) as h13.
  apply trans_le with x; assumption.
  (* ~ In X c *)
  assert (h13: Ensembles.In (Setminus S X) c). auto with sets.
  pose proof (h3l _ h13) as h14.
  apply trans_le with x'; assumption.
apply h2r; assumption.
Qed.

Lemma decompose_inf': forall (S X:Ensemble Bt) (x x':Bt), Included X S -> 
  inf X x -> inf (Setminus S X) x' -> inf S (x * x').
intros S X x x' h1 h2 h3.
pose proof decompose_inf.
case (classic (exists s, inf S s)) as [h4 | h5].
(* h4 *)
destruct h4 as [s h4].
assert (h5:s = x * x').
  pose proof (decompose_inf _ _ _ _ _ h1 h4 h3 h2) as h5.
  pose proof (lat_prod x x') as h6.
  pose proof (inf_unq _ _ _ h5 h6).
  assumption.
rewrite h5 in h4.
assumption.
(* h5 *)
contradict h5.
exists (x * x').
(* maybe turn into a separate lemma *)
assert (h6:S = Union (Setminus S X) X).
  apply Extensionality_Ensembles. red. split.
  (* <= *)
  red. intros y h7.
  case (classic (Ensembles.In X y)) as [h8 | h9].
  (* h8 *) apply Union_intror; assumption.
  (* h9 *) apply Union_introl. unfold Setminus. red. split; assumption.
  (* >= *)
  red. intros y h7.
  pose proof (Union_inv _ _ _ _ h7) as h8.
  destruct h8 as [h8l | h8r].  
  unfold Setminus in h8l. red in h8l. tauto.
  auto with sets.
rewrite h6.
pose proof (inf_union _ _ _ _ h3 h2) as h7.
rewrite comm_prod in h7.
assumption. 
Qed.

(* Probably not needed. *)
Lemma inf_times_cons : forall (E:Ensemble Bt) (a b:Bt), Ensembles.In E a -> 
  inf (Subtract E a) b -> inf E (a * b).
intros E a b h1 h2.
pose proof decompose_inf'.
pose proof (inf_singleton a) as h3.
unfold Subtract in h2.
assert (h4:Included (Singleton a) E). 
  red. intros y h5. destruct h5; assumption. 
apply (decompose_inf' _ _ _ _ h4 h3 h2).
Qed.


Lemma inf_times_finite : forall (l:list Bt), 
  inf (list_to_set l) (times_list l).
intros l.
induction l.
(* nil *)
simpl.
red. split. 
red.  intros. contradiction.
auto.
(* a::l *)
simpl.
unfold Add.
rewrite comm_sum_psa.
apply inf_union.
apply inf_singleton.
assumption.
Qed.

Lemma sup_plus_finite : forall (l:list Bt), 
  sup (list_to_set l) (plus_list l).
intros l.
induction l.
(* nil *)
simpl.
red. split. 
red.  intros. contradiction.
auto.
(* a::l *)
simpl.
unfold Add.
rewrite comm_sum_psa.
apply sup_union.
apply sup_singleton.
assumption.
Qed.

Lemma le_times_finite_member : forall (l:list Bt) (x:Bt), 
  (In x l) -> le (times_list l) x.
intros l x h1.
induction l as [|a l h2].
(* nil *)
simpl in h1.
contradiction.
(* a::l *)
apply in_inv in h1.
destruct h1 as [h1l | h1r].
  (* a = x *)
  subst.
  simpl.
  apply times_le.
  (* In x l *)
  specialize (h2 h1r).
  simpl.
  pose proof (times_le (times_list l) a) as h3.
  rewrite comm_prod in h3.
  apply trans_le with (times_list l); assumption.
Qed.

Lemma le_member_plus_finite : forall (l:list Bt) (x:Bt), 
  (In x l) -> le x (plus_list l).
intros l x h1.
induction l as [|a l h2].
(* nil *)
simpl in h1.
contradiction.
(* a:: l*)
apply in_inv in h1.
destruct h1 as [h1l | h1r].
  (* a = x *)
  subst.
  simpl.
  apply le_plus.
  (* In x l *)
  specialize (h2 h1r).
  simpl.
  pose proof (le_plus (plus_list l) a) as h3.
  rewrite comm_sum in h3.
  apply trans_le with (plus_list l); assumption.
Qed.


Definition transfer_gen {E:Ensemble Bt} (b:Btype (Bc (Gen E))) : Bt.
unfold SubAlg in b.
unfold Bc' in b.
simpl in b.
unfold SubBtype in b.
refine (proj1_sig b).
Defined.

Definition transfer_eq (B1 B2:Bool_Alg) (pf:B1 = B2) (x:Btype (Bc B1)) : Btype (Bc (B2)).
rewrite pf in x.
refine x.
Defined.






Lemma one_times_list : forall (l:list Bt), 
  @times_list B (map (fun i:Bt => 1) l) = 1.
intro l.
induction l as [| ? ? h1].
simpl. reflexivity.
simpl.
rewrite h1.
auto.
Qed.

Lemma zero_plus_list : forall (l:list Bt), 
  @plus_list B (map (fun i:Bt => 0) l) = 0.
intros l.
induction l as [| ? ? h1].
simpl. reflexivity.
simpl.
rewrite h1.
auto.
Qed.

(*This is 2^l, the set of all functions from l to 2 = signl*)
Definition twol (l:list Bt) := list_fun l signl.

(*This corresponds to /\p(i, a(i)) as i runs through l, 
  from the normal form Theorem : 11.2 Might delete.*)
Definition p_ {l:list Bt} (a:(twol l)) := 
  times_list (map (fun i:{xl:Bt | In xl l} => 
  eps (proj1_sig i)(proj1_sig (a i))) (map_sig l)).

(*This is an alternate form of the same thing (/\p(i, a(i)) in a less 
  functional and more list-oriented form.*)
Definition p_' (la:list (Bt*sign)) :=
  times_list (map (fun (pr:Bt*sign) => eps' pr) la).


(*eps b (proj1_sig (a (exist (fun x : Bt => b = x \/ In x l) b (in_eq b l)))) *
times_list
  (map (fun pr : Bt * sign => eps' pr) (proj1_sig (lfun_to_pair_list a'))) =
times_list
  (map (fun pr : Bt * sign => eps' pr) (proj1_sig (lfun_to_pair_list a)))*)

Lemma times_list_map_cons : forall {T:Type} {l:list T} {b:T}
  (f:{x:T | In x (b::l)}->Bt), times_list (map f (map_sig (b::l))) = 
  (f (exist (fun x:T => (In x (b::l))) b (in_eq b l))) * 
  times_list (map (f_list_pop f) (map_sig l)).
intros T l b f.
simpl.
rewrite map_map.
unfold f_list_pop.
reflexivity.
Qed.

(* Commenting out to allow compilation *)
(*
Lemma p_equiv : forall (l:list Bt) (a:(twol l)), p_ a = 
  p_' (proj1_sig (lfun_to_pair_list a)).
intros l a.
induction l as [|b l h2].
(* nil *)
unfold p_. unfold p_'.
unfold twol in a.
simpl.
assert (h1:nil = (proj1_sig (lfun_to_pair_list a))).
  pose proof (proj2_sig (lfun_to_pair_list a)) as h2.
  simpl in h2.
  destruct h2; [assumption | contradiction].
assert (h2:1 = times_list (map (fun pr : Bt * sign => eps' pr) nil)).
  simpl. reflexivity.
rewrite h1 in h2.
assumption.
(* b::l *)
unfold p_. (*simpl.*)
unfold p_ in h2. 
pose (f_list_pop a) as a'.
specialize (h2 a').
simpl.
rewrite map_map.
simpl.
(*
assert (h3: (fun x0:Bt => In x0 (b::l)) = (fun x0:Bt => b = x0 \/ In x0 l)).
  apply functional_extensionality.
  intros x.
  assert (h4:In x (b :: l) <-> (b = x \/ In x l)).
    split. apply in_inv.
    intro h5. destruct h5 as [h5l | h5r].
    rewrite h5l. auto with datatypes. apply in_cons; assumption.
  apply Prop_Extensionality; assumption.*)
unfold a' in h2 at 1.
assert (h4:
  map (fun x : {x : Bt | In x l} =>
      eps (proj1_sig x)
        (proj1_sig
           (a
              (exist (fun x0 : Bt => b = x0 \/ In x0 l) (proj1_sig x)
                 (in_cons b (proj1_sig x) l (proj2_sig x)))))) (map_sig l) = 
  map (fun i : {xl : Bt | In xl l} =>
           eps (proj1_sig i)
             (proj1_sig
                (a
                   (exist (fun x : Bt => In x (b :: l)) (proj1_sig i)
                      (in_cons b (proj1_sig i) l (proj2_sig i)))))) (map_sig l)).
  apply map_ext.
  intro x.
  assert (h5:(exist (fun x0 : Bt => b = x0 \/ In x0 l) (proj1_sig x)
           (in_cons b (proj1_sig x) l (proj2_sig x))) = 
           (exist (fun x0 : Bt => In x0 (b :: l)) (proj1_sig x)
           (in_cons b (proj1_sig x) l (proj2_sig x)))).
    apply existTexist.
    apply subsetT_eq_compat; reflexivity.
  rewrite h5.
  reflexivity.
unfold f_list_pop in h2.
rewrite <- h4 in h2.
simpl in h2.
rewrite h2.
unfold p_'.
apply times_list_map_cons.



unfold times_list. simpl.


pose proof (proj2_sig (lfun_to_pair_list a)) as h5.
simpl in h5.
rewrite in_flat_map in h5.





assert (h4:(exist (fun x0 : Bt => b = x0 \/ In x0 l) (proj1_sig x)
                 (in_cons b (proj1_sig x) l (proj2_sig x))) = 
  (exist (fun x0 : Bt => b = x0 \/ In x0 l) (proj1_sig x)
                 (in_cons b (proj1_sig x) l (proj2_sig x)))).

rewrite <- h3.
    
    



Theorem normal_form : forall (l:list Bt), 
  forall (pf:B' = Gen (list_to_set l)), 
  (forall (p:BA'.Bt), atom p <-> (p <> 0) /\   
  (exists a:(twol l), times_list (map (fun pr:Bt*sign => eps' pr) 
    (proj1_sig (lfun_to_pair_list a))) = 
    (transfer_gen (transfer_eq _ _ pf p))) /\
  (forall (b:BA'.Bt), exists !(S:Ensemble BA'.Bt),   
   exists (pfs: Finite _ S),  
  (forall x:BA'.Bt, Ensembles.In S x -> atom x) /\    
  plus_list (proj1_sig (set_to_list S pfs)) = b)).
Proof.
intros l twol h1.
(*case (classic (exists a:{xl:Bt | In xl l} -> sign, True)).
intro h2.
destruct h2 as [a ?].*)
(*pose (list_power l signl) as twol.*)

(*pose ({xl:Bt | In xl l}->{s:sign | In s signl}) as twol.*)

(*assert (h2:forall a b:twol *)

pose (fun a:twol => 
  times_list (map (fun i:{xl:Bt | In xl l} => 
  eps (proj1_sig i)(proj1_sig (a i))) (map_sig l))) as
  p_.


pose
  (fun la:list (Bt*sign) => 
  times_list (map (fun (pr:Bt*sign) => eps' pr) la)) as p_'.

(*
  assert (h5:le (p_ (proj1_sig (lfun_to_pair_list a))) (eps' pr_ai)).
    unfold p_.
    apply le_times_finite_member.
    rewrite in_map_iff.    
    exists pr_ai. split; try reflexivity.
    pose proof (proj2_sig (lfun_to_pair_list a)) as h6.
    simpl in h6.
    induction l. simpl in h6. destruct h6 as [h6l | h6r].
    pose proof (proj2_sig i') as h7.
    simpl in h7. contradiction. contradiction.
    simpl in h6.
    rewrite in_flat_map in h6.*)

assert (h2:forall a b:twol, (a <> b) -> p_ a * p_ b = 0).
  intros a b h3.
  pose proof (distinct_functions_differ_at_point _ _ h3) as h4.
  destruct h4 as [i' h4].
  Print eps'.
(*  pose ((proj1_sig i'), (proj1_sig (a i'))) as pr_ai.
  pose ((proj1_sig i'), (proj1_sig (b i'))) as pr_bi.*)
  assert (h5:le (p_ a) (eps (proj1_sig i') (proj1_sig (a i')))).
    unfold p_.
    apply le_times_finite_member.
    rewrite in_map_iff.
    exists i'.
    split; [reflexivity | apply map_sig_in].
  assert (h6:le (p_ b) (eps (proj1_sig i') (proj1_sig (b i')))).
    unfold p_.
    apply le_times_finite_member.
    rewrite in_map_iff.
    exists i'.
    split; [reflexivity | apply map_sig_in].
  apply le_x_0.
  destruct (a i'); destruct (b i'). 
  destruct x; destruct x0. 
  elim h4.
  apply proj1_sig_injective. reflexivity.
  simpl in h5.
  simpl in h6.
  pose proof (mono_prod _ _ _ _ h5 h6) as h7.
  rewrite comp_prod in h7. assumption.
  simpl in h5. simpl in h6.
  pose proof (mono_prod _ _ _ _ h5 h6) as h7.
  rewrite (comm_prod _ (- proj1_sig i') (proj1_sig i')) in h7.
  rewrite comp_prod in h7. assumption.
  elim h4. apply proj1_sig_injective. reflexivity.

assert (h3:1 = plus_list (map p_' (list_power l signl))).
(*  assert (h4:forall (i:Bt), plus_list (map (eps i) signl) = 1).
    intro i.
    unfold signl. simpl. rewrite zero_sum. apply comp_sum.*)
  assert (h5:1 = times_plus_list_of_lists (map (fun i => (map (eps i) signl)) l)).
    unfold times_plus_list_of_lists. 
    rewrite map_map.
    simpl.
    assert (h6:(fun x : Bt => x + (- x + 0)) = (fun i:Bt => 1)).
      apply functional_extensionality.
      intro x. rewrite zero_sum. auto.
    rewrite h6.
    rewrite one_times_list. reflexivity.  
  rewrite complete_dist_list_times in h5.
  unfold plus_times_list_of_lists in h5.
  unfold p_'.
  unfold list_of_lists_seqs in h5. simpl.
*)


End SA.



(*Examples*)
Variable fos:Field_of_Sets.
Definition psa_fos := psa (Xt fos).

(*BPF = Boolean algebra Power set algebra Field of sets*)
Declare Module BPF:BAt with Definition B:=psa_fos.
(*SPF = Subalgebra Power set algebra Field of sets *)
Module SPF := SA BPF.
Import SPF.
Import BPF.

Lemma fos_closed : A_closed (F fos).
constructor.
(* + *)
red.
intros A B.
destruct A.
destruct B. 
unfold Bplus_sub.
simpl.
apply (Union_closed fos); assumption.
(* * *)
red.
intros A B.
destruct A.
destruct B.
unfold Btimes_sub.
simpl.
apply (Int_closed fos); assumption.
(* 1 *)
red. simpl. apply full_in_F.
(* 0 *)
red. simpl. apply empty_in_F.
(* - *)
red.
intro A.
destruct A as [? h1].
unfold Bcomp_sub.
simpl.
apply (Comp_closed fos). assumption.
Qed.

Definition fos_sub_psa := SubAlg (F fos) fos_closed.

Variable Xt:Type.
Definition fin_cof_fos' := fin_cof_fos Xt.
(*Definition psa_fin_cof := psa Xt.*)


(*Boolean algebra Powerset algebra Finite Cofinite *)
Declare Module BPFC:BAt with Definition B:=(psa Xt).
Module SPFC := SA BPFC.
Import SPFC.
Import BPFC.

Lemma fin_cof_closed : A_closed (F_fc Xt).
constructor.
(* + *)
red.
intros A B.
destruct A.
destruct B. 
unfold Bplus_sub.
simpl.
apply (Union_closed_fc); assumption.
(* * *)
red.
intros A B.
destruct A.
destruct B.
unfold Btimes_sub.
simpl.
apply (Int_closed_fc); assumption.
(* 1 *)
red. simpl. apply full_in_F_fc.
(* 0 *)
red. simpl. apply empty_in_F_fc.
(* - *)
red.
intro A.
destruct A.
unfold Bcomp_sub.
simpl.
apply (Comp_closed_fc); assumption.
Qed.

Definition fin_cof_sub_psa := SubAlg (F_fc Xt) fin_cof_closed. 

(* ". . . consider the field A of finite and cofinite subsets of
a set X. We shall prove that it is generated in P(X) by the set E of one-
element (that is, singleton) subsets of X." *)

Lemma fin_cof_gen_singletons : Gen (SingF (Full_set Xt)) = fin_cof_sub_psa.
unfold fin_cof_sub_psa.
unfold Gen.
pose (SA_cont_A (SingF (Full_set Xt))) as F.
unfold SingF in F.
pose (int_saf_A F) as S.
simpl.
pose fin_cof_closed as h1.
pose proof (P_c _ h1) as h2.
assert (h3:Included (SingF (Full_set Xt)) (F_fc Xt)).
  red.
  intros A h4.
  inversion h4 as [x ? ? h5].
  rewrite h5.
  pose proof (Singleton_is_finite _ x) as h6.
  constructor. left; assumption.
assert (h4:int_saf_A (SA_cont_A (SingF (Full_set Xt))) = 
  (F_fc Xt)).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red.
  intros sing h5.
  inversion h5 as [? h6].
  apply h6.
  constructor.
  split.
    (* left *)  
    red.
    intros sing'.
    unfold Bt in sing'.
    simpl in sing'.
    intro h7.
    inversion h7 as [y ? ? h8].
    rewrite h8.
    constructor.
    left.
    apply Singleton_is_finite.
    (* right *) assumption.
  (* >= *)
  red.
  intros S' h4.
  pose proof (union_singF_eq S') as h5.
  inversion h4 as [h6].
  assert (hfin : forall fin:Bt, In (F_fc Xt) fin -> Finite _ fin
    -> In S fin).
    intros fin h4' h6'.
    pose proof (union_singF_eq fin) as h5'. 
  induction h6' as [|fin' h7 h8 x h9].
  (* Empty *)
    constructor.
    intros E h6''.
    inversion h6'' as [h7].
    destruct h7 as [? h8].  
    pose proof (Z_c _ h8) as h9.
    red in h9.
    simpl in h9.
    assumption.
  (*Add*)
  unfold SingF in h5'.
  pose proof (Im_add _ _ fin' x (Singleton (U:=Xt))) as h10.
  unfold SingF.
  rewrite h10 in h5'.
  pose proof (family_union_add 
    (Im fin' (Singleton (U:=Xt))) (Singleton x)) as h11.
  rewrite h11 in h5'.
  rewrite h5'.
  pose proof (union_singF_eq fin') as h12.
  unfold SingF in h12.
  assert (h13:In (F_fc Xt) fin').
    constructor.
    left; assumption.
  pose proof (h8 h13 h12) as h14.
  rewrite <- h12.
  pose (Im (Full_set Xt) (Singleton (U:=Xt))) as J.
  assert (h15:In (int_saf_A (SA_cont_A J)) (Singleton x)).
    assert (h16:In J (Singleton x)).
      unfold J.
      apply Im_intro with x. apply Full_intro. trivial. 
      unfold SA_cont_A in F.
    pose (saf_A F) as EF.
    unfold Bt in EF.
    simpl in EF.
    unfold int_saf_A.
    assert (h17:forall (G:Family Xt), In EF G -> Included J G).
      intros G h18.
      inversion h18 as [h19].
      destruct h19 as [h19l].
      unfold J.
      assumption.
    constructor.
    unfold Bt. simpl.
    intros G h18.
    unfold EF in h17.
    pose proof (h17 G h18) as h19.
    auto with sets.

  pose proof (saf_closed F) as h20.    
  constructor.
  intros S0 h21.
  pose proof (h20 S0 h21) as h22.
  pose proof (P_c _ h22) as h23.
  red in h23.
  unfold SubBtype in h23.  unfold Bplus_sub in h23.
  unfold Bt in h23.
  simpl in h23.
  assert (h24: In S0 (Singleton x)).
    inversion h15 as [? h16].
    apply (h16 S0 h21).
  assert (h25: In S0 fin').
    inversion h14.
    apply (H S0 h21).
    apply (h23 (exist _ (Singleton x) h24) (exist _ fin' h25)).
destruct h6 as [h6l | h6r].
  (*h6l*)
  apply hfin; assumption.
  (*h6r*)
  assert (h7:In (F_fc Xt) (Complement S')).
    constructor.
    left; assumption.
  pose proof (hfin _ h7 h6r) as h8.
  pose proof (int_saf_A_closed F) as h9.
  pose proof (C_c _ h9) as h10.
  red in h10.
  unfold SubBtype in h10. unfold Bcomp_sub in h10.
  unfold Bt in h10. simpl in h10.
  pose proof (h10 (exist _ (Complement S') h8)) as h11.
  simpl in h11.
  pose proof (Complement_Complement _ S') as h12.
  rewrite <- h12.
  assumption.
pose proof (subsetT_eq_compat _ _ _ _
  (int_saf_A_closed (SA_cont_A (SingF (Full_set Xt))))
  fin_cof_closed h4) as h5.
dependent rewrite -> h5. reflexivity.
Qed.


(* *)
