(* Copyright (C) 2014, Daniel Wyckoff, except for the portions so labeleled
which I got from Daniel Schepler*)



(*		   GNU LESSER GENERAL PUBLIC LICENSE
                       Version 3, 29 June 2007

 Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>
 Everyone is permitted to copy and distribute verbatim copies
 of this license document, but changing it is not allowed.


  This version of the GNU Lesser General Public License incorporates
the terms and conditions of version 3 of the GNU General Public
License, supplemented by the additional permissions listed below.

  0. Additional Definitions. 

  As used herein, "this License" refers to version 3 of the GNU Lesser
General Public License, and the "GNU GPL" refers to version 3 of the GNU
General Public License.

  "The Library" refers to a covered work governed by this License,
other than an Application or a Combined Work as defined below.

  An "Application" is any work that makes use of an interface provided
by the Library, but which is not otherwise based on the Library.
Defining a subclass of a class defined by the Library is deemed a mode
of using an interface provided by the Library.

  A "Combined Work" is a work produced by combining or linking an
Application with the Library.  The particular version of the Library
with which the Combined Work was made is also called the "Linked
Version".

  The "Minimal Corresponding Source" for a Combined Work means the
Corresponding Source for the Combined Work, excluding any source code
for portions of the Combined Work that, considered in isolation, are
based on the Application, and not on the Linked Version.

  The "Corresponding Application Code" for a Combined Work means the
object code and/or source code for the Application, including any data
and utility programs needed for reproducing the Combined Work from the
Application, but excluding the System Libraries of the Combined Work.

  1. Exception to Section 3 of the GNU GPL.

  You may convey a covered work under sections 3 and 4 of this License
without being bound by section 3 of the GNU GPL.

  2. Conveying Modified Versions.

  If you modify a copy of the Library, and, in your modifications, a
facility refers to a function or data to be supplied by an Application
that uses the facility (other than as an argument passed when the
facility is invoked), then you may convey a copy of the modified
version:

   a) under this License, provided that you make a good faith effort to
   ensure that, in the event an Application does not supply the
   function or data, the facility still operates, and performs
   whatever part of its purpose remains meaningful, or

   b) under the GNU GPL, with none of the additional permissions of
   this License applicable to that copy.

  3. Object Code Incorporating Material from Library Header Files.

  The object code form of an Application may incorporate material from
a header file that is part of the Library.  You may convey such object
code under terms of your choice, provided that, if the incorporated
material is not limited to numerical parameters, data structure
layouts and accessors, or small macros, inline functions and templates
(ten or fewer lines in length), you do both of the following:

   a) Give prominent notice with each copy of the object code that the
   Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the object code with a copy of the GNU GPL and this license
   document.

  4. Combined Works.

  You may convey a Combined Work under terms of your choice that,
taken together, effectively do not restrict modification of the
portions of the Library contained in the Combined Work and reverse
engineering for debugging such modifications, if you also do each of
the following:

   a) Give prominent notice with each copy of the Combined Work that
   the Library is used in it and that the Library and its use are
   covered by this License.

   b) Accompany the Combined Work with a copy of the GNU GPL and this license
   document.

   c) For a Combined Work that displays copyright notices during
   execution, include the copyright notice for the Library among
   these notices, as well as a reference directing the user to the
   copies of the GNU GPL and this license document.

   d) Do one of the following:

       0) Convey the Minimal Corresponding Source under the terms of this
       License, and the Corresponding Application Code in a form
       suitable for, and under terms that permit, the user to
       recombine or relink the Application with a modified version of
       the Linked Version to produce a modified Combined Work, in the
       manner specified by section 6 of the GNU GPL for conveying
       Corresponding Source.

       1) Use a suitable shared library mechanism for linking with the
       Library.  A suitable mechanism is one that (a) uses at run time
       a copy of the Library already present on the user's computer
       system, and (b) will operate properly with a modified version
       of the Library that is interface-compatible with the Linked
       Version. 

   e) Provide Installation Information, but only if you would otherwise
   be required to provide such information under section 6 of the
   GNU GPL, and only to the extent that such information is
   necessary to install and execute a modified version of the
   Combined Work produced by recombining or relinking the
   Application with a modified version of the Linked Version. (If
   you use option 4d0, the Installation Information must accompany
   the Minimal Corresponding Source and Corresponding Application
   Code. If you use option 4d1, you must provide the Installation
   Information in the manner specified by section 6 of the GNU GPL
   for conveying Corresponding Source.)

  5. Combined Libraries.

  You may place library facilities that are a work based on the
Library side by side in a single library together with other library
facilities that are not Applications and are not covered by this
License, and convey such a combined library under terms of your
choice, if you do both of the following:

   a) Accompany the combined library with a copy of the same work based
   on the Library, uncombined with any other library facilities,
   conveyed under the terms of this License.

   b) Give prominent notice with the combined library that part of it
   is a work based on the Library, and explaining where to find the
   accompanying uncombined form of the same work.

  6. Revised Versions of the GNU Lesser General Public License.

  The Free Software Foundation may publish revised and/or new versions
of the GNU Lesser General Public License from time to time. Such new
versions will be similar in spirit to the present version, but may
differ in detail to address new problems or concerns.

  Each version is given a distinguishing version number. If the
Library as you received it specifies that a certain numbered version
of the GNU Lesser General Public License "or any later version"
applies to it, you have the option of following the terms and
conditions either of that published version or of any later version
published by the Free Software Foundation. If the Library as you
received it does not specify a version number of the GNU Lesser
General Public License, you may choose any version of the GNU Lesser
General Public License ever published by the Free Software Foundation.

  If the Library as you received it specifies that a proxy can decide
whether future versions of the GNU Lesser General Public License shall
apply, that proxy's public statement of acceptance of any version is
permanent authorization for you to choose that version for the
Library.*)



Require Export Ensembles.
Require Export EnsemblesImplicit.
Require Import FunctionProperties.
Require Export Finite_sets.
Require Import Image.
Require Import ImageImplicit.
Require Import DecidableDec.
Require Import SetUtilities.
Require Import ClassicalChoice.
Require Import Arith.
Require Import CSB.
Require Import LogicUtilities.
Require Import Basics.
Require Import ArithUtilities.
Require Import Omega. 
Require Import Description.
Require Import ClassicalEpsilon.

(*Many of these functions are copied and pasted from Daniel Schepler's
  "Zorn's Lemma" user contribution.  Those which are are commented 
  "Schepler."  Those which aren't are commented "Wyckoff"*)
Section FiniteTypes.

(*Schepler*)
Inductive FiniteT : Type -> Prop :=
  | empty_finite: FiniteT False
  | add_finite: forall T:Type, FiniteT T -> FiniteT (option T)
  | bij_finite: forall (X Y:Type) (f:X->Y), FiniteT X ->
    invertible f -> FiniteT Y.

(*Schepler*)
Lemma True_finite: FiniteT True.
Proof.
apply bij_finite with (option False)
  (fun _ => I).
constructor; constructor.
exists (True_rect None).
destruct x as [[]|].
remember (True_rect (@None False) I) as LHS.
destruct LHS as [[]|].
reflexivity.

exact (fun y:True => match y with
  | I => refl_equal I
  end).
Qed.

(*Schepler*)
Lemma finite_dec_exists: forall (X:Type) (P:X->Prop),
  FiniteT X -> (forall x:X, {P x} + {~ P x}) ->
  { exists x:X, P x } + { forall x:X, ~ P x }.
Proof.
intros.
apply exclusive_dec.
red; intro.
destruct H0.
destruct H0.
contradiction (H1 x).

revert P X0.
induction H.
right.
destruct x.

intros.
case (IHFiniteT (fun x:T => P (Some x))
  (fun x:T => X0 (Some x))).
left.
destruct H0.
exists (Some x).
assumption.
intro.
case (X0 None).
left.
exists None.
assumption.
right.
destruct x.
apply H0.
assumption.
destruct H0.
intros.
case (IHFiniteT (fun x:X => P (f x))
  (fun x:X => X0 (f x))).
left.
destruct H2.
exists (f x).
assumption.
right.
intro.
rewrite <- H1 with x.
apply H2.
Qed.

Lemma finite_dec_forall: forall (X:Type) (P:X->Prop),
  FiniteT X -> (forall x:X, { P x } + { ~ P x }) ->
  { forall x:X, P x } + { exists x:X, ~ P x }.
Proof.
intros.
apply exclusive_dec.
intuition.
destruct H2.
contradiction (H1 x).

revert P X0.
induction H.
left.
destruct x.
intros.
case (IHFiniteT (fun x:T => P (Some x))
  (fun x:T => X0 (Some x))).
intro.
case (X0 None).
left.
destruct x.
apply H0.
assumption.
right.
exists None.
assumption.
right.
destruct H0.
exists (Some x).
assumption.

intros.
destruct H0.
case (IHFiniteT (fun x:X => P (f x))
  (fun x:X => X0 (f x))).
left.
intro y.
rewrite <- H1.
apply H2.
right.
destruct H2.
exists (f x).
assumption.
Qed.


(*Schepler*)
Lemma FiniteT_img: forall (X Y:Type) (f:X->Y),
  FiniteT X -> (forall y1 y2:Y, y1=y2 \/ y1<>y2) ->
  Finite _ (Im (Full_set _) f).
Proof.
intros.
induction H.
assert (Im (Full_set _) f = Empty_set _).
apply Extensionality_Ensembles.
red; split.
red; intros.
destruct H.
destruct x.
auto with sets.
rewrite H.
constructor.

assert ({exists x:T, f (Some x) = f None} +
        {forall x:T, f (Some x) <> f None}).
apply finite_dec_exists.
assumption.
intro.
apply decidable_dec.
apply H0.
case H1.
intro.
pose (g := fun (x:T) => f (Some x)).
assert (Im (Full_set _) f =
        Im (Full_set _) g).
apply Extensionality_Ensembles.
red; split.
red; intros.
destruct H2.
destruct x.
exists t.
constructor.
assumption.
destruct e.
exists x.
constructor.
transitivity (f None).
assumption.
symmetry; assumption.
red; intros.

destruct H2.
exists (Some x).
constructor.
assumption.
rewrite H2.
apply IHFiniteT.

intros.
pose (g := fun x:T => f (Some x)).
assert (Im (Full_set _) f =
  Add (Im (Full_set _) g) (f None)).
apply Extensionality_Ensembles.
red; split.
red; intros.
destruct H2.
destruct x.
left.
exists t.
constructor.
assumption.
right.
auto with sets.
red; intros.
destruct H2.
destruct H2.
exists (Some x).
constructor.
assumption.
destruct H2.
exists None.
constructor.
reflexivity.
rewrite H2.
constructor.
apply IHFiniteT.
red; intro.
destruct H3.
contradiction (n x).
symmetry; assumption.

pose (g := fun (x:X) => f (f0 x)).
assert (Im (Full_set _) f = Im (Full_set _) g).
apply Extensionality_Ensembles.
red; split.
red; intros.
destruct H2.
destruct H1.
rewrite H3.
rewrite <- H4 with x.
exists (g0 x).
constructor.
unfold g.
reflexivity.
red; intros.
destruct H2.
exists (f0 x).
constructor.
assumption.

rewrite H2.
apply IHFiniteT.
Qed.

(*Schepler*)
Lemma finite_subtype: forall (X:Type) (P:X->Prop),
  FiniteT X -> (forall x:X, P x \/ ~ P x) ->
  FiniteT {x:X | P x}.
Proof.
intros.
induction H.
apply bij_finite with False (False_rect _).
constructor.
exists (@proj1_sig _ _).
destruct x.
intro s; destruct s; destruct x.

destruct (H0 None).
pose (g := fun (x:option {x:T | P (Some x)}) =>
  match x return {x:option T | P x} with
  | Some (exist x0 i) => exist (fun x:option T => P x) (Some x0) i
  | None => exist (fun x:option T => P x) None H1
  end).
apply bij_finite with _ g.
apply add_finite.
apply IHFiniteT.
intro; apply H0.
pose (ginv := fun (s:{x0:option T | P x0}) =>
                       match s return option {x:T | P (Some x)} with
  | exist (Some x0) i => Some (exist (fun y:T => P (Some y)) x0 i)
  | exist None _ => None
  end).
exists ginv.
destruct x as [[x0]|].
simpl.
reflexivity.
simpl.
reflexivity.
destruct y as [[x0|]].
simpl.
reflexivity.
simpl.
destruct (proof_irrelevance _ H1 p).
reflexivity.

pose (g := fun (x:{x:T | P (Some x)}) =>
  match x return {x:option T | P x} with
    | exist x0 i => exist (fun x:option T => P x) (Some x0) i
  end).
apply bij_finite with _ g.
apply IHFiniteT.
intro; apply H0.
pose (ginv := fun s:{x0:option T | P x0} =>
  match s return {x:T | P (Some x)} with
    | exist (Some x0) i => exist (fun x:T => P (Some x)) x0 i
    | exist None i => False_rect _ (H1 i)
  end).
exists ginv.
destruct x; simpl.
reflexivity.
destruct y as [[x0|]].
simpl.
reflexivity.
contradiction H1.

pose (g := fun (x:{x:X | P (f x)}) =>
  match x with
  | exist x0 i => exist (fun x:Y => P x) (f x0) i
        end).
apply bij_finite with _ g.
apply IHFiniteT.
intro; apply H0.
destruct H1.
assert (forall y:Y, P y -> P (f (g0 y))).
intros; rewrite H2; assumption.
pose (ginv := fun (y:{y:Y | P y}) =>
  match y with
  | exist y0 i => exist (fun x:X => P (f x)) (g0 y0) (H3 y0 i)
  end).
exists ginv.
destruct x; simpl.
generalize (H3 (f x) p).
rewrite H1.
intro; destruct (proof_irrelevance _ p p0).
reflexivity.

destruct y; simpl.
generalize (H3 x p).
rewrite H2.
intro; destruct (proof_irrelevance _ p p0).
reflexivity.
Qed.

(*Wyckoff*)                   
Lemma FiniteT_Finite : forall (T:Type), FiniteT T -> 
  (forall (S:Ensemble T), Finite _ S).
intros T h1 S.
assert (h2: forall x:T, In S x \/ ~In S x).
  intro. tauto.
pose proof (finite_subtype T _ h1 h2) as h3.
assert (h4: forall y1 y2 : T, y1 = y2 \/ y1 <> y2).
  intros. tauto.
pose proof 
  (FiniteT_img {x : T | In S x} T (@proj1_sig T _) h3 h4) as h5.
assert (h6: (Im (Full_set {x : T | In S x}) (proj1_sig (P:=fun x : T => In S x)))
  = S).
  apply Extensionality_Ensembles. red. unfold Included.
  split.
  (*left*)
  intros x h7.
  inversion h7 as [y h8 z h9].
  pose proof (proj2_sig y) as h10.
  rewrite <- h9 in h10.
  assumption.
  (*right*)
  intros x h11.
  apply Im_intro with (exist _ x h11).
  apply Full_intro.
  simpl. reflexivity.
rewrite h6 in h5.
assumption.
Qed.

(*Wyckoff*)
Lemma FiniteT_sig_Finite : forall {T:Type} (S:Ensemble T), 
  FiniteT {x:T | In S x} -> Finite _ S.
intros T S h1.
pose proof (FiniteT_img _ _ (@proj1_sig _ _) h1) as h2.
assert (h3: forall y1 y2 : T, y1 = y2 \/ y1 <> y2).
  intros. apply classic.
apply h2 in h3.
assert (h4:(Im (Full_set {x : T | In S x}) 
  (proj1_sig (P:=fun x : T => In S x))) = 
    S).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h4.
  inversion h4 as [? ? ? h5].
  rewrite h5.
  apply proj2_sig.
  (* >= *)
  red. intros x h4.
  eapply Im_intro.
  apply (Full_intro _ (exist _ x h4)).
  reflexivity.
rewrite h4 in h3; assumption.
Qed.

(*Schepler*)
Lemma Finite_ens_type: forall {X:Type} (S:Ensemble X),
  Finite _ S -> FiniteT { x:X | In S x }.
Proof.
intros.
induction H.
apply bij_finite with False (False_rect _).
constructor.
assert (g:{x:X | In (Empty_set _) x}->False).
intro.
destruct X0.
destruct i.
exists g.
destruct x.
destruct y.
destruct g.

assert (Included A (Add A x)).
auto with sets.
assert (In (Add A x) x).
auto with sets.
pose (g := fun (y: option {x:X | In A x}) =>
  match y return {x0:X | In (Add A x) x0} with
                                          | Some (exist y0 i) => exist (fun x2:X => In (Add A x) x2) y0 (H1 y0 i)
  | None => exist (fun x2:X => In (Add A x) x2) x H2
  end).
apply bij_finite with _ g.
apply add_finite.
assumption.

assert (h:forall x0:X, In (Add A x) x0 ->
  { In A x0 } + { x0 = x }).
intros; apply exclusive_dec.
intuition.
destruct H6; auto.
destruct H3.
left; assumption.
right; destruct H3; reflexivity.

pose (ginv := fun s:{x0:X | In (Add A x) x0} =>
  match s return option {x:X | In A x} with
  | exist x0 i => match (h x0 i) with
                  | left iA => Some (exist _ x0 iA)
                  | right _ => None
   end
  end).
exists ginv.
intro; destruct x0.
destruct s.
simpl.
remember (h x0 (H1 x0 i)) as sum; destruct sum.
destruct (proof_irrelevance _ i i0).
reflexivity.
contradiction H0.
rewrite <- e; assumption.
simpl.
remember (h x H2) as sum; destruct sum.
contradiction H0.
reflexivity.
                   
intro.
unfold ginv.
destruct y.
destruct (h x0 i).
simpl.
generalize (H1 x0 i0); intro.
destruct (proof_irrelevance _ i i1).
reflexivity.
simpl.
destruct e.
destruct (proof_irrelevance _ H2 i).
reflexivity.
Qed.

(*Wyckoff*)
Lemma FiniteT_Full : forall T:Type, 
  FiniteT T <->  FiniteT {x:T | In (Full_set T) x}.
intros T.
pose (fun (a:T) => exist _ a (Full_intro _ a)) as f.
assert (h2:invertible f).
  apply bijective_impl_invertible.
  red. split.
  (*inj*)
  red.
  intros x1 x2 h3.
  unfold f in h3.
  assert (h4: x1 = proj1_sig 
    (exist (In (Full_set T)) x1 (Full_intro T x1))).
    simpl. reflexivity.
  assert (h5: x2 = proj1_sig
   (exist (In (Full_set T)) x2 (Full_intro T x2))).
    simpl. reflexivity.
  congruence.
  (*surj*)
  red.
  intro y.
  exists (proj1_sig y).
  destruct y as [z h1].
  unfold f.
  simpl.
  pose proof 
    (proof_irrelevance _ h1 (Full_intro T z)) as h2.
  rewrite h2. 
  reflexivity.

pose (fun (a:{x:T | In (Full_set T) x}) => proj1_sig a) as f'.
assert (h3:invertible f').
  apply bijective_impl_invertible.
  red. split.
  (*inj*)
  red.
  intros x1 x2 h3.
  unfold f' in h3.
  apply proj1_sig_injective; trivial.
  (*surj*)
  red.
  intro y.
  exists (exist _ y (Full_intro _ y)).
  unfold f'.
  simpl.
  reflexivity.

split.
intro h1.  apply (bij_finite _ _ f);  trivial.
intro h1.  apply (bij_finite _ _ f'); trivial.
Qed.

(*Wyckoff*)
Lemma Finite_FiniteT : forall (T:Type), 
  Finite _ (Full_set T) -> FiniteT T.
intros T h1.
rewrite FiniteT_Full.
apply Finite_ens_type; trivial.
Qed.

(*Schepler*)
Lemma finite_eq_dec: forall X:Type, FiniteT X ->
  forall x y:X, {x=y} + {x<>y}.
Proof.
intros.
apply decidable_dec.
induction H.
destruct x.
decide equality.

destruct H0.
case (IHFiniteT (g x) (g y)).
left.
rewrite <- H1.
rewrite <- H1 with x.
rewrite H2.
reflexivity.
right.
contradict H2.
rewrite H2.
reflexivity.
Qed.

(*Schepler*)
Lemma finite_inj_surj: forall (X:Type) (f:X->X),
  FiniteT X -> FunctionProperties.injective f -> surjective f.
Proof.
intros.
induction H.
red.
destruct y.

remember (f None) as f0; destruct f0 as [a|].
assert (forall x:T, f (Some x) <> Some a).
unfold not; intros.
assert (Some x = None).
apply H0.
congruence.
discriminate H2.
pose (g := fun x:T => match f (Some x) with
  | Some y => y
  | None => a
end).
assert (surjective g).
apply IHFiniteT.
red; intros.
remember (f (Some x1)) as fx1; destruct fx1;
remember (f (Some x2)) as fx2; destruct fx2.
unfold g in H2.
rewrite <- Heqfx1 in H2; rewrite <- Heqfx2 in H2.
destruct H2; assert (f (Some x1) = f (Some x2)).
congruence.
apply H0 in H2.
injection H2; trivial.

unfold g in H2; rewrite <- Heqfx1 in H2; rewrite <- Heqfx2 in H2.
destruct H2.
contradiction (H1 x1).
symmetry; assumption.

unfold g in H2; rewrite <- Heqfx1 in H2; rewrite <- Heqfx2 in H2.
destruct H2.
contradiction (H1 x2).
symmetry; assumption.
assert (Some x1 = Some x2).
apply H0.
congruence.
injection H3; trivial.

red; intro.
destruct y.
case (finite_eq_dec _ H t a).
exists None.
congruence.
destruct (H2 t).
exists (Some x).
unfold g in H3.
destruct (f (Some x)).
congruence.
contradiction n.
symmetry; assumption.
destruct (H2 a).
exists (Some x).
unfold g in H3.
remember (f (Some x)) as fx; destruct fx.
destruct H3.
contradiction (H1 x).
symmetry; assumption.
reflexivity.

assert (forall x:T, { y:T | f (Some x) = Some y }).
intros.
remember (f (Some x)) as fx; destruct fx.
exists t; reflexivity.
assert (Some x = None).
apply H0.
congruence.
discriminate H1.
pose (g := fun x:T => proj1_sig (X x)).
assert (surjective g).
apply IHFiniteT.
red; intros.
unfold g in H1.
repeat destruct X in H1.
simpl in H1.
assert (Some x1 = Some x2).
apply H0.
congruence.
injection H2; trivial.

red; intro.
destruct y.
destruct (H1 t).
unfold g in H2; destruct X in H2.
simpl in H2.
exists (Some x).
congruence.
exists None.
symmetry; assumption.

destruct H1.

pose (f' := fun (x:X) => g (f (f0 x))).
assert (surjective f').
apply IHFiniteT.
red; intros.
unfold f' in H3.
assert (f (f0 x1) = f (f0 x2)).
congruence.
apply H0 in H4.
congruence.

red; intro.
destruct (H3 (g y)).
unfold f' in H4.
exists (f0 x).
congruence.
Qed.

(*Wyckoff*)
Lemma FiniteT_Finite_Union : forall {It T:Type} 
  (F:IndexedFamily It T),   (forall i:It, Finite _ (F i)) 
  -> FiniteT It -> Finite _ (IndexedUnion F).
Proof.
intros It T F h1 h2.
induction h2.
(*Empty*)
assert (h3:(IndexedUnion F) = Empty_set _ ).
  apply Extensionality_Ensembles.
  red. split. 
  (* <= *)
  red. intros x h4.
  inversion h4.
  contradiction.
  (* >= *)
  red. intros; contradiction.
rewrite h3. constructor.
(*Add*)
rename T0 into It.
pose (fun (i:It) => F (Some i)) as f.
assert (h4:(IndexedUnion F) = Union (F None) (IndexedUnion f)).
  apply Extensionality_Ensembles.
  red. split.
  (*<=*)
  red. intros x h5.   
  inversion h5 as [i ? h6].
  destruct i.
    (*Some*)
    right.
    unfold f.
    apply indexed_union_intro with i. assumption.
    (*None*)
    constructor. assumption.
  (*>=*)
  red. intros x h5.
  pose proof (Union_inv _ _ _ x h5) as h6.
  case h6 as [h7 | h8].
    apply indexed_union_intro with None; assumption.
  inversion h8.
    apply indexed_union_intro with (Some i); assumption.
assert (h5: forall i:It, Finite _ (f i)).
  intro i.
  apply (h1 (Some i)).
pose proof (IHh2 f h5) as h9.
assert (h10: Finite _ (F None)).
  apply (h1 None).
rewrite h4.
apply Union_preserves_Finite; assumption.
(*Bij*)
rename Y into It.
rename H into h3.
pose (fun (x:X) => F (f x)) as Ff.
assert (h4: forall x:X, Finite _ (Ff x)).
  intro x.
  apply (h1 (f x)).
pose proof (IHh2 Ff h4) as h5.
assert (h6:IndexedUnion Ff = IndexedUnion F).
  apply Extensionality_Ensembles.
  red. split.
  (*<=*)
  red.
  intros x h7.
  inversion h7 as [? ? h8].
  unfold Ff in h8.
  apply indexed_union_intro with (f i); assumption.
  (*>=*)  
  pose proof (invertible_impl_bijective _ h3) as h6.
  red. intros t h7.
  inversion h7 as [i y h8].
  unfold bijective in h6.
  destruct h6 as [h9 h10].
  red in h10.
  pose proof (h10 i) as h11.
  elim h11. intros x h12.
  rewrite <- h12 in h8.
  unfold Ff.
  apply indexed_union_intro with x; assumption.
rewrite <- h6.
assumption.
Qed.

(*Schepler*)  
Lemma finite_dep_choice: forall (A:Type) (B:forall x:A, Type)
  (R:forall x:A, B x->Prop),
  FiniteT A -> (forall x:A, exists y:B x, R x y) ->
  exists f:(forall x:A, B x), forall x:A, R x (f x).
Proof.
intros.
revert B R H0.
induction H.
intros.
exists (fun x:False => False_rect (B x) x).
destruct x.
intros.
pose proof (IHFiniteT (fun x:T => B (Some x))
  (fun x:T => R (Some x))
  (fun x:T => H0 (Some x))).
destruct H1.
pose proof (H0 None).
destruct H2.
exists (fun y:option T =>
  match y return (B y) with
  | Some y0 => x y0
  | None => x0
  end).
destruct x1.
apply H1.
assumption.

intros.
destruct H0.
pose proof (IHFiniteT (fun x:X => B (f x))
  (fun x:X => R (f x))
  (fun x:X => H1 (f x))).
destruct H3.
pose (f0 := fun y:Y => x (g y)).
pose (conv := fun (y:Y) (a:B (f (g y))) =>
  eq_rect (f (g y)) B a y (H2 y)).

exists (fun y:Y => conv y (x (g y))).
intro.
unfold conv; simpl.
generalize (H2 x0).
pattern x0 at 2 3 6.
rewrite <- H2.
intro.
rewrite <- eq_rect_eq.
apply H3.
Qed.

(*Schepler*)
Lemma finite_choice : forall (A B:Type) (R:A->B->Prop),
  FiniteT A -> (forall x:A, exists y:B, R x y) ->
  exists f:A->B, forall x:A, R x (f x).
Proof.
intros.
apply finite_dep_choice.
assumption.
assumption.
Qed.
End FiniteTypes.

(*Whole section copied and pasted from file of same name in Schepler's Zorn's Lemma.*)
Section InfiniteTypes.

(*Schepler*)
Lemma finite_nat_initial_segment: forall n:nat,
  FiniteT { m:nat | m < n }.
Proof.
intros.
apply Finite_ens_type.
rewrite <- characteristic_function_to_ensemble_is_identity.
induction n.
assert ([x:nat | S x <= 0] = (Empty_set _)).
apply Extensionality_Ensembles; split; auto with sets.
red; intros.
destruct H.
contradict H.
auto with arith.
rewrite H; constructor.

assert ([x:nat | S x <= S n] = Add [x:nat | x < n] n).
apply Extensionality_Ensembles; split.
red; intros.
destruct H.
assert (x <= n); auto with arith.
apply le_lt_or_eq in H0.
case H0.
left; constructor; trivial.
right; auto with sets.
red; intros.
case H.
intros.
destruct H0; constructor.
auto with arith.
intros.
destruct H0.
constructor.
auto with arith.
rewrite H; constructor; trivial.
red; intro.
destruct H0.
contradict H0.
auto with arith.
Qed.

(*Schepler*)
Lemma nat_infinite: ~ FiniteT nat.
Proof.
red; intro.
assert (surjective S).
apply finite_inj_surj; trivial.
red; intros.
injection H0; trivial.

destruct (H0 0).
discriminate H1.
Qed.

(*Schepler*)
Lemma infinite_nat_inj: forall X:Type, ~ FiniteT X ->
  exists f:nat->X, FunctionProperties.injective f.
Proof.
intros.
assert (inhabited (forall S:Ensemble X, Finite _ S ->
  { x:X | ~ In S x})).
pose proof (choice (fun (x:{S:Ensemble X | Finite _ S}) (y:X) =>
  ~ In (proj1_sig x) y)).
simpl in H0.
match type of H0 with | ?A -> ?B => assert B end.
apply H0.
intros.
apply NNPP.
red; intro.
pose proof (not_ex_not_all _ _ H1); clear H1.
destruct x.
assert (x = (Full_set _)).
apply Extensionality_Ensembles; red; split; auto with sets.
intro x0; constructor.
symmetry in H1; destruct H1.
contradiction H.
clear H2.
apply bij_finite with (f:=@proj1_sig _ (fun x:X => In (Full_set _) x)).
apply Finite_ens_type; assumption.
exists (fun x:X => exist _ x (Full_intro _ x)).
destruct x; simpl.
generalize (Full_intro X x).
intro i0; destruct (proof_irrelevance _ i i0); trivial.
trivial.
clear H0.
destruct H1.
exists.
intros.
exists (x (exist _ S H1)).
exact (H0 (exist _ S H1)).
destruct H0.

assert (forall (n:nat) (g:forall m:nat, m<n -> X),
  { x:X | forall (m:nat) (Hlt:m<n), g m Hlt <> x }).
intros.
assert (Finite _ (fun x:X => exists m:nat, exists Hlt:m<n,
           g m Hlt = x)).
pose (h := fun x:{m:nat | m<n} =>
  g (proj1_sig x) (proj2_sig x)).

match goal with |- Finite X ?S => assert (S =
  Im (Full_set _) h) end.
apply Extensionality_Ensembles; red; split; red; intros.
destruct H0.
destruct H0.
exists (exist (fun m:nat => m < n) x0 x1).
constructor.
unfold h; simpl.
symmetry; assumption.
destruct H0.
destruct x.
unfold h in H1; simpl in H1.
exists x; exists l; symmetry; assumption.

rewrite H0; apply FiniteT_img.
apply finite_nat_initial_segment.
intros; apply classic.

destruct (X0 _ H0).
unfold In in n0.
exists x.
intros; red; intro.
contradiction n0; exists m; exists Hlt; exact H1.

pose (f := Fix lt_wf (fun n:nat => X)
  (fun (n:nat) (g:forall m:nat, m<n->X) => proj1_sig (X1 n g))).
simpl in f.
assert (forall n m:nat, m<n -> f m <> f n).
pose proof (Fix_eq lt_wf (fun n:nat => X)
  (fun (n:nat) (g:forall m:nat, m<n->X) => proj1_sig (X1 n g))).
fold f in H0.
simpl in H0.
match type of H0 with | ?A -> ?B => assert (B) end.
apply H0.
intros.
assert (f0 = g).
Require Import FunctionalExtensionality.
extensionality y; extensionality p; apply H1.
destruct H2; trivial.
intros.
pose proof (H1 n).
destruct X1 in H3.
simpl in H3.
destruct H3.
auto.

exists f.
red; intros m n ?.
destruct (lt_eq_lt_dec m n) as [[Hlt|Heq]|Hlt]; trivial.
contradiction (H0 n m).
contradiction (H0 m n).
symmetry; assumption.
Qed.


Lemma inf_comp_fin : forall {T:Type}, ~FiniteT T -> 
  forall U:Ensemble T, Finite _ (Ensembles.Complement U) -> 
    ~ Finite _ U.
intros T h0 U h1.
case (classic (Finite _ U)); trivial.
intro h2.
pose proof (Union_preserves_Finite T U (Ensembles.Complement U) h2 h1) as h3.
rewrite excl_middle_full in h3.
pose proof (Finite_ens_type (Full_set T) h3) as h4.
pose (fun (a:{x:T | In (Full_set T) x}) => proj1_sig a) as f.
assert (h5:invertible f).
  apply bijective_impl_invertible.
  red. split.
  (*left*)
  red.
  intros x1 x2 h6.
  unfold f in h6.
  apply proj1_sig_injective; trivial.
  (*right*)
  red.
  intro y.
  exists (exist _ y (Full_intro _ y)).
  unfold f.
  simpl.
  reflexivity.
pose proof (bij_finite _ _ f h4 h5). 
contradiction.
Qed.

End InfiniteTypes.


Section CountableT.

(*Schepler*)
Inductive CountableT (X:Type) : Prop :=
  | intro_nat_injection: forall f:X->nat, FunctionProperties.injective f -> CountableT X.

(*Schepler*)
Lemma CountableT_is_FiniteT_or_countably_infinite:
  forall X:Type, CountableT X ->
      {FiniteT X} + {exists f:X->nat, bijective f}.
Proof.
intros.
apply exclusive_dec.
red; intro.
destruct H0 as [? [f ?]].
contradiction nat_infinite.
apply bij_finite with _ f; trivial.
apply bijective_impl_invertible; trivial.

case (classic (FiniteT X)).
left; trivial.
right.
apply infinite_nat_inj in H0.
destruct H.
destruct H0 as [g].
apply CSB with f g; trivial.
Qed.

(*Schepler*)
Lemma countable_nat_product: CountableT (nat*nat).
Proof.
pose (sum_1_to_n := fix sum_1_to_n n:nat := match n with
  | O => O
  | S m => (sum_1_to_n m) + n
end).
exists (fun p:nat*nat => let (m,n):=p in
  (sum_1_to_n (m+n)) + n).
assert (forall m n:nat, m<n ->
  sum_1_to_n m + m < sum_1_to_n n).
intros.
induction H.
simpl.
auto with arith.
apply lt_trans with (sum_1_to_n m0).
assumption.
simpl.
assert (0 < S m0); auto with arith.
assert (sum_1_to_n m0 + 0 < sum_1_to_n m0 + S m0); auto with arith.
assert (sum_1_to_n m0 + 0 = sum_1_to_n m0); auto with arith.
rewrite H2 in H1; assumption.

red; intros.
destruct x1 as [x1 y1].
destruct x2 as [x2 y2].
Require Import Compare_dec.
case (lt_eq_lt_dec (x1+y1) (x2+y2)); intro.
case s; intro.
assert (sum_1_to_n (x1+y1) + y1 < sum_1_to_n (x2+y2) + y2).
apply le_lt_trans with (sum_1_to_n (x1+y1) + (x1+y1)).
assert (sum_1_to_n (x1+y1) + (x1+y1) =
  (sum_1_to_n (x1+y1) + y1) + x1).
Require Import ArithRing.
ring.
auto with arith.
apply lt_le_trans with (sum_1_to_n (x2+y2)).
apply H; trivial.
auto with arith.
rewrite H0 in H1.
contradict H1.
auto with arith.

assert (y1=y2).
rewrite e in H0.
Require Import Arith.
apply plus_reg_l in H0.
assumption.
f_equal; trivial.
rewrite H1 in e.
rewrite plus_comm in e.
rewrite (plus_comm x2 y2) in e.
apply plus_reg_l in e.
assumption.

assert (sum_1_to_n (x2+y2) + y2 < sum_1_to_n (x1+y1) + y1).
apply le_lt_trans with (sum_1_to_n (x2+y2) + (x2+y2)).
auto with arith.
apply lt_le_trans with (sum_1_to_n (x1+y1)); auto with arith.
rewrite H0 in H1.
contradict H1.
auto with arith.
Qed.

(*Schepler*)
Lemma countable_sum: forall X Y:Type,
  CountableT X -> CountableT Y -> CountableT (X+Y).
Proof.
intros.
destruct H as [f].
destruct H0 as [g].
destruct countable_nat_product as [h].
exists (fun s:X+Y => match s with
  | inl x => h (0, f x)
  | inr y => h (1, g y)
end).
red; intros s1 s2 ?.
destruct s1 as [x1|y1]; destruct s2 as [x2|y2];
  apply H1 in H2; try discriminate H2;
  intros; f_equal; (apply H || apply H0); injection H2; trivial.
Qed.

(*Schepler*)
Lemma countable_product: forall X Y:Type,
  CountableT X -> CountableT Y -> CountableT (X*Y).
Proof.
intros.
destruct H as [f].
destruct H0 as [g].
pose (fg := fun (p:X*Y) => let (x,y):=p in (f x, g y)).
destruct countable_nat_product as [h].
exists (fun p:X*Y => h (fg p)).
red; intros.
apply H1 in H2.
destruct x1 as [x1 y1].
destruct x2 as [x2 y2].
unfold fg in H2.
injection H2; intros.
apply H0 in H3.
apply H in H4.
f_equal; trivial.
Qed.

Require Import FunctionalExtensionality.

(*Schepler*)
Lemma countable_exp: forall X Y:Type,
  FiniteT X -> CountableT Y -> CountableT (X->Y).
Proof.
intros.
induction H.
exists (fun _ => 0).
red; intros.
extensionality f.
destruct f.

destruct (countable_product (T->Y) Y); trivial.

exists (fun (g:option T->Y) =>
  f (fun x:T => g (Some x), g None)).
red; intros g1 g2 ?.
apply H1 in H2.
extensionality o.
destruct o.
injection H2; intros.
pose proof (equal_f H4).
simpl in H5.
apply H5.
injection H2; trivial.

destruct H1.
destruct IHFiniteT.
exists (fun (h:Y0->Y) => f0 (fun x:X => h (f x))).
red; intros h1 h2 ?.
apply H3 in H4.
pose proof (equal_f H4).
simpl in H5.
extensionality y.
rewrite <- (H2 y).
apply H5.
Qed.

(*Schepler*)
Definition Countable {X:Type} (S:Ensemble X) : Prop :=
  CountableT {x:X | In S x}.

(*Schepler*)
Lemma inj_countable: forall {X Y:Type} (f:X->Y),
  CountableT Y -> FunctionProperties.injective f -> CountableT X.
Proof.
intros.
destruct H as [g].
exists (fun x:X => g (f x)).
red; intros; auto.
Qed.

(*Schepler*)
Lemma surj_countable: forall {X Y:Type} (f:X->Y),
  CountableT X -> surjective f -> CountableT Y.
Proof.
intros.
Require Import ClassicalChoice.

pose proof (choice (fun (y:Y) (x:X) => f x = y)).
destruct H1 as [finv].
exact H0.

apply inj_countable with finv.
assumption.
red; intros.
congruence.
Qed.

(*Schepler*)
Lemma countable_downward_closed: forall {X:Type} (S T:Ensemble X),
  Countable T -> Included S T -> Countable S.
Proof.
intros.
destruct H.
exists (fun x:{x:X | In S x} => match x with
  | exist x0 i => f (exist _ x0 (H0 _ i))
  end).
red; intros.
destruct x1 as [x1].
destruct x2 as [x2].
apply H in H1.
injection H1; intros.
destruct H2.
destruct (proof_irrelevance _ i i0).
trivial.
Qed.

(*Schepler*)
Lemma countable_img: forall {X Y:Type} (f:X->Y) (S:Ensemble X),
  Countable S -> Countable (Im S f).
Proof.
intros.
assert (forall x:X, In S x -> In (Im S f) (f x)).
auto with sets.
pose (fS := fun x:{x:X | In S x} =>
  match x return {y:Y | In (Im S f) y} with
  | exist x0 i => exist _ (f x0) (H0 x0 i)
  end).
apply surj_countable with fS; trivial.
red; intros.
destruct y.
destruct i.
exists (exist _ x i).
simpl.
generalize (H0 x i); intro.
generalize (Im_intro X Y S f x i y e); intro.
destruct e.
destruct (proof_irrelevance _ i0 i1).
trivial.
Qed.

(*slightly modified from Schepler*)
Lemma countable_type_ensemble: forall {X:Type} (S:Ensemble X),
  CountableT X -> Countable S.
Proof.
intros.
red.
apply inj_countable with (@proj1_sig _ (fun x:X => In S x)).
assumption.
red.
intros a1 a2 h1.
destruct a1. destruct a2.
apply existTexist.
apply subsetT_eq_compat.
simpl in h1. assumption.
Qed.

(*Schepler*)
Lemma FiniteT_impl_CountableT: forall X:Type,
  FiniteT X -> CountableT X.
Proof.
intros.
induction H.
exists (False_rect nat).
red; intros.
destruct x1.
destruct IHFiniteT.
exists (fun x:option T => match x with
  | Some x0 => S (f x0)
  | None => 0
end).
red; intros.
destruct x1; destruct x2; try (injection H1 || discriminate H1); trivial.
intro.
apply H0 in H2.
destruct H2; trivial.

destruct IHFiniteT as [g].
destruct H0 as [finv].
exists (fun y:Y => g (finv y)).
red; intros y1 y2 ?.
apply H1 in H3.
congruence.
Qed.

(*Schepler*)
Lemma Finite_impl_Countable: forall {X:Type} (S:Ensemble X),
  Finite _ S -> Countable S.
Proof.
intros.
apply FiniteT_impl_CountableT.
apply Finite_ens_type; trivial.
Qed.

(* Schepler *)
Lemma countable_union: forall {X A:Type}
  (F:IndexedFamily A X), CountableT A ->
    (forall a:A, Countable (F a)) ->
    Countable (IndexedUnion F).
Proof.
intros.
destruct (choice_on_dependent_type (fun (a:A)
                               (f:{x:X | In (F a) x} -> nat) =>
  FunctionProperties.injective f)) as [choice_fun_inj].
intro.
destruct (H0 a).
exists f; trivial.

destruct (choice (fun (x:{x:X | In (IndexedUnion F) x}) (a:A) =>
  In (F a) (proj1_sig x))) as [choice_fun_a].
destruct x as [x [a]].
exists a.
assumption.

destruct countable_nat_product as [g].
destruct H as [h].
exists (fun x:{x:X | In (IndexedUnion F) x} =>
  g (h (choice_fun_a x), choice_fun_inj (choice_fun_a x)
                                   (exist _ (proj1_sig x) (H2 x)))).
red; intros.
apply H3 in H4.
injection H4; intros.
apply H in H6.
revert H5.
generalize (H2 x1).
generalize (H2 x2).
rewrite H6.
intros.
apply H1 in H5.
injection H5.
apply proj1_sig_injective.
Qed.


(*Wyckoff*)
Lemma countable_union2 : forall (T:Type) (S1 S2:Ensemble T), 
  Countable S1 -> Countable S2 -> Countable (Union S1 S2).
intros T S1 S2 h1 h2.
pose (option (option False)) as twoset.
pose (fun (t:twoset) => 
  match t with | None => S1 | Some _ => S2 end) as F.
assert (h3:FiniteT twoset).
  apply add_finite. apply add_finite. apply empty_finite.
pose proof (FiniteT_impl_CountableT _ h3) as h4.
assert (h5: forall t:twoset, Countable (F t)).
  destruct t; simpl; assumption.
pose proof (countable_union F h4 h5) as h6.
assert (h7:IndexedUnion F =  Union S1 S2).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x h7. 
  (* <= *) 
  inversion h7 as [i t h8].
  destruct i. simpl in h8. apply Union_intror; assumption.
  simpl in h8. apply Union_introl; assumption.
  (* >= *)
  red. intros x h7.
  pose proof (Union_inv _ _ _ _ h7) as h8.
  case h8. 
    unfold F. intro h9. 
      apply indexed_union_intro with None; assumption.
    unfold F. intro h9. 
      apply indexed_union_intro with (Some None); assumption.
rewrite h7 in h6. assumption.
Qed.
   


(*Wyckoff*)
(*A leftover from before I wrote previous lemma, included for
  conservation's sake.*)
Lemma countable_union2': forall (T:Type) (S1 S2:Ensemble T), 
  Countable S1 -> Countable S2 -> Countable (Union S1 S2).
intros T S1 S2 h1 h2.
red in h1. red in h2.
pose proof (countable_sum _ _ h1 h2) as h3.
red.
pose (fun s:{x:T | In S1 x} + {x:T | In S2 x} => match s with
  | inl x => (exist _ (proj1_sig x) (Union_introl _ S1 S2 (proj1_sig x) (proj2_sig x)))
  | inr y => (exist _ (proj1_sig y) (Union_intror _ S1 S2 (proj1_sig y) (proj2_sig y)))
end) as f.
apply (surj_countable f). assumption.
red.
intros y.
destruct y as [z h4].
pose proof (Union_inv _ _ _ z h4) as h5.
case h5.
  (*left*)
  intro h6.
  exists (inl (exist _ z h6)).
  simpl.
  pose proof (proof_irrelevance _ h4 (Union_introl T S1 S2 z h6)) as h7.
  rewrite h7. reflexivity.
  (*right*)
  intro h8.
  exists (inr (exist _ z h8)).
  simpl.
  pose proof (proof_irrelevance _ h4 (Union_intror T S1 S2 z h8)) as h9.
  rewrite h9. reflexivity.
Qed.

(*Wyckoff*)
Lemma countable_intersection: forall (T:Type) (S1 S2:Ensemble T), 
  Countable S1 -> Countable S2 -> Countable (Intersection S1 S2).
intros T S1 S2 h1 h2.
apply (countable_downward_closed (Intersection S1 S2) S1).
assumption.
auto with sets.
Qed.

(*Schepler*)
Lemma Countable_Full : forall {T:Type}, 
  CountableT T->Countable (Full_set T).
intros T h1.
red.
pose (fun (x:T) => (exist _ x (Full_intro _ x))) as f.
apply surj_countable with f. assumption.
red. intro y.
exists (proj1_sig y).
unfold f.
destruct y.
simpl.
pose proof (proof_irrelevance _ i (Full_intro T x)) as h2.
rewrite h2. reflexivity.
Qed.

(*Schepler*)
Lemma Countable_Sig : forall {T:Type} (P:T->Prop), CountableT T -> 
  CountableT {t:T|P t}.
intros T P h1.
pose [t:T|P t] as S.
pose proof (countable_downward_closed S (Full_set _)) as h2.
pose proof (Countable_Full h1) as h3.
assert (h4:Included S (Full_set _)).
  red. intros. constructor.
pose proof (h2 h3 h4) as h5.
red in h5.
assert (h6:forall x:T, In S x <-> P x).
  intro x.
  split.
  (*->*)
  intro h7.
  inversion h7. assumption.
  (*<-*)
  intro h8.
  constructor. assumption.
assert (h7:forall x:T, In S x -> P x).
  intros x h8.
  rewrite h6 in h8. assumption.

pose (fun a:{x:T | In S x} => 
  exist _ (proj1_sig a) (h7 (proj1_sig a) (proj2_sig a))) as f.
apply surj_countable with f.
assumption.

red.
intros y.
destruct y as [t h8].
assert (h9:In S t).
  rewrite h6. assumption.
exists (exist _ t (h9)).
unfold f. simpl.
pose proof (proof_irrelevance _ (h7 t h9) h8) as h10.
rewrite h10. reflexivity.
Qed.

(*Wyckoff*)
Corollary CountableT_Countable : forall (T:Type), CountableT T -> 
  (forall (S:Ensemble T), Countable S).
intros T h1 S.
red.
apply Countable_Sig; assumption.
Qed.

End CountableT.

Section FiniteT_nat_cardinal.

Lemma FiniteT_has_nat_cardinal: forall X:Type, FiniteT X ->
  exists! n:nat, cardinal _ (@Full_set X) n.
Proof.
intros.
apply -> unique_existence; split.
apply finite_cardinal.
pose (idX := fun x:X => x).
assert (Im (Full_set _) idX = Full_set _).
apply Extensionality_Ensembles.
red; split.
red; intros; constructor.
red; intros.
exists x.
constructor.
trivial.

rewrite <- H0.
apply FiniteT_img with (f:=fun x:X => x).
assumption.
intros.
case (finite_eq_dec X H y1 y2); tauto.

red; intros.
apply cardinal_unicity with X (Full_set _); trivial.
Qed.



Definition FiniteT_nat_cardinal (X:Type) (H:FiniteT X) : nat :=
  proj1_sig (constructive_definite_description _
              (FiniteT_has_nat_cardinal X H)).
Lemma FiniteT_nat_cardinal_def: forall (X:Type) (H:FiniteT X),
  cardinal _ (@Full_set X) (FiniteT_nat_cardinal X H).
Proof.
intros; unfold FiniteT_nat_cardinal.
destruct constructive_definite_description.
assumption.
Qed.

Lemma FiniteT_nat_cardinal_cond: forall (X:Type) (H:FiniteT X)
  (n:nat),
  cardinal _ (@Full_set X) n ->
  FiniteT_nat_cardinal X H = n.
Proof.
intros.
pose proof (FiniteT_has_nat_cardinal X H).
destruct H1.
red in H1.
destruct H1.
transitivity x.
symmetry; apply H2.
apply FiniteT_nat_cardinal_def.
apply H2; trivial.
Qed.

Lemma FiniteT_nat_cardinal_False:
  FiniteT_nat_cardinal False empty_finite = 0.
Proof.
apply FiniteT_nat_cardinal_cond.
assert (@Full_set False = @Empty_set False).
apply Extensionality_Ensembles; red; split; auto with sets.
red; intros.
destruct x.
rewrite H.
constructor.
Qed.

Lemma injection_preserves_cardinal: forall (X Y:Type)
  (f:X->Y) (n:nat) (S:Ensemble X), cardinal _ S n ->
  FunctionProperties.injective f -> cardinal _ (Im S f) n.
Proof.
intros.
induction H.
assert (Im (Empty_set _) f = (Empty_set _)).
apply Extensionality_Ensembles; split; auto with sets.
red; intros.
destruct H.
destruct H.
rewrite H; constructor.
assert (Im (Add A x) f = Add (Im A f) (f x)).
apply Extensionality_Ensembles; split.
red; intros.
destruct H2.
symmetry in H3; destruct H3.
destruct H2.
left; exists x0; auto with sets.
destruct H2; right; auto with sets.
red; intros.
destruct H2.
destruct H2.
exists x0.
left; auto with sets.
assumption.
destruct H2.
exists x; trivial; right; auto with sets.
rewrite H2.
constructor; trivial.
red; intro H3; inversion H3.
apply H0 in H5; destruct H5.
contradiction H1.
Qed.

Lemma FiniteT_nat_cardinal_option:
  forall (X:Type) (H:FiniteT X),
  FiniteT_nat_cardinal (option X) (add_finite X H) =
  S (FiniteT_nat_cardinal X H).
Proof.
intros.
apply FiniteT_nat_cardinal_cond.
assert ((Full_set _) =
        Add (Im (Full_set _) (@Some X)) None).
apply Extensionality_Ensembles; split.
red; intros.
destruct x.
left; exists x; constructor.
right; constructor.
red; intros; constructor.
rewrite H0.
constructor.
apply injection_preserves_cardinal.
apply FiniteT_nat_cardinal_def.
red; intros x1 x2 Heq; injection Heq; trivial.
red; intro.
inversion H1.
discriminate H3.
Qed.

Lemma FiniteT_nat_cardinal_bijection:
  forall (X Y:Type) (H:FiniteT X) (g:X->Y) (Hinv:invertible g),
    FiniteT_nat_cardinal Y (bij_finite X Y g H Hinv) =
    FiniteT_nat_cardinal X H.
Proof.
intros.
apply FiniteT_nat_cardinal_cond.
apply invertible_impl_bijective in Hinv.
destruct Hinv as [g_inj g_surj].
assert ((Full_set _)= Im (Full_set _) g).
apply Extensionality_Ensembles; split; red; intros;
  try constructor.
destruct (g_surj x).
exists x0; try constructor; auto.
rewrite H0; apply injection_preserves_cardinal; trivial.
apply FiniteT_nat_cardinal_def.
Qed.

End FiniteT_nat_cardinal.

(*This section might be deleted some time later; it was written before 
  I knew about Schepler's better formulation.  It redoes the definitions 
  and lemmas about cardinality included in the standard library, 
  only for the FiniteT Type*)


Section Cardinality_FiniteT.

(*Wyckoff*)
Inductive cardinalT : Type->nat->Prop :=
  | card_empty : cardinalT False O
  | card_add : forall (T:Type) (n:nat),   
      cardinalT T n -> cardinalT (option T) (S n)
  | card_bij : forall (T U:Type) (n:nat) (f:T->U), 
    invertible f -> cardinalT T n -> cardinalT U n.

(*Wyckoff*)
Lemma cardinalT_FiniteT : forall (T:Type) (n:nat), 
  cardinalT T n ->FiniteT T.
Proof.
intros T n h1.
induction h1. constructor.
apply (add_finite _ IHh1).
apply (bij_finite _ _ f IHh1 H).
Qed.

(*Wyckoff*)
Definition bij_eq (T:Type) (U:Type) := exists f:T->U, bijective f.

(*Wyckoff*)
Lemma bij_eq_refl : forall (T:Type), bij_eq T T.
intros T.
red. exists id.
red.
split.
(*inj*)
red. unfold id. intros; trivial.
(*surj*)
red. unfold id. intros y; exists y; reflexivity.
Qed.

(*Wyckoff*)
Lemma bij_eq_trans : forall (T U V:Type), 
  bij_eq T U -> bij_eq U V -> bij_eq T V.
intros T U V h1 h2.
red in h1. red in h2.
elim h1. intros f h3.
elim h2. intros g h4.
red.
exists (compose g f).
red.
split.
(*inj*)
red.
intros t1 t2 h5.
unfold compose in h5.
red in h4.
destruct h4 as [h4].
red in h4.
pose proof (h4 (f t1) (f t2)) as h6.
pose proof (h6 h5) as h7.
red in h3.
destruct h3 as [h3].
red in h3.
pose proof (h3 t1 t2) as h8.
apply (h8 h7).
(*surj*)
red.
intro v.
red in h4.
destruct h4 as [? h4].
red in h4.
pose proof (h4 v) as h5.
elim h5.
intros u h6.
red in h3.
destruct h3 as [? h3].
red in h3.
pose proof (h3 u) as h7.
elim h7.
intros t h8.
exists t.
unfold compose.
rewrite h8.
assumption.
Qed.

(*Wyckoff*)
Lemma bij_eq_sym : forall (T U:Type), bij_eq T U -> bij_eq U T.
intros T U.
intro h1.
red in h1.
elim h1.
intros f h2.
pose proof (bijective_impl_invertible _ h2) as h3.
inversion h3 as [g h4 h5].
red.
exists g.
red.
split.
(*inj*)
red.
intros u1 u2 h6.
pose proof (f_equal f h6) as h7.
pose proof (h5 u1) as h8.
pose proof (h5 u2) as h9.
congruence.
(*surj*)
red.
intro t.
exists (f t).
apply h4.
Qed.

(*Wyckoff*)
Lemma bij_eq_rel : Equivalence _ bij_eq.
constructor.
red.  apply bij_eq_refl.
red.  apply bij_eq_trans.
red.  apply bij_eq_sym.
Qed.

(*Wyckoff*)
Lemma cardinalT_invert : forall (T:Type) (p:nat), cardinalT T p ->
  match p with
    | 0 => bij_eq T False
    | S n => exists U : _, bij_eq T (option U )/\ cardinalT U n
  end.
induction 1 as [|T ? h1 h2|T U n f h3 h4 h5]. 
(*False*) apply bij_eq_refl.  
(*option*)exists T. split. 
apply bij_eq_refl. assumption.
(*bij*)
pose proof (invertible_impl_bijective _ h3) as h6.
destruct n.
  (* O *)
  assert (h7:bij_eq T U).
    red. exists f. assumption.
  pose proof (bij_eq_sym T U h7) as h8.
  apply (bij_eq_trans _ _ _ h8 h5).
  (* S n *)
  elim h5.
  intros V h7.
  assert (h8:bij_eq T U).
    red. exists f. assumption.
  pose proof (bij_eq_sym _ _ h8) as h9.
  destruct h7 as [h7l h7r].
  pose proof (bij_eq_trans _ _ _ h9 h7l) as h10.
  exists V. split; assumption.
Qed.

(*Wyckoff*)
Lemma cardinalT_elim : forall (T:Type) (p:nat), cardinalT T p ->
  match p with
    | O => bij_eq T False
    | S n => Inhabited (Full_set T)
  end.
intros T p h1.
pose proof (cardinalT_invert T p h1) as h2.
inversion h1 as [h3 h4 | U n h5 h6 h7 |U V n f h8 h9 h10 h11].
(*False*)  apply bij_eq_refl.
(*Option*) 
apply Inhabited_intro with None.
apply Full_intro.
(*Bij*)
destruct p.
assumption.
elim h2. intros W h12.
destruct h12 as [h12l h12r].
pose proof (bij_eq_sym _ _ h12l) as h13.
elim h13.
intros g h14.
apply Inhabited_intro with (g None).
apply Full_intro.
Qed.

(*Wyckoff*)
Lemma FiniteT_CardinalT :
  forall T:Type, FiniteT T -> exists n : nat, cardinalT T n.
intros T h1.
induction h1 as [?|T h2 h3|T U f h5 h6 h7].
exists 0. constructor.
elim h3. intros n h4.
exists (S n).
constructor. assumption.
elim h6. intros n h8.
exists n. 
apply (card_bij _ _ n f h7 h8).
Qed.

(*Wyckoff*)
Lemma cardinalTO_False : forall T:Type, cardinalT T 0 -> bij_eq T False.
intros T h1.
apply (cardinalT_invert _ 0 h1).
Qed.

(*Wyckoff*)
Lemma inh_cardT_gt_O : forall T:Type, Inhabited (Full_set T) ->
  forall n:nat, cardinalT T n -> n > 0.
intros T h1 n h2.
induction n as [|n h6].
(* O *)
inversion h1.
pose proof (cardinalTO_False) as h3.
pose proof (h3 _ h2) as h4.
red in h4.
elim h4.
intros f h5.
pose (f x).
contradiction.
(* S n*)
apply gt_Sn_O.
Qed.
End Cardinality_FiniteT.

Section Index_of_Finite_Set.
Definition f_add {T:Type} (A:Ensemble T) (n:nat) 
  (f:{m:nat | m < n} -> {x:T | In A x}) (t:T) : 
    {m:nat | m < S n} -> {x:T| In (Add A t) x}.
intro a.
inversion a as [m h1].
red in h1.
pose proof (le_S_n m n h1) as h2.
case (le_lt_eq_dec _ _ h2).
(* m < n *)
intro h3.
refine (exist _ (proj1_sig (f (exist _ m h3))) _).
apply Add_intro1. apply proj2_sig. 
(*m = n*)
intro h3. 
pose proof (Add_intro2 _ A t) as h4.
refine (exist _ t h4).
Defined.

Lemma index_element' : forall {T:Type} (A:Ensemble T) (n:nat), 0 < n -> 
  cardinal _ A n -> 
  exists ind:{m:nat | m < n}->{x:T|In A x}, FunctionProperties.injective ind.
intros T A n h1 h2.
induction h2 as [|A n h3 h4 t h5].
pose proof (lt_n_0 0). contradiction.
pose proof (Add_intro2 _ A t) as h6.
case (classic (n = 0)).
(* n = 0 *)
intro h7.
subst.
exists (fun a:{m:nat | m < 1} => (exist _ t h6)).
red. intros x y h7.
destruct x as [x h8]. destruct y as [y h9].
pose proof (lt_n_1 _ h8) as h10. pose proof (lt_n_1 _ h9) as h11.
rewrite <- h11 in h10. 
apply existTexist. apply subsetT_eq_compat; assumption.
(* n > 1*)
intro h7.
assert (h8: n > 0). omega.
pose proof (h4 h8) as h9.
elim h9. intros ind h10.
exists (f_add A n ind t).
red.
intros x y h11.
unfold f_add in h11.
destruct x as [x h12]. destruct y as [y h13].
destruct (le_lt_eq_dec x n (le_S_n x n h12)) as [h14|h15].
destruct (le_lt_eq_dec y n (le_S_n y n h13)) as [h16|h17].
(* x < n, y < n *)
pose proof (f_equal (@proj1_sig _ _) h11) as h18.
simpl in h18.
pose proof (proj1_sig_injective _ _ _ h18) as h19.
red in h10.
pose proof (h10 _ _ h19) as h20.
pose proof (f_equal (@proj1_sig _ _) h20) as h21.
simpl in h21.
apply existTexist.
apply subsetT_eq_compat. assumption.
(* x < n, y = n*)
pose proof (f_equal (@proj1_sig _ _) h11) as h18.
simpl in h18.
assert (h19: In A t).
  rewrite <- h18.  apply proj2_sig.
contradiction.
destruct (le_lt_eq_dec y n (le_S_n y n h13)) as [h16|h17].
(* x = n, y < n *)
pose proof (f_equal (@proj1_sig _ _) h11) as h18.
simpl in h18.
assert (h19: In A t).
  rewrite h18. apply proj2_sig.
contradiction.
(* x = n, y = n*)
rewrite <- h17 in h15.
apply existTexist.
apply subsetT_eq_compat. assumption.
Qed.

Lemma card_init_segment : forall (n:nat), let A := [m:nat | m < n] in
  cardinal _ A n.
intros n A.
induction n as [|n h2].
(* 0 *)
assert (h1: A = (Empty_set _)).
  apply Extensionality_Ensembles. red. split.
  (* <= *)
  red. intros x h1. inversion h1. omega.
  (* >= *)
  auto with sets.
rewrite h1. constructor.
(* S n *)
assert (h3: A = Add [m:nat | m < n] n).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h3.
  inversion h3 as [h4].
  red in h4.
  pose proof (le_S_n _ _ h4) as h5.
  destruct (le_lt_eq_dec _ _ h5) as [h6 | h7].
    (* < *)
    constructor. constructor. assumption.
    (* = *)
    rewrite h7.
    apply Add_intro2.
  (* >= *)
  red. intros m h3.
  pose proof (Add_inv _ _ _ _ h3) as h4.
  destruct h4 as [h5 | h6].
    inversion h5. constructor. omega.
    rewrite <- h6.
    constructor. omega.
rewrite h3.
assert (h4: ~(In [m:nat | m < n] n)).
  intro h5. inversion h5. omega.
constructor. assumption. assumption.
Qed.

Lemma card_full : forall {T:Type} (B:Ensemble T) (n:nat), cardinal _ B n <->
  cardinal _ (Full_set {t :T | In B t}) n.
intros T B n. 
split.
(* -> *) 
intro h1.
pose proof (cardinal_finite _ _ _ h1) as h2.
pose proof (Finite_ens_type _ h2) as h3.
pose proof (FiniteT_has_nat_cardinal _ h3) as h4.
elim h4. intros m h5.
red in h5. destruct h5 as [h5l h5r].
pose proof (injection_preserves_cardinal _ _ (@proj1_sig _ _) 
  m (Full_set {t:T | In B t}) h5l) as h6.
assert (h7: FunctionProperties.injective (@proj1_sig T (fun t:T => In B t))). 
  red. apply proj1_sig_injective.
pose proof (h6 h7) as h8.
pose proof (full_sig B) as h9.
rewrite <- h9 in h8.
pose proof (cardinal_unicity _ _ _ h1 _ h8) as h10.
rewrite <- h10 in h5l.
assumption.
(* -> *)
intro h1.
pose proof (injection_preserves_cardinal _ _ (@proj1_sig _ _) n 
  _ h1) as h2.
assert (h3: FunctionProperties.injective 
  (@proj1_sig T (fun t:T => In B t))).
  red. apply proj1_sig_injective.
pose proof (h2 h3) as h4.
pose proof (full_sig B) as h5.
rewrite <- h5 in h4. assumption.
Qed.

Lemma card_initial_segment_sig : forall (n:nat), 
  cardinal _ (Full_set {m:nat | m < n}) n.
intro n.
pose proof (card_init_segment n) as h1.
pose proof (card_full [m:nat | m < n] n) as h2.
rewrite h2 in h1.
pose (fun m:{t:nat | In [m:nat | m < n] t} => 
  exist _ (proj1_sig m) match (proj2_sig m) with
  | intro_characteristic_sat P => P end) as f.
pose proof (injection_preserves_cardinal _ _ f n _ h1) as h3.
assert (h4:FunctionProperties.injective f).
  unfold f.
  red. intros x1 x2 h4.
  pose proof (exist_injective _ _ _ _ _ h4) as h5.
  apply proj1_sig_injective. assumption.
pose proof (h3 h4) as h5.
assert (h6: Im (Full_set {t:nat | In [m:nat | m < n] t}) f = 
  (Full_set {m:nat | m < n})).
  apply Extensionality_Ensembles.
  red. split.
  (* <= *)
  red. intros x h6. constructor.
  (* >= *)
  red. intros x h6.
  assert (h7: In [m:nat | m < n] (proj1_sig x)).
    constructor. apply proj2_sig.
  assert (h8: x = f (exist _ (proj1_sig x) h7)).
    unfold f.
    simpl.
    destruct x.
    simpl.
    apply existTexist.
    apply subsetT_eq_compat. reflexivity.
  assert (h9: In (Full_set _ ) (exist _ (proj1_sig x) h7)).
    constructor.
  apply (Im_intro _ _ (Full_set _) f _ h9 _ h8).
rewrite <- h6.
assumption.
Qed.

Lemma cardinal_inj_impl_surj : forall {T:Type} (A:Ensemble T) (n:nat)
  (f:{m:nat | m < n}->{x:T|In A x}), FunctionProperties.injective f -> 
  cardinal _ A n -> surjective f.
intros T A n f h1 h2.
apply NNPP.
intro h3.
pose proof (surj_eq f) as h4.
rewrite h4 in h3.
pose proof (full_sig A) as h6.
assert (h7:Strict_Included 
  (Im (Full_set {m:nat | m < n}) f) 
  (Full_set {x:T | In A x})).
  red. split.
  (*incl*)
  red. intros x ?. apply Full_intro. 
  (* <> *) assumption.
pose proof (card_initial_segment_sig n) as h8.
pose proof (injection_preserves_cardinal _ _ _ _ _ h8 h1) as h9.
pose proof (card_full A n) as h10.
rewrite h10 in h2.
pose proof (incl_st_card_lt _ _ _ h9 _ _ h2 h7).
omega.
Qed.

Lemma index_element : forall {T:Type} (A:Ensemble T) (n:nat), 0 < n -> 
  cardinal _ A n -> 
  exists ind_el:{m:nat | m < n}->{x:T|In A x}, bijective ind_el.
intros T A n h1 h2.
pose proof (index_element' A n h1 h2) as h3.
elim h3.
intros f h4.
exists f.
split. assumption. apply cardinal_inj_impl_surj; assumption.
Qed. 

Lemma element_index : forall {T:Type} (A:Ensemble T) (n:nat), 0 < n ->
  cardinal _ A n -> exists el_ind:{x:T|In A x}->{m:nat | m < n}, 
  bijective el_ind.
intros T A n h1 h2.
pose proof (index_element A n h1 h2) as h3.
elim h3. intros ind_el h4.
pose proof (bijective_impl_invertible _ h4) as h5.
inversion h5 as [el_ind h6 h7].
assert (h8: invertible el_ind).
  apply intro_invertible with ind_el; assumption.
pose proof (invertible_impl_bijective _ h8) as h9.
exists el_ind; assumption.
Qed.

Definition ind_el {T:Type} (A:Ensemble T) (n:nat) (pf0:0 < n)  
  (pfc:cardinal _ A n) :
  {f:{m:nat | m < n}->{x:T|In A x} |  bijective f}.
refine (constructive_indefinite_description _ 
(index_element A n pf0 pfc)).
Qed. (*sic*)

End Index_of_Finite_Set.
